-- Premake script

addoption("enable-debug", "Enables program debugging")
addoption("force-console", "Forces use of the background console in Windows")
addoption("static-link", "Uses the static-link version of Allegro in Windows")
addoption("extra-opts", "Adds options to pass, untranslated, to the compiler")


project.name = "ZETH"

project.bindir = "."
project.objdir = "obj"

package.name = "ZETH"
package.language = "c++"
package.kind = "winexe"
package.target = "zeth"

package.includepaths = { "src" }

package.files = {
	matchfiles("src/*.c", "src/*.cpp")
}


if(OS == "windows") then
	if(options["static-link"]) then
		package.defines = {
			"ALLEGRO_STATICLINK"
		}
		package.config["Debug"].links = { "alld_s", "kernel32", "user32", "gdi32", "comdlg32", "ole32", "dinput", "ddraw", "dxguid", "winmm", "dsound" }
		package.config["Release"].links = { "alleg_s", "kernel32", "user32", "gdi32", "comdlg32", "ole32", "dinput", "ddraw", "dxguid", "winmm", "dsound" }
	else
		package.config["Debug"].links = { "alld" }
		package.config["Release"].links = { "alleg" }
	end
else
	package.config["Debug"].linkoptions = { "`allegro-config --libs debug`" }
	package.config["Release"].linkoptions = { "`allegro-config --libs`" }
end

package.links = { "png", "z", "m" }

package.config["Release"].linkflags = {
	"strip"
}


package.buildflags = {
	"warnings",
	"extra-warnings"
}

if (options["extra-opts"]) then
	package.buildoptions = {
		options["extra-opts"][1]
	}
end

package.config["Release"].buildflags = {
	"optimize-speed",
	"no-frame-pointer"
}

package.config["Debug"].buildflags = {
	"fatal-warnings",
	"with-symbols"
}

if (OS == "windows") then
	if (options["force-console"]) then
		package.defines = {
			"ALLEGRO_USE_CONSOLE"
		}
	end
else
	package.defines = { "strnicmp=strncasecmp" }
end

package.config["Debug"].defines = {
	"DEBUGMODE"
}
