/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.

   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.

   This program comes with absolutely no warranty.

   DEU.C - Main program execution routines.
*/


// the includes
#include <time.h>
#include <signal.h>
#include <limits.h>
#include <errno.h>

#include <stdio.h>

#include <sys/stat.h>

#include "deu.h"

void add_decorate(void);

extern "C" {

// global variables
extern int Fullscreen;
FILE  *logst =					NULL;	// filepointer to the debug log jff
FILE  *logfile =				NULL;	// filepointer to the error log
BCINT InitialScale = 			8;		// initial zoom factor for map
BCINT VideoMode = 				2;		// default video mode for VESA/SuperVGA
BCINT RealVideoMode = 			3;		// default video mode for VESA/SuperVGA
BCINT DefaultLargeScroll = 		10;		// default large scroll factor

BCINT DefaultFloorHeight = 		0;		// default floor height
BCINT DefaultCeilingHeight = 	128;	// default ceiling height
BCINT DefaultLighting =			128;	// default sector lighting

char  DefaultWallTexture[9] = 		"STARTAN3";	/* default normal wall texture */
char  DefaultUpperTexture[9] = 		"STARTAN3";	/* default upper wall texture */
char  DefaultLowerTexture[9] = 		"STARTAN3"; /* default lower wall texture */
char  DefaultFloorTexture[9] = 		"FLOOR7_1";	/* default floor texture */
char  DefaultCeilingTexture[9] = 	"RROCK09\0";/* default ceiling texture */

char  DefaultDoorBase[9] =			"-\0\0\0\0\0\0\0";
char  DefaultDoorTrack[9] = 		"DOORTRAK";		/* default door track texture */
char  DefaultDoorTexture[9] = 		"DOOR1\0\0\0";	/* default door texture */

char  DefaultUpperStepTexture[9] = 	"STEP5\0\0\0";	/* default step texture */
char  DefaultLowerStepTexture[9] = 	"STEP5\0\0\0";	/* default step texture */

char  DefaultPlatTexture[9] =		"SHAWN2\0\0";	/* default plat (lift) texture */

TeleportPrefab DefaultTeleport;
TeleportPrefab DefaultTeleDest;

char  *MainWad =				NULL; 	/* name of the main wad file */
BCINT CloseToLine = 			1;
BCINT TempCloseToLine = 		1;
BCINT LastLineDefTag = 			0;
BCINT LastSectorTag = 			0;
BCINT Increment = 				16;
BCINT MinGrid = 				4;
BCINT MaxGrid = 				512;
BCINT InitGrid = 				64;
BCINT GridScale = 				0;
BCINT Gamma = 					0;
BCINT StretchDelay =			5;
BCINT Count = 					0;
BCINT Sensitivity;
BCINT MoveSpeed = 				10;
BCINT MenuDefault =				0;
BCINT NumPatchWads = 			0;
BCINT NumConfigPatchWads =		0;
BCINT EdMode = 					0;
int	  NumSingleStarts;
Bool  ChooseGame = 				TRUE;
Bool  GraphFront = 				TRUE;
Bool  Registered = 				FALSE;		/* registered or shareware game? */
Bool  Debug = 					FALSE;		/* are we debugging? */
Bool  SwapButtons =				FALSE;		/* swap right and middle mouse buttons */
Bool  Quiet = 					FALSE;		/* don't play a sound when an object is selected */
Bool  Quieter =					FALSE;		/* don't play any sound, even when an error occurs */
Bool  Expert = 					FALSE;		/* don't ask for confirmation for some operations */
Bool  Colour2 = 				FALSE;		/* use the alternate set for things colors */
Bool  InfoShown = 				TRUE;		/* should we display the info bar? */
Bool  ShowSprite = 				FALSE;	
Bool  DraggingObject = 			FALSE;
Bool  BrowseMap = 				FALSE;
Bool  ShowDistance = 			FALSE;
Bool  AdditiveSelBox = 			FALSE;		/* additive selection box or select in box only? */
Bool  ThingAngle = 				FALSE;   	/* draw things with directional arrow */
Bool  DispSprite = 				TRUE;
Bool  AltInfo = 				TRUE;
Bool  VertConf = 				FALSE;		/* don't ask for confirmation for some operations */
Bool  QisQuit = 				FALSE;		
Bool  Select0 = 				TRUE;		/* select object 0 by default when switching modes */
Bool  Reminder = 				FALSE;		/* display a funny message when HETH starts */
Bool  Failed = 					FALSE;
Bool  CheckReport =				FALSE;		/* just list errors, don't stop on them */
Bool  DoingBehavior =			FALSE;
Bool  KeepInfo = 				FALSE;
Bool  Caching = 				FALSE;
Bool  UsePromptOnly = 			FALSE;
Bool  GridShown = 				TRUE;
Bool  FloorGridShown =			TRUE;
Bool  GridAccurate =			TRUE;
Bool  NestMenus = 				TRUE;
Bool  AutoMap = 				FALSE;
Bool  ShowThings = 				FALSE;
Bool  GridDashed = 				FALSE;
Bool  SaveIniFile = 			TRUE;
Bool  CopySideDefs =			FALSE;
Bool  CreatingNewLevel =		FALSE;
Bool  DeleteSideLines =			TRUE;		/* can deleting sidedefs possibly delete lines? */
Bool  DeleteLineVerts =			TRUE;		/* can deleting linedefs possibly delete vertices? */
Bool  SpecialEffects =			FALSE;		/* turn off some error checks */
Bool  SmoothMouseScroll =		FALSE;
Bool  PersistantMenus =			TRUE;
Bool PauseBeforeCompile =		FALSE;	// 2000/01/29 & 2002/12/08 Ty - Notepad returns immediately--need an optional pause
char *CfgFile = 				NULL;//DEU_CONFIG_FILE;	/* name of the configuration file */
char **PatchWads = 				NULL;		/* list of patch wad files */
char **ConfigPatchWads =		NULL;		/* list of patch wad files */
char BackupText[64] = 			"Nothing To Undo";
char *Doom = 					NULL;
char *Doom2 = 					NULL;
char *Heretic = 				NULL;
char *Hexen = 					NULL;
char *Strife = 					NULL;
char *Game = 					NULL;//"DOOM2";
char LevelName[9] = 			"\0\0\0\0\0\0\0\0";
char *RegTest = 				NULL;
char *CurrentScheme =			NULL;
char *DoomPwads =				NULL;
char *Doom2Pwads =				NULL;
char *HereticPwads =			NULL;
char *StrifePwads =				NULL;
char *HexenPwads =				NULL;
char **Pwads =					NULL;
char Cwd[256] =					".";
char *TextEditor =				NULL;
char *ACCDirectory =			NULL;
char *ACCCommand =				NULL;
Bool NonHexenWarning =			TRUE;
Bool ShowPaths =				TRUE;
Bool DetailScroll =				TRUE;
Bool HideGrid;	/* for compatibility with old zeth.ini */
Bool AlwaysMScroll =			FALSE;
Bool SeperateTypes =			TRUE;
char exe_path[256];

SList LevelNameFormat = 			(SList)NULL;
OptDesc options[] =				/* description of the command line options */
{
	//  short & long names  type            message if true/changed        message if false    where to store the value
	//--------------------------------------------------------------------------------------------------------------------
    { "f",   "file",		OPT_STRINGLIST, "Patch WAD file",              NULL,                         &PatchWads },
    { "pw", "pwad",			OPT_STRINGACC,  "Patch WAD file",              NULL,                         &PatchWads },
	{ "",	"specialeffects",OPT_BOOLEAN,	"Special Effect Checking ON",  "Special Effect Checking OFF",&SpecialEffects },
	{ "",	"texteditor",	OPT_STRING,		NULL,						   NULL,						 &TextEditor },
	{ "",	"accdirectory", OPT_STRING,		NULL,						   NULL,						 &ACCDirectory },
	{ "",	"acccommand",	OPT_STRING,		NULL,						   NULL,						 &ACCCommand },
    { "",   "config",      	OPT_STRING,     "Config file",                 NULL,                         &CfgFile },
    { "",	"main",        	OPT_STRING,     "Main WAD file",               NULL,                         &MainWad },
    { "",	"doom",	   		OPT_STRING,     NULL,						   NULL,                         &Doom },
    { "",	"doom2",	   	OPT_STRING,     NULL,						   NULL,                         &Doom2 },
    { "",	"heretic",	   	OPT_STRING,     NULL,						   NULL,                         &Heretic },
    { "",	"hexen",	  	OPT_STRING,     NULL,						   NULL,                         &Hexen },
    { "",	"strife",	  	OPT_STRING,     NULL,						   NULL,                         &Strife },
    { "",	"game",	   		OPT_STRING,     NULL,						   NULL,                         &Game },
    { "",	"doompwads",	OPT_STRING,     NULL,						   NULL,                         &DoomPwads },
    { "",	"doom2pwads",	OPT_STRING,     NULL,						   NULL,                         &Doom2Pwads },
    { "",	"hereticpwads",	OPT_STRING,     NULL,						   NULL,                         &HereticPwads },
    { "",	"strifepwads",	OPT_STRING,     NULL,						   NULL,                         &StrifePwads },
    { "",	"hexenpwads",	OPT_STRING,     NULL,						   NULL,                         &HexenPwads },
	{ "",	"pwads",		OPT_STRINGLIST,	"Patch WAD files",			   NULL,						 &ConfigPatchWads },
    { "z",	"zoom",        	OPT_INTEGER,    "Initial zoom factor",         NULL,                         &InitialScale },
    { "",	"closetoline",  OPT_INTEGER,    "Close to Line to Split",	   NULL,                         &CloseToLine },
    { "v",  "video",       	OPT_INTEGER,    "Default video mode",          NULL,                         &VideoMode },
    { "",   "fullscreen",   OPT_BOOLEAN,    "",                            NULL,                         &Fullscreen },
	{ "",   "increment",   	OPT_INTEGER,    NULL,						   NULL,						 &Increment },
	{ "",   "mingrid",     	OPT_INTEGER,    NULL,						   NULL,						 &MinGrid },
	{ "",   "maxgrid",     	OPT_INTEGER,    NULL,						   NULL,						 &MaxGrid },
	{ "",   "initgrid",   	OPT_INTEGER,    NULL,						   NULL,						 &InitGrid },
	{ "",	"lighting",		OPT_INTEGER,	"Default lighting",			   NULL,						 &DefaultLighting },
    { "",   "floorheight", 	OPT_INTEGER,    "Default floor height",        NULL,                         &DefaultFloorHeight },
    { "",   "ceilheight",  	OPT_INTEGER,    "Default ceiling height",      NULL,                         &DefaultCeilingHeight },
    { "",   "largescroll", 	OPT_INTEGER,    "Default large scroll factor", NULL, 						 &DefaultLargeScroll },
	{ "",   "gamma",   		OPT_INTEGER,    "",							   "",							 &Gamma },
    { "",   "stretchdelay",	OPT_INTEGER,    "",							   NULL,						 &StretchDelay },
    { "",   "edmode",		OPT_INTEGER,    "",							   NULL,						 &EdMode },
    { "",	"choosegame",  	OPT_BOOLEAN,    "",							   NULL,						 &ChooseGame },
    { "",	"graphfront",  	OPT_BOOLEAN,    "",							   NULL,						 &GraphFront },
    { "d",  "debug",       	OPT_BOOLEAN,    "Debug mode ON",               "Debug mode OFF",             &Debug },
    { "q",	"quiet",       	OPT_BOOLEAN,    "Quiet mode ON",               "Quiet mode OFF",             &Quiet },
    { "qq",	"quieter",     	OPT_BOOLEAN,    "Quieter mode ON",             "Quieter mode OFF",           &Quieter },
    { "e",	"expert",      	OPT_BOOLEAN,    "Expert mode ON",              "Expert mode OFF",            &Expert },
    { "",	"qisquit",    	OPT_BOOLEAN,    "Quit on Q",                   "Save on Q",                  &QisQuit },
    { "sb",	"swapbuttons", 	OPT_BOOLEAN,    "Mouse buttons swapped",       "Mouse buttons restored",     &SwapButtons },
    { "",	"vertconf",    	OPT_BOOLEAN,    "Confirm Vertices Merge",      NULL,            			 &VertConf },
    { "",	"select0",     	OPT_BOOLEAN,    "Select 0 by default",         "No default selection",       &Select0 },
    { "", 	"thingangle",  	OPT_BOOLEAN,    "Show Thing angle",            NULL,						 &ThingAngle },
    { "i",	"infobar",     	OPT_BOOLEAN,    "Info bar shown",              "Info bar hidden",            &InfoShown },
    { "a",	"addselbox",   	OPT_BOOLEAN,    "Additive selection box",      "Select objects in box only", &AdditiveSelBox },
	{ "",	"dispsprite",  	OPT_BOOLEAN,    "Show graphics",   		  	   "",							 &DispSprite },
	{ "",	"altinfo",     	OPT_BOOLEAN,    "Using alternative info bar",  "",							 &AltInfo },
	{ "",	"keepinfo",		OPT_BOOLEAN,    "",							   "",							 &KeepInfo },
	{ "",	"caching", 		OPT_BOOLEAN,    "",							   "",							 &Caching },
	{ "",	"usepromptonly",OPT_BOOLEAN,    "",							   "",							 &UsePromptOnly },
	{ "",	"nestmenus",	OPT_BOOLEAN,    "",							   "",							 &NestMenus },
	{ "",	"automap",		OPT_BOOLEAN,    "",							   "",							 &AutoMap },
	{ "",	"showthings",	OPT_BOOLEAN,    "",							   "",							 &ShowThings },
	{ "",	"griddashed",	OPT_BOOLEAN,    "",							   "",							 &GridDashed },
	{ "",	"gridshown",	OPT_BOOLEAN,    "",							   "",							 &GridShown },
	{ "",	"floorgridshown", OPT_BOOLEAN,	"",							   "",							 &FloorGridShown },
	{ "",	"gridaccurate",	OPT_BOOLEAN,	"",							   "",							 &GridAccurate },
	{ "",	"saveinifile",	OPT_BOOLEAN,    "",							   "",							 &SaveIniFile },
	{ "",	"showdistance",	OPT_BOOLEAN,    "",							   "",							 &ShowDistance },
	{ "",	"copysidedefs",	OPT_BOOLEAN,    "",							   "",							 &CopySideDefs },
	{ "",	"nonhexenwarning", OPT_BOOLEAN,	NULL,						   NULL,						 &NonHexenWarning },
	{ "",	"smoothscroll", OPT_BOOLEAN,	NULL,						   NULL,						 &SmoothMouseScroll },
	{ "",	"drawthingcircles", OPT_BOOLEAN, NULL,						   NULL,						 &DrawThingCircles },
	{ "",	"showpaths",	OPT_BOOLEAN,	NULL,						   NULL,						 &ShowPaths },
	{ "",	"detailscroll",	OPT_BOOLEAN,	NULL,						   NULL,						 &DetailScroll },
	{ "",	"deletelineverts", OPT_BOOLEAN,	NULL,						   NULL,						 &DeleteLineVerts },
	{ "",	"deletesidelines", OPT_BOOLEAN, NULL,						   NULL,						 &DeleteSideLines },
	{ "",	"hidegrid",		OPT_BOOLEAN,	NULL,						   NULL,						 &HideGrid },
	{ "",	"persistantmenus", OPT_BOOLEAN,	NULL,						   NULL,						 &PersistantMenus },
	{ "",   "pausebeforecompile", OPT_BOOLEAN, "Pause before script compile step", "Compile immediately after text editor exits", &PauseBeforeCompile }, // 2000/01/29 & 2002/12/08 Ty
	{ "",	"alwaysmscroll",	OPT_BOOLEAN, NULL,						   NULL,						 &AlwaysMScroll },
	{ "",	"seperatetypes",	OPT_BOOLEAN, NULL,						   NULL,						 &SeperateTypes },
    { NULL, NULL,          	OPT_END,        NULL,						   NULL,                         NULL }
};

// close request callback function
int quitnow;
static void close_callback(void)
{
	quitnow = 1; 
}


DATAFILE *zethdat;
extern BITMAP *mouse_bmp;

// Find a file, depending on where we are and what's available
char *find_file(const char *filename)
{
	static char fname[256];

	memset(fname, 0, sizeof(fname));

	sprintf(fname, "%s%s", exe_path, filename);
    if(file_exists(fname, ~FA_DIREC, NULL))
		return fname;

	sprintf(fname, "./%s", filename);
	if(file_exists(fname, ~FA_DIREC, NULL))
		return fname;

#ifdef ALLEGRO_UNIX
	sprintf(fname, "~/.zeth/%s", filename);
	return fname;
#else
	sprintf(fname, "%s%s", exe_path, filename);
	return fname;
#endif
}

// Cleanup
static void cleanup(void)
{
	ForgetAllResources();

	if(zethdat)
	{
		unload_datafile(zethdat);
		zethdat = NULL;
	}

	set_mouse_sprite(NULL);
	if(mouse_bmp)
	{
		destroy_bitmap(mouse_bmp);
		mouse_bmp = NULL;
	}

	delete video;
	video = NULL;

	if(Doom)
	{
		free(Doom);
		Doom = NULL;
	}
	if(Doom2)
	{
		free(Doom2);
		Doom2 = NULL;
	}
	if(Heretic)
	{
		free(Heretic);
		Heretic = NULL;
	}
	if(Hexen)
	{
		free(Hexen);
		Hexen = NULL;
	}
	if(Strife)
	{
		free(Strife);
		Strife = NULL;
	}
	if(Game)
	{
		free(Game);
		Game = NULL;
	}
	if(PatchWads)
	{
		int i = 0;
		while(PatchWads[i])
		{
			free(PatchWads[i]);
			++i;
		}
		free(PatchWads);
		PatchWads = NULL;
	}
	if(ConfigPatchWads)
	{
		int i = 0;
		while(ConfigPatchWads[i])
			free(ConfigPatchWads[i]);
		free(ConfigPatchWads);
		ConfigPatchWads = NULL;
	}

	if(RegTest)
	{
		free(RegTest);
		RegTest = NULL;
	}
	if(CurrentScheme)
	{
		free(CurrentScheme);
		CurrentScheme = NULL;
	}
	if(DoomPwads)
	{
		free(DoomPwads);
		DoomPwads = NULL;
	}
	if(Doom2Pwads)
	{
		free(Doom2Pwads);
		Doom2Pwads = NULL;
	}
	if(HereticPwads)
	{
		free(HereticPwads);
		HereticPwads = NULL;
	}
	if(HexenPwads)
	{
		free(HexenPwads);
		HexenPwads = NULL;
	}
	if(StrifePwads)
	{
		free(StrifePwads);
		StrifePwads = NULL;
	}
	if(TextEditor)
	{
		free(TextEditor);
		TextEditor = NULL;
	}
	if(ACCDirectory)
	{
		free(ACCDirectory);
		ACCDirectory = NULL;
	}
	if(ACCCommand)
	{
		free(ACCCommand);
		ACCCommand = NULL;
	}
	if(CfgFile)
	{
		free(CfgFile);
		CfgFile = NULL;
	}

	chdir(Cwd);
}

static void ParseConfigFileOptions(const char *filename);
static void ParseCommandLineOptions( int argc, const char *argv[]);

//
// the main program
//
int main(int argc, const char *argv[])
{
    int i;
	char **p, *t;
	char *usedCfg = NULL;
	Bool ask = TRUE;

	if(allegro_init() != 0)
	{
		fprintf(stderr, "Error initializing Allegro!\n");
		exit(-1);
	}

	if(install_timer() != 0)
	{
		allegro_message("Error installing timer!\n");
		exit(1);
	}

	set_window_title("ZETH v"DETH_VERSION);

#ifdef ALLEGRO_UNIX
	if(!file_exists("~/.zeth", 0xFF, &i) || !(i & FA_DIREC))
		mkdir("~/.zeth", 644);
#endif

	get_executable_name(exe_path, sizeof(exe_path));
	if((t = strrchr(exe_path, '/')) != NULL || (t = strrchr(exe_path, '\\')) != NULL)
		t[1] = 0;
	else
		strcpy(exe_path, "./");

	// Load zeth.dat
	zethdat = load_datafile(find_file("zeth.dat"));
	if(!zethdat)
	{
		allegro_message("Unable to load zeth.dat!\n");
		exit(1);
	}

	atexit(cleanup);

    argv++;
    argc--;

    /* quick and dirty check for a "-config" and/or "-noask" option */
    for (i = 0; i < argc; i++)
	{
		if(stricmp(argv[i], "-config") == 0 && i < argc-1)
		{
			free(CfgFile);
			CfgFile = strdup(argv[i + 1]);
		}
		if(stricmp(argv[i], "-noask") == 0)
			ask = FALSE;
	}

	if(!CfgFile)
		CfgFile = strdup(DEU_CONFIG_FILE);

	/* read config file and command line options */
	ParseConfigFileOptions(CfgFile);
	ParseCommandLineOptions(argc, argv);

	strupr(Game);
#ifdef ALLEGRO_CONSOLE_OK
	if (ChooseGame && ask)
	{
		char input[10];

		Credits(stdout);
		printf("\n[Enter game or hit enter]> ");

		if(fgets(input, sizeof(input), stdin))
		{
			strupr(input);
			if (strchr(input, '2'))
				strcpy(Game, "DOOM2");
			else if (input[0] == 'D' || input[0] == '1')
				strcpy(Game, "DOOM");
			else if (input[0] == '3' || (input[0] == 'H' && !strchr(input, 'X')))
				strcpy(Game, "HERETIC");
			else if (input[0] == '4' || (input[0] == 'H' && strchr (input, 'X')))
				strcpy(Game, "HEXEN");
			else if (input[0] == 'S' || input[0] ==  '5')
				strcpy(Game, "STRIFE");
			else if (strchr(input, 'Q'))
				return 0; 
		}
	}
#endif

    if (stricmp(Game, "DOOM2") == 0)
	{
		readcfg(usedCfg = Doom2);
		readcfg("doom2.fab");
		Pwads = &Doom2Pwads;
	}
	else if (stricmp(Game, "DOOM") == 0)
	{
		readcfg(usedCfg = Doom);
		readcfg("doom.fab");
		Pwads = &DoomPwads;
	}
	else if (stricmp(Game, "HERETIC") == 0)
	{
		readcfg(usedCfg = Heretic);
		readcfg("heretic.fab");
		Pwads = &HereticPwads;
	}
	else if (stricmp(Game, "HEXEN") == 0)
	{
		readcfg(usedCfg = Hexen);
		readcfg("hexen.fab");
		Pwads = &HexenPwads;
	}
	else if (stricmp(Game, "STRIFE") == 0)
	{ /* Ha ha! As if Strife was ever supported by DETH! */
		readcfg(usedCfg = Strife);
		readcfg("strife.fab");
		Pwads = &StrifePwads;
	}
	else
		ProgError ("Unsupported Game %s\n", Game);

	LogMessage(": Welcome to ZETH!\n");

    if (Quieter)
		Quiet = TRUE;

#ifdef ALLEGRO_CONSOLE_OK
	// [CR] Check if the PWAD dir exists
	if(ask)
	{
		while(!(*Pwads) || !file_exists(*Pwads, ~0, &i) || !(i&FA_DIREC))
		{
			char wad[256] = "";
			int namelen;

			if(*Pwads)
				strcpy(wad, *Pwads);

			fprintf (stderr, "\n%s doesn't exist or is not a directory.\nPlease enter the full path to your PWAD directory:\n", wad);
			fprintf (stderr, "(Just press Enter for current directory)\n> ");
			if (!fgets(wad, sizeof(wad), stdin))
				ProgError ("Error requesting pwad directory: %s", strerror(errno));

			namelen = strlen(wad);
			while(namelen > 0 && (wad[namelen-1] == '\n' || wad[namelen-1] == '\r' ||
			                      wad[namelen-1] == '\\' || wad[namelen-1] == '/'))
				wad[--namelen] = 0;

			if(!namelen)
				strcpy(wad, ".");

			if(*Pwads)
				free(*Pwads);
			*Pwads = strdup(wad);
		}

		if(!GraphFront)
			Credits(stdout);
	}
	else
#endif
	if(!(*Pwads) || !file_exists(*Pwads, FA_DIREC, &i))
	{
		if(*Pwads)
			free(*Pwads);
		*Pwads = strdup(".");
	}

	getcwd(Cwd, sizeof(Cwd));
	chdir(*Pwads); 

#ifdef ALLEGRO_CONSOLE_OK
	if(ask)
	{
		/* [RH] check if the iwad is present */
		if(!MainWad || !file_exists(MainWad, 0xFF, &i) || (i & FA_DIREC))
		{
			char wad[256];
			int namelen;

			do {
				fprintf (stderr, "\nCannot open %s. Please enter the full path to your IWAD:\n", MainWad);
				fprintf (stderr, "(Just press Enter to quit)\n> ");
				if (!fgets(wad, sizeof(wad), stdin))
					ProgError ("Error requesting main wad file: %s", strerror(errno));

				namelen = strlen(wad);

				while(namelen > 0 && (wad[namelen-1] == '\n' || wad[namelen-1] == '\r'))
					wad[--namelen] = 0;

				if (namelen == 0)
					ProgError ("No main wad specified");

				free(MainWad);
				MainWad = strdup(wad);
			} while(!(file_exists(MainWad, 0xFF, &i) && !(i & FA_DIREC)));

			fixcfg (usedCfg);
		}

		// [CR] Check if ACC exists
		char acc_cmd[256];

		strcpy(acc_cmd, ACCDirectory);
		append_filename(acc_cmd, acc_cmd, ACCCommand, sizeof(acc_cmd));

		if(file_exists(acc_cmd, 0xFF, &i) == 0 || (i & FA_DIREC))
		{
			int namelen;

			do {
				fprintf (stderr, "\nCannot open %s. Please enter the full path to your ACC command:\n", acc_cmd);
				fprintf (stderr, "(Just press Enter to ignore)\n> ");
				if(!fgets(acc_cmd, sizeof(acc_cmd), stdin))
					ProgError("Error requesting ACC executable: %s", strerror(errno));

				namelen = strlen(acc_cmd);

				while(namelen > 0 && (acc_cmd[namelen-1] == '\n' || acc_cmd[namelen-1] == '\r'))
					acc_cmd[--namelen] = 0;

				if(!namelen)
					break;
			} while(!(file_exists(acc_cmd, 0xFF, &i) && !(i & FA_DIREC)));

			// Split command from directory
			if(strlen(acc_cmd))
			{
				if((t = strrchr(acc_cmd, '\\')) != NULL ||
				   (t = strrchr(acc_cmd, '/')) != NULL)
				{
					free(ACCCommand);
					ACCCommand = strdup(t+1);

					*t = 0;

					free(ACCDirectory);
					ACCDirectory = strdup(acc_cmd);
				}
			}
		}
	}
	else
#endif
	{
		int l = strlen(ACCDirectory);
		i = 0;

		while(i < l)
		{
			if(ACCDirectory[i] == '/' || ACCDirectory[i] == '\\')
				ACCDirectory[i] = OTHER_PATH_SEPARATOR;

			++i;
		}

		if(l > 0 && ACCDirectory[l-1] == OTHER_PATH_SEPARATOR)
			ACCDirectory[l-1] = '\0';
	}

    /* load the wad files */
	OpenMainWad (MainWad);

	if((p=ConfigPatchWads))
	{
		while(*p)
		{
			OpenPatchWad(*p);
			p++;
			NumConfigPatchWads++;
		}
	}
    if((p=PatchWads))
	{
		while (*p)
		{
			OpenPatchWad(*p);
			p++;
			NumPatchWads++;
		}
	}
    /* sanity check */
    CloseUnusedWadFiles();
	RemapIcons ();

	add_decorate();

    /* SO: here seems like a good place to make the ld_types index */
    /* not so good now that the data structures have changed! */
    index_ld_types();
    index_things();
    index_spawns();
    index_schemes();
	index_keys();

    /* all systems go! */
	RealVideoMode = VideoMode;

	quitnow = 0;
	set_close_button_callback(close_callback);

#ifdef ALLEGRO_CONSOLE_OK
	if (!GraphFront)
		MainLoop2();
	else
#endif
		MainLoop();

	/* that's all, folks! */
	CloseWadFiles();
	LogMessage(": The end!\n\n\n");
	if (logfile != NULL)
		fclose(logfile);

	return 0;
}
END_OF_MAIN()


//
// Append a string to a null-terminated string list
//
static void AppendItemToList( char ***list, const char *item)
{
	int i = 0;
    if (*list != NULL)
	{
		/* count the number of elements in the list (last = null) */
		while ((*list)[i] != NULL)
			i++;
	}

	/* expand the list */
	*list = (char**)myrealloc(*list, (i + 2) * sizeof(char *));

    /* append the new element */
    (*list)[i] = strdup(item);
    (*list)[i + 1] = NULL;
}


//
// Handle command line options
//
static void ParseCommandLineOptions(int argc, const char *argv[])
{
	int optnum;

	while (argc > 0 && argv[0][0] != '-' && argv[0][0] != '+')
	{
		AppendItemToList((char***)&PatchWads, argv[0]);
		argv++;
		argc--;
	}

    while (argc > 0)
	{
		if (!strcmp( argv[ 0], "-?") || !stricmp( argv[ 0], "-h") ||
			!stricmp( argv[ 0], "-help"))
		{
			Usage(stdout);
			exit(0);
		}
		for (optnum = 0; options[ optnum].opt_type != OPT_END; optnum++)
		{
			if (!stricmp(argv[0]+1, options[ optnum].short_name) ||
				!stricmp(argv[0]+1, options[ optnum].long_name))
			{
				switch (options[ optnum].opt_type)
				{
				case OPT_BOOLEAN:
					if (argv[0][ 0] == '-')
						*((Bool *) (options[ optnum].data_ptr)) = TRUE;
					else 
						*((Bool *) (options[ optnum].data_ptr)) = FALSE;
					break;
				case OPT_INTEGER:
					if(argc > 1)
					{
						argv++;
						argc--;
						*((BCINT *) (options[ optnum].data_ptr)) = atoi( argv[ 0]);
					}
					break;
				case OPT_STRING:
					if (argc > 1)
					{
						argv++;
						argc--;
						*((char **) (options[ optnum].data_ptr)) = strdup(argv[ 0]);
					}
					break;
				case OPT_STRINGACC:
					if (argc > 1)
					{
						argv++;
						argc--;
						AppendItemToList( (char ***) options[ optnum].data_ptr, argv[ 0]);
					}
					break;
				case OPT_STRINGLIST:
					while (argc > 1 && argv[ 1][ 0] != '-' && argv[ 1][ 0] != '+')
					{
						argv++;
						argc--;
						AppendItemToList( (char ***) options[ optnum].data_ptr, argv[ 0]);
					}
					break;
				case OPT_END:
					allegro_message("Unknown argument: \"%s\"", argv[0]);
					break;
//				default:
//					ProgError( "unknown option type (BUG!)");
				}
				break;
			}
		}
		argv++;
		argc--;
    }
}



/* read the config file */
static void ParseConfigFileOptions(const char *filename)
{
    FILE *cfgfile;
    char line[1024];
    char *value;
    char *option;
    char *p;
	int linenum;
    BCINT   optnum;

    if ((cfgfile = fopen (find_file(filename), "r")) == NULL)
	{
		if ((cfgfile = fopen (find_file(DEU_CONFIG_FILE), "r")) == NULL)
			ProgError("Configuration file not found (%s)\n", filename);
		filename = DEU_CONFIG_FILE;
	}

	linenum = 0;
    while (fgets (line, 1024, cfgfile) != NULL)
	{
		linenum++;

		// skip comments and lines less than 2 char long
		if (line[0] == '#' || strlen( line) < 2)
			continue;

		// remove end-of-line terminators
		while(line[strlen(line) - 1] == '\n' || line[strlen(line) - 1] == '\r')
			line[strlen(line) - 1] = '\0';

		// skip blanks before the option name
		option = line;
		while (isspace(option[0]))
			option++;

		// skip the option name
		value = option;
		while (value[0] && value[0] != '=' && !isspace(value[0]))
			value++;
		if (!value[0])
			allegro_message("Invalid line in %s (ends prematurely)", filename);
		// mark the end of the option name
		value[0] = '\0';
		value++;

		/* skip blanks after the option name */
		while (isspace( value[ 0]))
			value++;
		if (value[ 0] != '=')
			allegro_message("Invalid line in %s (no '=')", filename);
		value++;

		/* skip blanks after the equal sign */
		while (isspace( value[ 0]))
			value++;
		for (optnum = 0; options[ optnum].opt_type != OPT_END; optnum++)
		{
			if (!stricmp(option, options[optnum].long_name))
			{
				switch (options[ optnum].opt_type)
				{
				case OPT_BOOLEAN:
					if(!stricmp(value, "yes") || !stricmp(value, "true") ||
					   !stricmp(value, "on") || atoi(value))
						*((Bool *) (options[ optnum].data_ptr)) = TRUE;
					else
						*((Bool *) (options[ optnum].data_ptr)) = FALSE;

					break;
				case OPT_INTEGER:
					*((BCINT *)options[ optnum].data_ptr) = atoi(value);
					break;
				case OPT_STRING:
					p = (char*)myalloc( (strlen(value) + 1) * sizeof(char));
					strcpy(p, value);
					*((char **)options[optnum].data_ptr) = p;
					break;
				case OPT_STRINGACC:
					AppendItemToList((char ***)options[ optnum].data_ptr, value);
					break;
				case OPT_STRINGLIST:
					while(value[0])
					{
						option = value;
						while (option[0] && !isspace( option[ 0]))
							option++;
						option[0] = '\0';
						option++;
						while (isspace(option[0]))
							option++;

						AppendItemToList( (char ***) options[ optnum].data_ptr, value);
						value = option;
					}
					break;
				default:
					allegro_message("Unknown option type on line %d", linenum);
				}
				break;
			}
		}
		if (options[ optnum].opt_type == OPT_END)
		{
/*			printf ("Your %s is either old or trashed. Try editing it to remove\n", filename);
			printf ("the line with the \"%s\" option. If that doesn't work, delete\n", option);
			printf ("%s and let ZETH rebuild it.\n", filename);
			ProgError( "Invalid option on line %d in %s: \"%s\"", linenum, filename, option);*/
			allegro_message("Your %s is either old or trashed. Try editing it to remove\n"
							"the line with the \"%s\" option. If that doesn't work, delete\n"
							"%s and let ZETH rebuild it.\n"
							"Invalid option on line %d: \"%s\"\n", filename, option, filename, linenum, option);
		}
    }
    fclose( cfgfile);
    if (StretchDelay < 1)
    	StretchDelay = 1;
    if (StretchDelay > 100)
    	StretchDelay = 100;
    if (CloseToLine < 0)
    	CloseToLine = 0;
    if (CloseToLine > 16)
    	CloseToLine = 16;
    GridScale = InitGrid;
}


//
// output the program usage to the specified file
//
void Usage(FILE *where)
{
#ifndef ALLEGRO_CONSOLE_OK
	if(where == stdout || where == stderr)
	{
		allegro_message("Usage:\n"
			"ZETH [-w <main_wad_file>] [-d] [-sb] [-c] [-q] [-qq] [-sq] [-e] [-a] [-i]\n"
			"     [-z <zoom>] [-v <mode>] [-config <ini_file>] [-pw <pwad_file>]\n"
			"     [-file <pwad_files>...]\n"
			"   -w    Gives the name of the main wad file (also -main).  Default is DOOM.WAD\n"
			"   -d    Enter debug mode (also -debug).\n"
			"   -q    Suppresses some sounds (also -quiet).\n"
			"   -qq   Suppresses all sounds (also -quieter).\n"
			"   -e    Stops prompts for confirmation (also -expert).\n"
			"   -a    To have an additive selection box (also -addselbox).\n"
			"   -i    Show the info bar in the editors (also -infobar).\n"
			"   -z    Set the initial zoom factor for the editors (also -zoom).\n"
			"   -v    Set the default video mode number (also -video).\n"
			"   -sq   To draw ruler with squares instead of circles (also -squarecircles).\n"
			"   -sb   Swaps the mouse buttons (also -swapbuttons).\n"
			"   -pw   To add one patch wad file to be loaded; may be repeated (also -pwad).\n"
			"   -file To add a list of patch wad files to be loaded.\n"
			"   -config Gives the name of the config file.\n"
			"Put a '+' instead of a '-' before boolean options to reverse their effect.\n"
			"** THIS LIST IS HORRIBLY OUTDATED **\n");
		return;
	}
#endif

	fputs  ("Usage:\n"
			"ZETH [-w <main_wad_file>] [-d] [-sb] [-c] [-q] [-qq] [-sq] [-e] [-a] [-i]\n"
			"     [-z <zoom>] [-v <mode>] [-config <ini_file>] [-pw <pwad_file>]\n"
			"     [-file <pwad_files>...]\n"
			"   -w    Gives the name of the main wad file (also -main).  Default is DOOM.WAD\n"
			"   -d    Enter debug mode (also -debug).\n"
			"   -q    Suppresses some sounds (also -quiet).\n"
			"   -qq   Suppresses all sounds (also -quieter).\n"
			"   -e    Stops prompts for confirmation (also -expert).\n"
			"   -a    To have an additive selection box (also -addselbox).\n"
			"   -i    Show the info bar in the editors (also -infobar).\n"
			"   -z    Set the initial zoom factor for the editors (also -zoom).\n"
			"   -v    Set the default video mode number (also -video).\n"
			"   -sq   To draw ruler with squares instead of circles (also -squarecircles).\n"
			"   -sb   Swaps the mouse buttons (also -swapbuttons).\n"
			"   -pw   To add one patch wad file to be loaded; may be repeated (also -pwad).\n"
			"   -file To add a list of patch wad files to be loaded.\n"
			"   -config Gives the name of the config file.\n"
			"Put a '+' instead of a '-' before boolean options to reverse their effect.\n"
			"** THIS LIST IS HORRIBLY OUTDATED **\n", where);
}


//
// output the credits of the program to the specified file
//
void Credits( FILE *where)
{
    fprintf( where, "\nZETH - ZDoom Editor for Total Headcases, Version %s\n\n", DETH_VERSION);
    fputs  ("The Professional DEU-GCC Editor of Choice for Discriminating Authors.\n\n"
			"Authors:-\n"
			"    Antony J. Burden (100765.3240@compuserve.com)\n"
			"    and    Simon Oke (loony@demon.co.uk)\n\n"
			"Hacks by:-\n"
			"    Jim F. Flynn  (jflynn@pacbell.net)\n"
			"With Help From:-\n"
			"    John W. Anderson, Jim F. Flynn.\n"
			"    Sverre A. Kvernmo, Ty Halderman.\n"
			"    Ken F. Henderson and Kjell Rilbe.\n\n"
			"ZETH additions by Randy Heit.\n\n", where);
//	fprintf( where, "Visit Dr Sleep's Doom Apothecary and Quake Armamentarium:\n");
//	fprintf( where, "The offical Home of DETH and HETH - http://www.pennet.com/drsleep\n\n");
	fprintf( where, "Editing: %s\n\n", Game);
}


//
// play a fascinating tune
//

void DEU_Beep()
{
#ifdef DJGPP
	if (!Quieter)
	{
		sound(640);
		rest(100);
		nosound();
	}
#endif
}



//
// play a sound
//

void PlaySound( BCINT freq, BCINT msec)
{
#ifdef DJGPP
	if (!Quiet)
	{
		sound(freq);
		rest(msec);
		nosound();
	}
#else
	(void)freq;
	(void)msec;
#endif
}


//
// terminate the program reporting an error
//
void ProgError( char *errstr, ...)
{
	char text[256] = "";
    va_list args;
    
    DEU_Beep();
    DEU_Beep();

	TermGfx();

    va_start(args, errstr);
    vsprintf(text, errstr, args);
    va_end(args);

    if (Debug && logfile != NULL)
	{
		fprintf(logfile, "\nProgram Error: *** ");
		fprintf(logfile, text);
		fprintf(logfile, " ***\n");
    }

	allegro_message("*** Program Error ***\n%s", text);

    // clean up things
    ForgetLevelData();
    ForgetTextureNames();
    CloseWadFiles();

    exit(5); 
}

//
// write a message in the log file
//
// SO: changed this 13/3/95.
// we now open the logfile before, and close it after, writing each
// line. This way we guarantee that all messages get into the file
// (even if the program crashes).
//
void LogMessage( char *logstr, ...)
{
    va_list  args;
    time_t   tval;
    char    *tstr;

    if (Debug)
	{
		char fname[256];
		append_filename(fname, exe_path, DEU_LOG_FILE, sizeof(fname));

		logfile = fopen(fname, "a");
		va_start( args, logstr);
		/* if the messsage begins with ":",
		   output the current date & time first */
		if (logstr[ 0] == ':')
		{
		    time( &tval);
		    tstr = ctime( &tval);
		    tstr[ strlen( tstr) - 1] = '\0';
		    fprintf(logfile, "%s", tstr);
		}
		vfprintf( logfile, logstr, args);
		va_end( args);
		fclose(logfile);
    }
}

int lool()
{
	int dirx = 1, diry = 1;
	BCINT width, height, x, y;
	GetPicSize("M_DOOM", &width, &height);
	x = SCREEN_W/2 + width/2; y = SCREEN_H/2 + height/2;
	
	while(!keypressed())
	{
		DisplayPic( x, y, x+width, y+height, "M_DOOM");
		x+=dirx; y+=diry; 
		
		if(x > SCREEN_W - width) dirx = -dirx; 
		if(x < 0 ) dirx = -dirx;
		if(y > SCREEN_H - height) diry = -diry;
		if(y < 0 ) diry = -diry;

		vsync();
	}

	return 0;
}

//
// Graphic Front End
//
static void DrawFrontEnd()
{
	struct al_ffblk info;
	BCINT height, width;
    int wadcount, x;
	BITMAP *buffer;
    WadPtr wad;
    int count;
	int done;
	int c;

	video->buffer_drawing(true);
	buffer = video->get_buffer();

	// Draw the credits and logo
	video->draw_box_3d(0, 0, ScrMaxX-1, 149);

	GetPicSize("M_HTIC", &width, &height);
	DisplayPic( ScrMaxX - width - 10, 50, ScrMaxX - 10, 50 + height, "M_HTIC");
	GetPicSize("M_DOOM", &width, &height);
	DisplayPic( ScrMaxX - width - 10, 50, ScrMaxX - 10, 50 + height, "M_DOOM");

	video->draw_text(10, 10, TranslateToGameColor(BLACK), -1, false, false, "ZETH - ZDoom Editor for Total Headcases, Version %s", DETH_VERSION);

	c = TranslateToGameColor(DARKBLUE);
	video->draw_text(20, 24, c, TranslateToGameColor(LIGHTGRAY), false, false, "The Professional DEU-GCC Editor of Choice for Discriminating Authors.");

	video->draw_text(10, 44, c, -1, false, false, "Authors:-");
	video->draw_text(-1, -1, c, -1, false, false, "    Antony J. Burden (100765.3240@compuserve.com)");
	video->draw_text(-1, -1, c, -1, false, false, "    and    Simon Oke (loony@demon.co.uk)");
	video->draw_text(10, 84, c, -1, false,  true, "With Help From:-");
	video->draw_text(-1, -1, c, -1, false,  true, "    John W. Anderson, Jim F. Flynn.");
	video->draw_text(-1, -1, c, -1, false,  true, "    Sverre A. Kvernmo, Ty Halderman.");
	video->draw_text(-1, -1, c, -1, false,  true, "    Ken F. Henderson and Kjell Rilbe.");
	video->draw_text(-1, -1, c, -1, false,  true, "    Chris Robinson and Michael \"guru\" Zoller.");

	// Draw the main front end and buttons
	video->draw_box_3d(0, 150, ScrMaxX-1, ScrMaxY-51);

	c = TranslateToGameColor(GREEN);
	video->draw_text(10, 160, c, -1, false, false, "Current Settings");
	video->draw_text(10, 160, c, -1, false, false, "________________");
//	video->draw_text(10, 160, c, -1, false, true, "Current Settings");
//	video->draw_line(10, 171, 96 , 171, c);

	c = TranslateToGameColor(DARKBLUE);
	video->draw_text(10, 172, c, -1, false, true, " ");
	video->draw_text(-1,  -1, c, -1, false, true, "Video Mode");
	video->draw_text(-1,  -1, c, -1, false, true, "Editor Mode");
	video->draw_text(-1,  -1, c, -1, false, true, "Sounds");
	video->draw_text(-1,  -1, c, -1, false, true, "Info Bar");
	video->draw_text(-1,  -1, c, -1, false, true, "Nested Menus");
	video->draw_text(-1,  -1, c, -1, false, true, "Scroll Factor");
	video->draw_text(-1,  -1, c, -1, false, true, "Grid Settings");
	video->draw_text(-1,  -1, c, -1, false, true, "Initial Zoom");
	video->draw_text(-1,  -1, c, -1, false, true, "Close to Split");
	video->draw_text(-1,  -1, c, -1, false, true, "Marquee Delay");
	video->draw_text(-1,  -1, c, -1, false, true, "Debugging");

	video->draw_text(100, 172, c, -1, false, true, " ");
	video->draw_text( -1,  -1, c, -1, false, true,
	                 ((VideoMode <= 2) ? ": 640x480" :
	                 ((VideoMode == 3) ? ": 800x600" :
	                 ((VideoMode == 4) ? ": 1024x768" :
	                 ((VideoMode == 5) ? ": 1280x960" :
	                 ((VideoMode == 6) ? ": 1600x1200" :
	                                     ": Unknown"))))));

	video->draw_text(-1, -1, c, -1, false, true,
	                 (Expert ? ": Expert": ": Novice"));
	video->draw_text(-1, -1, c, -1, false, true,
	                 (Quieter ? ": Off" :
	                 (Quiet   ? ": Partial" : ": On")));
	video->draw_text(-1, -1, c, -1, false, true,
	                 (!InfoShown ? ": Off" :
					 (AltInfo ? ": On, Alternative" :
	                            ": On, Standard")));
	video->draw_text(-1, -1, c, -1, false, true, ": %s",
	                 (NestMenus ? "On": "Off"));
	video->draw_text(-1, -1, c, -1, false, true, ": %d", DefaultLargeScroll);
	video->draw_text(-1, -1, c, -1, false, true, ": %d, %d, %d%s", MinGrid, MaxGrid, InitGrid, GridDashed ? ", Dashed" : "");
	video->draw_text(-1, -1, c, -1, false, true, ": %d", InitialScale);
	video->draw_text(-1, -1, c, -1, false, true, ": %d", CloseToLine);
	video->draw_text(-1, -1, c, -1, false, true, ": %d", StretchDelay);
	video->draw_text(-1, -1, c, -1, false, true,
	                 (Debug ? ": On": ": Off"));

	video->draw_text(10, -1, c, -1, false, true, "");
	if (ThingAngle)
		video->draw_text(-1, -1, c, -1, false, true, "Showing Things with Angle");
	if (VertConf)
		video->draw_text(-1, -1, c, -1, false, true, "Confirming Merging of Vertices");
	if (DispSprite)
		video->draw_text(-1, -1, c, -1, false, true, "Showing Textures/Flats and Sprites");
	if (AdditiveSelBox)
		video->draw_text(-1, -1, c, -1, false, true, "Additive Selection Box");
	if (Caching)
		video->draw_text(-1, -1, c, -1, false, true, "Caching Textures");
	if (KeepInfo)
		video->draw_text(-1, -1, c, -1, false, true, "Keeping INFO Lumps");

	c = TranslateToGameColor(GREEN);
	video->draw_text(10, 360, c, -1, false, true, "Path: %s", ((strlen(*Pwads) <= 1) ? Cwd : *Pwads));

	count = 0;
	for (wad = WadFileList->next; wad && count <= 5; wad = wad->next)
	{
		if (count == 0)
		{
			video->draw_text(10, 370, c, -1, false, true, "Added:");
			video->draw_text(50, 370, c, -1, false, true, wad->filename);
		}
		else if (count == 5)
			video->draw_text(10, -1, c, -1, false, true, "More..");
		else
			video->draw_text(-1, -1, c, -1, false, true, wad->filename);
		count++;
	}

	if (UsePromptOnly)
	{
		wadcount = 0;
		x = 300;

		video->draw_text(x, 160, TranslateToGameColor(DARKRED), -1, false, true, "Pwads Present in %s", strlen(*Pwads) <= 1 ? Cwd : *Pwads);
		c = TranslateToGameColor(DARKBLUE);

		done = al_findfirst("*.wad", &info, 0);
		while(!done && wadcount <= 50)
		{
			video->draw_text(x, 180 + (wadcount * 10), c, -1, false, true, info.name);
			video->draw_text(x + 80, 180 + (wadcount * 10), c, -1, false, true, "%dK", (info.size / 1024));
			done = al_findnext(&info);
			wadcount++;
			if (wadcount == 24)
			{
				x += 130;
				wadcount = 0;
			}
		}
	}
	else
	{
		video->draw_text(ScrMaxX-240, 160, c, -1, false, false, "Quick Keys");
		video->draw_text(ScrMaxX-240, 160, c, -1, false, false, "__________");

		c = TranslateToGameColor(DARKBLUE);
		video->draw_text(ScrMaxX-240, 172, c, -1, false, true, "");
		video->draw_text(-1, -1, c, -1, false, true, "Enter");
		video->draw_text(-1, -1, c, -1, false, true, "E");
		video->draw_text(-1, -1, c, -1, false, true, "Shift E");
		video->draw_text(-1, -1, c, -1, false, true, "C");
		video->draw_text(-1, -1, c, -1, false, true, "Shift C");
		video->draw_text(-1, -1, c, -1, false, true, "W");
		video->draw_text(-1, -1, c, -1, false, true, "V");
		video->draw_text(-1, -1, c, -1, false, true, "Q");
		video->draw_text(ScrMaxX-150, 172, c, -1, false, true, "");
		video->draw_text(-1, -1, c, -1, false, true, "Command Line");
		video->draw_text(-1, -1, c, -1, false, true, "Edit Level 1");
		video->draw_text(-1, -1, c, -1, false, true, "Browse Levels to Edit");
		video->draw_text(-1, -1, c, -1, false, true, "Create Level 1");
		video->draw_text(-1, -1, c, -1, false, true, "Browse Levels to Create");
		video->draw_text(-1, -1, c, -1, false, true, "List Open Wads");
		video->draw_text(-1, -1, c, -1, false, true, "View Sprites");
		video->draw_text(-1, -1, c, -1, false, true, "Quit");

		video->draw_button(ScrMaxX-(640-586), ScrMaxY-80, 40, 16, "Help", true);
		video->draw_button(ScrMaxX-(640-525), ScrMaxY-80, 40, 16, "Quit", true);
		video->draw_button(ScrMaxX-(640-460), ScrMaxY-80, 40, 16, "Wads", true);
		video->draw_button(ScrMaxX-(640-400), ScrMaxY-80, 40, 16, "Dir", true);
#ifdef ALLEGRO_CONSOLE_OK
		video->draw_button(ScrMaxX-(640-340), ScrMaxY-80, 40, 16, "Shell", true);
#endif
		video->draw_button(ScrMaxX-(640-460), ScrMaxY-(480-342), 80, 16, "Edit", true);
		video->draw_button(ScrMaxX-(640-546), ScrMaxY-(480-342), 80, 16, "Browse", true);
		video->draw_button(ScrMaxX-(640-460), ScrMaxY-(480-370), 80, 16, "Create", true);
		video->draw_button(ScrMaxX-(640-546), ScrMaxY-(480-370), 80, 16, "Browse", true);
	}

	// Draw the command line area
	video->draw_box_3d(0, ScrMaxY-50, ScrMaxX-1, ScrMaxY-1);

	c = TranslateToGameColor(WHITE);
	video->draw_text(20, ScrMaxY - 26, c, -1, false, false, "Command Line:");
	video->draw_text(ScrMaxX - 170, ScrMaxY - 26, c, -1, false, false, "Enter \'?\' for Help");
	video->draw_button(130, ScrMaxY - 30, ScrMaxX - 315, 13, "", true);

	video->buffer_drawing(false);
	video->update(0, 0, ScrMaxX, ScrMaxY);
}

//
// the main program menu loop, in graphics
//
void MainLoop()
{
	char input[120];
	int k = 0;
	int buttons;

	InitGfx();
	DrawFrontEnd();
//	lool();

	for (;;)
	{
		input[0] = '\0';
		GetMouseCoords(&PointerX, &PointerY, &buttons);

		k = (keypressed() ? readkey() : 0);

		if (UsePromptOnly)
			k = ENTER;

		if (!UsePromptOnly && buttons == 1)
		{
			if (PointerY > ScrMaxY-48)
				k = ENTER;
			else if (MouseInArea(ScrMaxX-(640-586), ScrMaxY-80, 40, 16))
			{
				video->draw_button_in(ScrMaxX-(640-586), ScrMaxY-80, 40, 16, "Help");
				if (MouseClickedArea(ScrMaxX-(640-586), ScrMaxY-80, 40, 16))
				{
					ParseInput( "?");
					DrawFrontEnd();
				}
				video->draw_button(ScrMaxX-(640-586), ScrMaxY-80, 40, 16, "Help", true);
			}
			else if (MouseInArea(ScrMaxX-(640-525), ScrMaxY-80, 40, 16))
			{
				video->draw_button_in(ScrMaxX-(640-525), ScrMaxY-80, 40, 16, "Quit");
				if (MouseClickedArea(ScrMaxX-(640-525), ScrMaxY-80, 40, 16))
					k = 'q';
				video->draw_button(ScrMaxX-(640-525), ScrMaxY-80, 40, 16, "Quit", true);
			}
#ifdef ALLEGRO_CONSOLE_OK
			else if (MouseInArea(ScrMaxX-(640-340), ScrMaxY-80, 40, 16))
			{
				video->draw_button_in(ScrMaxX-(640-340), ScrMaxY-80, 40, 16, "Shell");
				if (MouseClickedArea(ScrMaxX-(640-340), ScrMaxY-80, 40, 16))
					k = '!';
				video->draw_button(ScrMaxX-(640-340), ScrMaxY-80, 40, 16, "Shell", true);
			}
#endif
			else if (MouseInArea(ScrMaxX-(640-460), ScrMaxY-80, 40, 16))
			{
				video->draw_button_in(ScrMaxX-(640-460), ScrMaxY-80, 40, 16, "Wads");
				if (MouseClickedArea(ScrMaxX-(640-460), ScrMaxY-80, 40, 16))
					k = 'w';
				video->draw_button(ScrMaxX-(640-460), ScrMaxY-80, 40, 16, "Wads", true);
			}
			else if (MouseInArea(ScrMaxX-(640-400), ScrMaxY-80, 40, 16))
			{
				video->draw_button_in(ScrMaxX-(640-400), ScrMaxY-80, 40, 16, "Dir");
				if (MouseClickedArea(ScrMaxX-(640-400), ScrMaxY-80, 40, 16))
				{
					ParseInput("DIR");
					DrawFrontEnd();
				}
				video->draw_button(ScrMaxX-(640-400), ScrMaxY-80, 40, 16, "Dir", true);
			}
			else if (MouseInArea(ScrMaxX-(640-460), ScrMaxY-(480-350), 80, 16))
			{
				video->draw_button_in(ScrMaxX-(640-460), ScrMaxY-(480-342), 80, 16, "Edit");
				if (MouseClickedArea(ScrMaxX-(640-460), ScrMaxY-(480-350), 80, 16))
					k = 'e';
				video->draw_button(ScrMaxX-(640-460), ScrMaxY-(480-342), 80, 16, "Edit", true);
			}
			else if (MouseInArea(ScrMaxX-(640-546), ScrMaxY-(480-350), 80, 16))
			{
				video->draw_button_in(ScrMaxX-(640-546), ScrMaxY-(480-342), 80, 16, "Browse");
				if (MouseClickedArea(ScrMaxX-(640-546), ScrMaxY-(480-350), 80, 16))
					k = 'E';
				video->draw_button(ScrMaxX-(640-546), ScrMaxY-(480-342), 80, 16, "Browse", true);
			}
			else if (MouseInArea(ScrMaxX-(640-460), ScrMaxY-(480-370), 80, 16))
			{
				video->draw_button_in(ScrMaxX-(640-460), ScrMaxY-(480-370), 80, 16, "Create");
				if (MouseClickedArea(ScrMaxX-(640-460), ScrMaxY-(480-370), 80, 16))
					k = 'c';
				video->draw_button(ScrMaxX-(640-460), ScrMaxY-(480-370), 80, 16, "Create", true);
			}
			else if (MouseInArea(ScrMaxX-(640-546), ScrMaxY-(480-370), 80, 16))
			{
				video->draw_button_in(ScrMaxX-(640-546), ScrMaxY-(480-370), 80, 16, "Browse");
				if (MouseClickedArea(ScrMaxX-(640-546), ScrMaxY-(480-370), 80, 16))
					k = 'C';
				video->draw_button(ScrMaxX-(640-546), ScrMaxY-(480-370), 80, 16, "Browse", true);
			}
		}

		if(quitnow)
			return;

		switch (k & 255)
		{
		case 'q':
			SaveConfigFileOptions();
			TermGfx();
			return;

		case 'E':
			ParseInput("E");
			break;

		case 'e':
			if (stricmp(Game, "DOOM2") == 0 || stricmp(Game, "HEXEN") == 0)
				ParseInput("E 1");
			else
				ParseInput("E 1 1");
			break;

		case 'C':
			ParseInput("C");
			break;

		case 'c':
			if (stricmp(Game, "DOOM2") == 0 || stricmp(Game, "HEXEN") == 0)
				ParseInput("C 1");
			else
				ParseInput("C 1 1");
			break;

		case '!':
			ParseInput("!");
			break;

		case 'v':
			HideMousePointer();
			ParseInput("V");
			ShowMousePointer();
			break;

		case 'w':
			ParseInput("W");
			DrawFrontEnd();
			break;

		case '\r':
		case '\n':
			// fullscreen/windowed
			if((key_shifts & KB_ALT_FLAG) && !UsePromptOnly)
			{
				Fullscreen = !Fullscreen;
				InitGfx();
				DrawFrontEnd();
				break;
			}

			video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));
			input[0] = '\0';
			InputCommand( 130, ScrMaxY - 30, "", MIN(((unsigned)(ScrMaxY-34-130)/8)+1, sizeof(input)), input);
			if (toupper(input[0]) == 'Q')
			{
				SaveConfigFileOptions();
			    TermGfx();
				return;
			}
			else if (toupper(input[0]) == 'O')
			{
				DoingBehavior = TRUE;
				ParseInput("I BEHAVIOR");
				DoingBehavior = FALSE;
				DrawFrontEnd();
			}
			else
				ParseInput(input);

			input[0] = '\0';
			DrawFrontEnd();
			break;
		}

		rest(1);
	}
}

//
// Front end command line parser for graphic front end
// [RH] Make input1 const, because some of the calls to ParseInput use string literals.
//
void ParseInput(const char *input1)
{
	static char input[192] = "";
    char *com, *out;
    FILE *file, *raw;
    WadPtr wad = NULL;
    int k = 0;
	int buttons = 0;
    Bool startcheck = FALSE;
    int x, c, wadcount;
	struct al_ffblk info;
	int done;

	/* [RH] Duplicate input1 as the original input for strtok */
	strncpy(input, input1, 192);

	/* eat the white space and get the first command word */
	com = strtok(input, " ");

	/* user just hit return */
	if (com == NULL)
	{
		DEU_Beep();
		return;
	}

	/* user inputting for help */
	if (!strcmp( com, "?"))
	{
		video->draw_box_3d(0, 150, ScrMaxX-1, ScrMaxY-51);
		video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));

		c = TranslateToGameColor(BLUE);
		video->draw_text(10, 160, c, -1, false, false, "Command Line Help");
		video->draw_text(10, 162, c, -1, false, false, "_________________");
		video->draw_text(-1,  -1, c, -1, false, false, " ");

		c = TranslateToGameColor(DARKBLUE);
		video->draw_text(-1, -1, c, -1, false, false, "B[uild] <WadFile>              -  Build new IWAD file");
		video->draw_text(-1, -1, c, -1, false, false, "C[reate] [level]               -  Create and edit a new level");
		video->draw_text(-1, -1, c, -1, false, false, "D[ump] <DirEntry> [outfile]    -  Dump a directory entry in hex");
		video->draw_text(-1, -1, c, -1, false, false, "E[dit] [level]                 -  Edit a level");
		video->draw_text(-1, -1, c, -1, false, false, "G[roup] <WadFile>              -  Group all PWADs into a file");
		video->draw_text(-1, -1, c, -1, false, false, "I[nsert] <RawFile> <WadFile>   -  Insert a raw file into a PWAD file");
		video->draw_text(-1, -1, c, -1, false, false, "L[ist] <WadFile> [outfile]     -  List the directory of a PWAD file");
		video->draw_text(-1, -1, c, -1, false, false, "M[aster] [outfile]             -  List the master directory");
		video->draw_text(-1, -1, c, -1, false, false, "R[ead] <WadFile>               -  Read a new PWAD file");
		video->draw_text(-1, -1, c, -1, false, false, "S[ave] <DirEntry> <WadFile>    -  Save one entry to a separate file");
		video->draw_text(-1, -1, c, -1, false, false, "V[iew] [SpriteName]            -  View the sprites");
		video->draw_text(-1, -1, c, -1, false, false, "W[ads]                         -  List the open PWADS");
		video->draw_text(-1, -1, c, -1, false, false, "X[tract] <DirEntry> <RawFile>  -  Extract one entry to a raw file");
		video->draw_text(-1, -1, c, -1, false, false, "O [myscript.raw]               -  Convert compiled script into BEHAVIOR.WAD");
		video->draw_text(-1, -1, c, -1, false, false, "DIR / LS                       -  List Pwads present in current path");
#ifdef ALLEGRO_CONSOLE_OK
		video->draw_text(-1, -1, c, -1, false, false, "!                              -  Shell, Type 'exit' to return");
#endif

		video->draw_text(10, ScrMaxY-70, TranslateToGameColor(DARKRED), -1, false, true, "Click Mouse Button or Hit Any Key...");
		for (;;)
		{
			GetMouseCoords(&PointerX, &PointerY, &buttons);
			if (startcheck)
				if (buttons == 1)
					break;
			if (keypressed())
				k = readkey();
			if ( k > 0)
			    break;
			k = 0;
			if (buttons == 0)
			    startcheck = TRUE;

			rest(1);
		}
	}
	
#ifdef ALLEGRO_CONSOLE_OK
	else if (!strcmp( com, "!"))
	{
		TermGfx();
		printf("Type Exit to return\n");
		system("");
		InitGfx();
		DrawFrontEnd();
	}
#endif

	/* user asked for list of open WAD files */
	else if (!stricmp( com, "WADS") || !stricmp( com, "W"))
	{
		video->draw_box_3d(0, 150, ScrMaxX-1, ScrMaxY-51);
		video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));

    	c = TranslateToGameColor(GFXBLUE);
		video->draw_text(10, 160, c, -1, false, true, "%-30s  IWAD  (Main wad file)", WadFileList->filename);
		for (wad = WadFileList->next; wad; wad = wad->next)
		{
			if(wad->directory[0].name[0] == 'E' &&
			   wad->directory[0].name[2] == 'M')
			{
				video->draw_text(10, -1, c, -1, false, true, "%-30s  PWAD  (Patch wad file for episode %c level %c)",
					wad->filename, wad->directory[0].name[1], wad->directory[0].name[3]);
			}
			else
			{
				// [RH] Do not copy to input, because we don't know that it's large enough
				char lumpname[9];
				strncpy(lumpname, wad->directory[0].name, 8);
				lumpname[8] = '\0';
				video->draw_text(10, -1, c, -1, false, true, "%-30s  PWAD  (Patch wad file for %s)",
					wad->filename, lumpname);
			}
		}
		SetColor( DARKRED);
		video->draw_text(10, ScrMaxY-70, c, -1, false, true, "Click Mouse Button or Hit Any Key...");
		for (;;)
		{
			rest(1);

			GetMouseCoords( &PointerX, &PointerY, &buttons);
			if (startcheck)
				if (buttons == 1)
					break;
			if (keypressed())
				k = readkey();
			if ( k > 0)
			    break;
			k = 0;
			if (buttons == 0)
			    startcheck = TRUE;
		}
	}
	
	/* user asked to edit a level */
	else if (!stricmp( com, "EDIT")   || !stricmp( com, "E") ||
			 !stricmp( com, "CREATE") || !stricmp( com, "C"))
	{
		SList tokens, t;

		Bool n = (stricmp(com, "EDIT") && stricmp(com, "E"));
		/* if com is neither of these, the user wants
		   to start from scratch */
		if (n) 
			CreatingNewLevel = TRUE;

		tokens = NULL;
		com = strtok( NULL, " ");
		while(com)
		{
			tokens = SList_append(tokens, com);
			com = strtok(NULL, " ");
		}
		if(tokens)
		{
			MakeLevelName(tokens);
			
			/* now forget the token list */
			while(tokens)
			{
				t = tokens->next;
				free(tokens->string);
				free(tokens);
				tokens = t;
			}
		}

		EditLevel(n);

		DrawFrontEnd();
	}

	/* user asked to build a new main WAD file */
	else if (!stricmp( com, "BUILD") || !stricmp( com, "B"))
	{
		com = strtok( NULL, " ");
		if (com == NULL)
		{
#define DRAW_ERROR(s)	\
			{ video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));	\
			video->draw_text(130, ScrMaxY - 42, TranslateToGameColor(DARKRED), -1, false, false, s); }
			DRAW_ERROR("Wad file name argument missing.");
			return;
		}
		for (wad = WadFileList; wad; wad = wad->next)
			if (!stricmp( com, wad->filename))
				break;
		if (wad)
		{
			char buf[128];
			sprintf(buf, "File \"%s\" is opened and cannot be overwritten.\n", com);
			DRAW_ERROR(buf);
			return;
		}
		video->draw_box_3d(0, 150, ScrMaxX-1, ScrMaxY-51);
		video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));

		BuildNewMainWad( com, FALSE);

		video->draw_text(10, 410, TranslateToGameColor(DARKRED), -1, false, true, "Click Mouse Button or Hit Any Key...");
		for (;;)
		{
			GetMouseCoords( &PointerX, &PointerY, &buttons);
			if (startcheck)
				if (buttons == 1)
					break;
			if (keypressed())
				k = readkey();
			if ( k > 0)
			    break;
			k = 0;
			if (buttons == 0)
			    startcheck = TRUE;

			rest(1);
		}
	}
	
	/* user asked to build a compound patch WAD file */
	else if (!stricmp( com, "GROUP") || !stricmp( com, "G"))
	{
		if (WadFileList->next == NULL || WadFileList->next->next == NULL)
		{
			DRAW_ERROR("You need at least two open PWAD files.");
			return;
		}
		com = strtok( NULL, " ");
		if (com == NULL)
		{
			DRAW_ERROR("Wad file name argument missing.");
			return;
		}
		for (wad = WadFileList; wad; wad = wad->next)
			if (!stricmp( com, wad->filename))
				break;
		if (wad)
		{
			char buf[128];
			sprintf(buf, "File \"%s\" is opened and cannot be overwritten.", com);
			DRAW_ERROR(buf);
			return;
		}
		video->draw_box_3d(0, 150, ScrMaxX-1, ScrMaxY-51);
		video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));

		BuildNewMainWad(com, TRUE);
		video->draw_text(10, 410, TranslateToGameColor(DARKRED), -1, false, true, "Click Mouse Button or Hit Any Key...");
		for (;;)
		{
			GetMouseCoords( &PointerX, &PointerY, &buttons);
			if (startcheck)
				if (buttons == 1)
					break;
			if (keypressed())
				k = readkey();
			if ( k > 0)
			    break;
			k = 0;
			if (buttons == 0)
			    startcheck = TRUE;

			rest(1);
		}
	}

	/* user ask for a listing of a WAD file */
	else if (!stricmp( com, "LIST") || !stricmp( com, "L"))
	{
		com = strtok( NULL, " ");
		if (com == NULL)
		{
			DRAW_ERROR("Wad file name argument missing.");
			return;
		}
		for (wad = WadFileList; wad; wad = wad->next)
			if (!stricmp( com, wad->filename))
				break;
		if (wad == NULL)
		{
			char buf[128];
			sprintf(buf, "Wad file \"%s\" is not open.", com);
			DRAW_ERROR(buf);
			return;
		}
		out = strtok( NULL, " ");
		if (out)
		{
			video->draw_box_3d(0, 150, ScrMaxX-1, ScrMaxY-51);
			c = TranslateToGameColor(GFXBLUE);
			video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, c);
			video->draw_text(10, 160, c, -1, false, false, "Outputting directory of \"%s\" to \"%s\".", wad->filename, out);

			if ((file = fopen( out, "wt")) == NULL)
				ProgError( "error opening output file \"%s\"", com);
			Credits( file);
			ListFileDirectory( file, wad);
			fclose( file);

			video->draw_text(10, 410, TranslateToGameColor(DARKRED), -1, false, true, "Click Mouse Button or Hit Any Key...");
			for (;;)
			{
				GetMouseCoords( &PointerX, &PointerY, &buttons);
				if (startcheck)
					if (buttons == 1)
						break;
				if (keypressed())
					k = readkey();
				if ( k > 0)
					break;
				k = 0;
				if (buttons == 0)
					startcheck = TRUE;

				rest(1);
			}
		}
		else
		{
			TermGfx();
			ListFileDirectory( stdout, wad);
			for (;;)
			{
				GetMouseCoords(&PointerX, &PointerY, &buttons);
				if (keypressed())
					k = readkey();
				if (k >= 1 || buttons == 1)
					break;

				rest(1);
			}
			InitGfx();
			DrawFrontEnd();
		}
	}

	/* user asked for the list of the master directory */
	else if (!stricmp( com, "MASTER") || !stricmp( com, "M"))
	{
		out = strtok( NULL, " ");
		if (out) {
			video->draw_box_3d(0, 150, ScrMaxX-1, ScrMaxY-51);
			video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));

			video->draw_text(10, 160, TranslateToGameColor(GFXBLUE), -1, false, false, "Outputting master directory to \"%s\".", out);
			if ((file = fopen( out, "wt")) == NULL)
				ProgError( "error opening output file \"%s\"", com);
			Credits( file);
			ListMasterDirectory( file);
			fprintf( file, "\nEnd of file.\n");
			fclose( file);
		}
		else
		{
			TermGfx();
			ListMasterDirectory( stdout);
			InitGfx();
			DrawFrontEnd();
		}
	}
	
	else if(!stricmp( com, "DIR") || !stricmp( com, "LS"))
	{
		video->draw_box_3d(0, 150, ScrMaxX-1, ScrMaxY-51);
		video->draw_text(10, 160, TranslateToGameColor(DARKRED), -1, false, true, "Pwads Present in %s", strlen(*Pwads) <= 1 ? Cwd : *Pwads);
		c = TranslateToGameColor(DARKBLUE);
		wadcount = 0;
		x = 10;
		done = al_findfirst("*.wad", &info, FA_RDONLY | FA_HIDDEN | FA_ARCH | FA_LABEL);
		while(!done && wadcount <= 100)
		{
			video->draw_text(x, 180 + (wadcount * 10), c, -1, false, true, info.name);
			video->draw_text(x + 80, 180 + (wadcount * 10), c, -1, false, true, "%dK", (info.size / 1024));
			done = al_findnext(&info);
			wadcount++;
			if (wadcount == 19)
			{
				x += 130;
				wadcount = 0;
			}
		}

		video->draw_text(10, 410, TranslateToGameColor(DARKRED), -1, false, true, "Click Mouse Button or Hit Any Key...");
		for (;;)
		{
			rest(1);

			GetMouseCoords( &PointerX, &PointerY, &buttons);
			if (startcheck)
				if (buttons == 1)
					break;
			if (keypressed())
				k = readkey();
			if (k > 0)
				break;
			if (buttons == 0)
				startcheck = TRUE;
		}
		return;
	}

	/* user asked to read a new patch WAD file */
	else if (!stricmp(com, "READ") || !stricmp(com, "R"))
	{
		com = strtok( NULL, " ");
		if (com == NULL)
		{
			DRAW_ERROR("Wad file name argument missing.");
			return;
		}
		out = (char*)myalloc((strlen(com) + 1) * sizeof(char));
		strcpy(out, com);
		OpenPatchWad(out);
		CloseUnusedWadFiles();
	}
	
	/* user asked to dump the contents of a WAD file */
	else if (!stricmp( com, "DUMP") || !stricmp( com, "D"))
	{
		com = strtok( NULL, " ");
		if (com == NULL) {
			DRAW_ERROR("Object name argument missing.");
			return;
		}
		out = strtok( NULL, " ");
		if (out)
		{
			video->draw_box_3d(0, 150, ScrMaxX-1, ScrMaxY-51);
			video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));
		    c = TranslateToGameColor(GFXBLUE);
			video->draw_text(10, 160, c, -1, false, false, "Outputting directory entry data to \"%s\".", out);

			if ((file = fopen( out, "wt")) == NULL)
				ProgError( "error opening output file \"%s\"", com);
			Credits( file);
			DumpDirectoryEntry( file, com);

			video->draw_text(10, -1, c, -1, false, false, "Complete.", wad->filename, out);
			fclose( file);
		}
		else
		{
			TermGfx();
			DumpDirectoryEntry(stdout, com);
			video->draw_text(10, 410, TranslateToGameColor(DARKRED), -1, false, false, "Click Mouse Button or Hit Any Key...");
			for (;;)
			{
				rest(1);

				GetMouseCoords( &PointerX, &PointerY, &buttons);
				if (startcheck)
					if (buttons == 1)
						break;
				if (keypressed())
					k = readkey();
				if ( k > 0)
					break;
				k = 0;
				if (buttons == 0)
					startcheck = TRUE;
			}
		}
		DrawFrontEnd();
	}
	
	/* user asked to view the sprites */
	else if (!stricmp( com, "VIEW") || !stricmp( com, "V"))
	{
		com = strtok( NULL, " ");
	    video->draw_box_3d(0, 150, ScrMaxX-1, ScrMaxY-51);

		video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));

		if (Caching)
		{
			Caching = FALSE;
			ChooseSprite( -1, 180, "Sprite viewer", com);
			Caching = TRUE;
		}
		else
			ChooseSprite( -1, 180, "Sprite viewer", com);

		DrawFrontEnd();
	}
	
	/* user asked to save an object to a separate PWAD file */
	else if (!stricmp( com, "SAVE") || !stricmp(com, "S"))
	{
		com = strtok( NULL, " ");
		if (com == NULL)
		{
			DRAW_ERROR("Object name argument missing.");
			return;
		}
		if (strlen( com) > 8 || strchr( com, '.') != NULL)
		{
			DRAW_ERROR("Invalid object name.");
			return;
		}
		out = strtok( NULL, " ");
		if (out == NULL)
		{
			DRAW_ERROR("Wad file name argument missing.");
			return;
		}
		for (wad = WadFileList; wad; wad = wad->next)
			if (!stricmp( out, wad->filename))
				break;
		if (wad)
		{
			DRAW_ERROR("This Wad file is already in use.");
			return;
		}
		video->draw_box_3d( 0, 150, ScrMaxX-1, ScrMaxY-51);

		video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));
		video->draw_text(10, 160, TranslateToGameColor(GFXBLUE), -1, false, false, "Saving directory entry data to \"%s\".", out);

		if ((file = fopen( out, "wb")) == NULL)
			ProgError( "error opening output file \"%s\"", out);
		SaveDirectoryEntry( file, com);
		fclose( file);

		video->draw_text(10, 410, TranslateToGameColor(DARKRED), -1, false, true, "Click Mouse Button or Hit Any Key...");
		for (;;)
		{
			rest(1);

			GetMouseCoords( &PointerX, &PointerY, &buttons);
			if (startcheck)
				if (buttons == 1)
					break;
			if (keypressed())
				k = readkey();
			if ( k > 0)
				break;
			if (buttons == 0)
				startcheck = TRUE;
		}
	}

	/* user asked to encapsulate a raw file in a PWAD file */
	else if(!stricmp( com, "INSERT") || !stricmp( com, "I"))
	{
		com = strtok( NULL, " ");
		if (com == NULL)
		{
			DRAW_ERROR("Raw file name argument missing.");
			return;
		}
		out = strtok( NULL, " ");
		if (out == NULL)
		{
			DRAW_ERROR("Object name argument missing.");
			return;
		}
		if (strlen( out) > 8 || strchr( out, '.') != NULL)
		{
			DRAW_ERROR("Invalid object name.");
			return;
		}
		if ((raw = fopen( com, "rb")) == NULL)
			ProgError( "error opening input file \"%s\"", com);
		// [RH] do not copy to input because there is no guarantee it is large enough
		{
			char wadf[255];

			sprintf (wadf, "%s.wad", input);
			for (wad = WadFileList; wad; wad = wad->next)
				if (!stricmp( wadf, wad->filename))
					break;
		}
		if (wad)
		{
			DRAW_ERROR("This Wad file is already in use.");
			return;
		}
		if (DoingBehavior)
		{
			if ((file = fopen( input, "wb")) == NULL)
				ProgError( "error opening output file \"%s\"", input);
			SaveEntryFromRawFile( file, raw, out);
			fclose( raw);
			fclose( file);
			ParseInput ("R BEHAVIOR.WAD");
			return;
		}
		else
		{
			video->draw_box_3d( 0, 150, ScrMaxX-1, ScrMaxY-51);

			video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));
			video->draw_text(10, 160, TranslateToGameColor(GFXBLUE), -1, false, false, "Including new object %s in \"%s\".", out, input);

			if ((file = fopen( input, "wb")) == NULL)
				ProgError( "error opening output file \"%s\"", input);
			SaveEntryFromRawFile( file, raw, out);
			fclose( raw);
			fclose( file);

			video->draw_text(10, 410, TranslateToGameColor(DARKRED), -1, false, true, "Click Mouse Button or Hit Any Key...");
			for (;;)
			{
				rest(1);

				GetMouseCoords( &PointerX, &PointerY, &buttons);
				if (startcheck)
					if (buttons == 1)
						break;
				if (keypressed())
					k = readkey();
				if ( k > 0)
					break;
				if (buttons == 0)
					startcheck = TRUE;
			}
		}
	}
	
	/* user asked to extract an object to a raw binary file */
	else if(!stricmp( com, "XTRACT") || !stricmp( com, "EXTRACT") ||
	        !stricmp( com, "X"))
	{
		com = strtok( NULL, " ");
		if (com == NULL)
		{
			DRAW_ERROR("Object name argument missing.");
			return;
		}
		if (strlen( com) > 8 || strchr( com, '.') != NULL)
		{
			DRAW_ERROR("Invalid object name.");
			return;
		}
		out = strtok( NULL, " ");
		if (out == NULL) {
			printf( "[Raw file name argument missing.]\n");
			return;
		}
		for (wad = WadFileList; wad; wad = wad->next)
			if (!stricmp( out, wad->filename))
				break;
		if (wad)
		{
			DRAW_ERROR("You may not overwrite an opened Wad file with raw data.");
			return;
		}
		video->draw_box_3d( 0, 150, ScrMaxX-1, ScrMaxY-51);

		video->draw_box(130, ScrMaxY - 42, ScrMaxX-1, ScrMaxY - 34, TranslateToGameColor(DARKGRAY));
		video->draw_text(10, 160, TranslateToGameColor(GFXBLUE), -1, false, false, "Saving directory entry data to \"%s\".", out);

		if ((file = fopen( out, "wb")) == NULL)
			ProgError( "error opening output file \"%s\"", out);
		SaveEntryToRawFile( file, com);
		fclose( file);

		video->draw_text(10, 410, TranslateToGameColor(DARKRED), -1, false, true, "Click Mouse Button or Hit Any Key...");
		for (;;)
		{
			rest(1);

			GetMouseCoords( &PointerX, &PointerY, &buttons);
			if (startcheck)
				if (buttons == 1)
					break;
			if (keypressed())
				k = readkey();
			if ( k > 0)
				break;
			if (buttons == 0)
				startcheck = TRUE;
		}
	}
}

//
// the main program menu loop, in text
//
void MainLoop2()
{
	char input[120];
    char *com, *out;
    FILE *file, *raw;
    WadPtr wad;
    int wadcount;
	struct al_ffblk info;
	int done;
	char k;

    for (;;)
	{
		/* get the input */
		printf( "\n[? for help]> ");
		fgets(input, sizeof(input), stdin);
		printf( "\n");

		/* eat the white space and get the first command word */
		com = strtok( input, " ");

		/* user just hit return  */
		if (com == NULL)
			printf( "[Please enter a command.]\n");

		else if (!strcmp( com, "?"))
		{
			printf( "?                                 -- to display this text\n");
			printf( "B[uild] <WadFile>                 -- to build a new main WAD file\n");
			printf( "C[reate] [level]                  -- to create and edit a new (empty) level\n");
			printf( "D[ump] <DirEntry> [outfile]       -- to dump a directory entry in hex\n");
			printf( "E[dit] [level]                    -- to edit a game level\n");
			printf( "G[roup] <WadFile>                 -- to group all patch wads in a file\n");
			printf( "I[nsert] <RawFile> <DirEntry>     -- to insert a raw file in a patch wad file\n");
			printf( "L[ist] <WadFile> [outfile]        -- to list the directory of a wadfile\n");
			printf( "M[aster] [outfile]                -- to list the master directory\n");
			printf( "Q[uit]                            -- to quit\n");
			printf( "R[ead] <WadFile>                  -- to read a new wad patch file\n");
			printf( "S[ave] <DirEntry> <WadFile>       -- to save one object to a separate file\n");
			printf( "W[ads]                            -- to display the open wads\n");
			printf( "X[tract] <DirEntry> <RawFile>     -- to save (extract) one object to a raw file\n");
			printf( "!                                 -- drop to dos\n");
			printf( "DIR / LS                          -- list wads in current directory\n");
		}
		/* user asked for list of open WAD files */
		else if(!stricmp( com, "WADS") || !stricmp( com, "W"))
		{
			printf( "%-20s  IWAD  (Main wad file)\n", WadFileList->filename);
			for (wad = WadFileList->next; wad; wad = wad->next)
			{
				if (wad->directory[ 0].name[ 0] == 'E' && wad->directory[ 0].name[ 2] == 'M')
					printf( "%-20s  PWAD  (Patch wad file for episode %c level %c)\n", wad->filename, wad->directory[ 0].name[ 1], wad->directory[ 0].name[ 3]);
				else
				{
					// [RH] Use a buffer we know to be large enough
					char lumpname[9];
					strncpy( lumpname, wad->directory[ 0].name, 8);
					lumpname[ 8] = '\0';
					printf( "%-20s  PWAD  (Patch wad file for %s)\n", wad->filename, lumpname);
				}
			}
		}
		
		/* user asked to quit */
		else if (!stricmp( com, "QUIT") || !stricmp( com, "Q"))
		{
			SaveConfigFileOptions();
			if (!Registered)
				printf("Remember to register your copy of DOOM!\n");
			printf("Zeth: Beyond DEU\n");
			break;
		}
		
		
		else if (!stricmp( com, "DIR") || !stricmp( com, "LS"))
		{
			printf("Pwads Present in %s\n", strlen(*Pwads) <= 1 ? Cwd : *Pwads);
			wadcount = 0;
			done = al_findfirst("*.wad", &info, FA_RDONLY | FA_HIDDEN | FA_ARCH | FA_LABEL);
			while(!done)
			{
				printf("%-12s  %ldK\n", info.name, info.size /1024);
				done = al_findnext(&info);
				wadcount++;
				if (wadcount == 21)
				{
					k = readkey()&255;
					if (k == 'Q' || k == 'q')
						break;
					wadcount = 0;
				}
			}
		}

		else if (!stricmp( com, "!"))
		{
			printf("Type Exit to return\n");
			remove_keyboard();
			system("");
			Credits(stdout);
			install_keyboard();
		}


		/* user asked to edit a level */
		else if(!stricmp( com, "EDIT")   || !stricmp( com, "E") ||
				!stricmp( com, "CREATE") || !stricmp( com, "C"))
		{
			SList tokens, t;
			
			Bool n = (stricmp(com, "EDIT") && stricmp(com, "E"));
			/* if com is neither of these, the user wants
			   to start from scratch */

			if (n)
				CreatingNewLevel = TRUE;
			
			tokens = NULL;
			com = strtok( NULL, " ");

			/* [RH] Allow specifying maps by name */
			if (com)
			{
				char *probe = com;

				while (*probe >= '0' && *probe <= '9')
					probe++;

				if (*probe == 0 || strlen(probe) > 8)
				{
					while(com)
					{
						tokens = SList_append(tokens, com);
						com = strtok(NULL, " ");
					}
				}
				else
				{
					strcpy (LevelName, com);
				}
			}
			if(tokens)
			{
				MakeLevelName(tokens);
				
				/* now forget the token list */
				while(tokens)
				{
					t = tokens->next;
					free(tokens->string);
					free(tokens);
					tokens = t;
				}
			}
			
			EditLevel(n);
			Credits(stdout);
		}
		
		/* user asked to build a new main WAD file */
		else if (!stricmp( com, "BUILD") || !stricmp( com, "B"))
		{
			com = strtok( NULL, " ");
			if (com == NULL)
			{
				printf( "[Wad file name argument missing.]\n");
				continue;
			}
			for (wad = WadFileList; wad; wad = wad->next)
				if (!stricmp( com, wad->filename))
					break;
			if (wad)
			{
				printf( "[File \"%s\" is opened and cannot be overwritten.]\n", com);
				continue;
			}
			BuildNewMainWad( com, FALSE);
		}
		
		/* user asked to build a compound patch WAD file */
		else if (!stricmp( com, "GROUP") || !stricmp( com, "G"))
		{
			if (WadFileList->next == NULL || WadFileList->next->next == NULL)
			{
				printf( "[You need at least two open wad files if you want to group them.]\n");
				continue;
			}
			com = strtok( NULL, " ");
			if (com == NULL)
			{
				printf( "[Wad file name argument missing.]\n");
				continue;
			}
			for (wad = WadFileList; wad; wad = wad->next)
				if (!stricmp( com, wad->filename))
					break;
			if (wad)
			{
				printf( "[File \"%s\" is opened and cannot be overwritten.]\n", com);
				continue;
			}
			BuildNewMainWad( com, TRUE);
		}
		
		/* user ask for a listing of a WAD file */
		else if (!stricmp( com, "LIST") || !stricmp( com, "L"))
		{
			com = strtok( NULL, " ");
			if (com == NULL)
			{
				printf( "[Wad file name argument missing.]\n");
				continue;
			}
			for (wad = WadFileList; wad; wad = wad->next)
				if (!stricmp( com, wad->filename))
					break;
			if (wad == NULL)
			{
				printf( "[Wad file \"%s\" is not open.]\n", com);
				continue;
			}
			out = strtok( NULL, " ");
			if (out)
			{
				printf( "Outputting directory of \"%s\" to \"%s\".\n", wad->filename, out);
				if ((file = fopen( out, "wt")) == NULL)
					ProgError( "error opening output file \"%s\"", com);
				Credits( file);
				ListFileDirectory( file, wad);
				fprintf( file, "\nEnd of file.\n");
				fclose( file);
			}
			else
				ListFileDirectory( stdout, wad);
		}
		
		/* user asked for the list of the master directory */
		else if (!stricmp( com, "MASTER") || !stricmp( com, "M"))
		{
			out = strtok( NULL, " ");
			if (out) {
				printf( "Outputting master directory to \"%s\".\n", out);
				if ((file = fopen( out, "wt")) == NULL)
					ProgError( "error opening output file \"%s\"", com);
				Credits( file);
				ListMasterDirectory( file);
				fprintf( file, "\nEnd of file.\n");
				fclose( file);
			}
			else
				ListMasterDirectory( stdout);
		}
		
		/* user asked to read a new patch WAD file */
		else if (!stricmp( com, "READ") || !stricmp( com, "R"))
		{
			com = strtok( NULL, " ");
			if (com == NULL)
			{
				printf( "[Wad file name argument missing.]\n");
				continue;
			}
			out = (char*)myalloc((strlen(com) + 1) * sizeof(char));
			strcpy(out, com);
			OpenPatchWad(out);
			CloseUnusedWadFiles();
		}
		
		/* user asked to dump the contents of a WAD file */
		else if (!stricmp( com, "DUMP") || !stricmp( com, "D"))
		{
			com = strtok( NULL, " ");
			if (com == NULL)
			{
				printf( "[Object name argument missing.]\n");
				continue;
			}
			out = strtok( NULL, " ");
			if (out)
			{
				printf( "Outputting directory entry data to \"%s\".\n", out);
				if ((file = fopen( out, "wt")) == NULL)
					ProgError( "error opening output file \"%s\"", com);
				Credits( file);
				DumpDirectoryEntry( file, com);
				fprintf( file, "\nEnd of file.\n");
				fclose( file);
			}
			else
				DumpDirectoryEntry( stdout, com);
		}
		
		/* user asked to save an object to a separate PWAD file */
		else if (!stricmp( com, "SAVE") || !stricmp( com, "S"))
		{
			com = strtok( NULL, " ");
			if (com == NULL)
			{
				printf( "[Object name argument missing.]\n");
				continue;
			}
			if (strlen( com) > 8 || strchr( com, '.') != NULL)
			{
				printf( "[Invalid object name.]\n");
				continue;
			}
			out = strtok( NULL, " ");
			if (out == NULL)
			{
				printf( "[Wad file name argument missing.]\n");
				continue;
			}
			for (wad = WadFileList; wad; wad = wad->next)
				if (!stricmp( out, wad->filename))
					break;
			if (wad)
			{
				printf( "[This Wad file is already in use.  You may not overwrite it.]\n");
				continue;
			}
			printf( "Saving directory entry data to \"%s\".\n", out);
			if ((file = fopen( out, "wb")) == NULL)
				ProgError( "error opening output file \"%s\"", out);
			SaveDirectoryEntry( file, com);
			fclose( file);
		}
		
		/* user asked to encapsulate a raw file in a PWAD file */
		else if (!stricmp( com, "INSERT") || !stricmp( com, "I"))
		{
			com = strtok( NULL, " ");
			if (com == NULL)
			{
				printf( "[Raw file name argument missing.]\n");
				continue;
			}
			out = strtok( NULL, " ");
			if (out == NULL)
			{
				printf( "[Object name argument missing.]\n");
				continue;
			}
			if (strlen( out) > 8 || strchr( out, '.') != NULL)
			{
				printf( "[Invalid object name.]\n");
				continue;
			}
			if ((raw = fopen( com, "rb")) == NULL)
				ProgError( "error opening input file \"%s\"", com);
			// [RH] Use a buffer we know to be large enough
			{
				char wadf[255];
				sprintf (wadf, "%s.WAD", out);
				for (wad = WadFileList; wad; wad = wad->next)
					if (!stricmp( wadf, wad->filename))
						break;
				if (wad)
				{
					printf( "[This Wad file is already in use (%s).  You may not overwrite it.]\n", wadf);
					continue;
				}
				printf( "Including new object %s in \"%s\".\n", out, wadf);
				if ((file = fopen( wadf, "wb")) == NULL)
					ProgError( "error opening output file \"%s\"", wadf);
				SaveEntryFromRawFile( file, raw, out);
				fclose( raw);
				fclose( file);
			}
		}
		
		/* user asked to extract an object to a raw binary file */
		else if(!stricmp( com, "XTRACT") || !stricmp( com, "EXTRACT") ||
				!stricmp( com, "X"))
		{
			com = strtok( NULL, " ");
			if (com == NULL)
			{
				printf( "[Object name argument missing.]\n");
				continue;
			}
			if (strlen( com) > 8 || strchr( com, '.') != NULL)
			{
				printf( "[Invalid object name.]\n");
				continue;
			}
			out = strtok( NULL, " ");
			if (out == NULL)
			{
				printf( "[Raw file name argument missing.]\n");
				continue;
			}
			for (wad = WadFileList; wad; wad = wad->next)
				if (!stricmp( out, wad->filename))
					break;
			if (wad)
			{
				printf( "[You may not overwrite an opened Wad file with raw data.]\n");
				continue;
			}
			printf( "Saving directory entry data to \"%s\".\n", out);
			if ((file = fopen( out, "wb")) == NULL)
				ProgError( "error opening output file \"%s\"", out);
			SaveEntryToRawFile( file, com);
			fclose( file);
		}
		
		/* unknown command */
		else 
			printf("[Unknown command \"%s\"!]\n", com);  
    }
}

void FileRead (const char *name, void **mem, int *size)
{
	FILE *file;
	int len;
	void *locmem;

	*size = 0;
	*mem = NULL;

	if ( (file = fopen (name, "rb")) )
	{
		if (0 == fseek (file, 0, SEEK_END))
		{
			len = ftell (file);
			if (len != -1)
			{
				if (0 == fseek (file, 0, SEEK_SET))
				{
					locmem = myalloc (len);
					if (1 == fread (locmem, len, 1, file))
					{
						*size = len;
						*mem = locmem;
					}
					else
						free (locmem);
				}
			}
		}
		fclose (file);
	}
}


void *myalloc (size_t len)
{
	void *base;

	if(len <= 0)
		len = 1;

	base = malloc(len);
	if(base == NULL)
		ProgError("Out of memory trying to alloc %d bytes\n", len);

	memset(base, 0, len);
	return base;
}

void *myrealloc (void *base, size_t len)
{
	if(len <= 0)
		len = 1;

	base = realloc(base, len);
	if(base == NULL)
		ProgError("Out of memory trying to realloc %d bytes\n", len);

	return base;
}

}	// extern "C"

/* end of file */
