/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.

   INFO.C - object display routines.
   */

/* the includes */
#include "deu.h"
#include "levels.h"

static void DrawArgs(char *argtype, char **argtext, BYTE *argnum, char **enums, int x, int y)
{
	char *foo;
	int i;

	for (i = 0; i < 5; ++i)
	{
		BYTE arg = argnum[i];
		DrawScreenText( -1, -1, "%9s: %d", argtext[i], arg);
	}

	DrawScreenTextSmall(x, y, "");
	for (i = 0; i < 5; ++i)
	{
		BYTE arg = argnum[i];

		switch (argtype[i])
		{
		case 'T':
			foo = GetSpawnName(arg);
			break;

		case 'A':
			foo = GetHexenAngleName(arg);
			break;

		case 'N':
			foo = GetNegativeName(arg);
			break;

		case 'E':
			foo = GetEnumName(arg, enums[i]);
			break;

		case 'F':
			foo = GetUserFlagsName(arg, enums[i]);
			break;

		default:
			foo = NULL;
			break;
		}

		if (foo && *foo)
			DrawScreenTextSmall(-1, -1, "(%s)", foo);
		else
			DrawScreenTextSmall(-1, -1, "");
	}
}

/*
   display the information about one object
   */
void DisplayObjectInfo( BCINT objtype, BCINT objnum) /* SWAP! */
{
	char ctexname[9];
	char ftexname[9];
	char texname[9];
	char foo[40], *argtext;
	BCINT  tag, n, val;
	BCINT  sd1 = -1, sd2 = -1, s1 = -1, s2 = -1;
	BCINT  x0, y0;
	BCINT  extra;
	char *args, **argtexts, **enums;
	BITMAP *buffer;
	buffer = GetBuffer();

	acquire_bitmap(buffer);

	ShowSprite = FALSE;

	switch (objtype)
	{
	case OBJ_THINGS:
		x0 = 0;
		y0 = ScrMaxY - 120;
		if (InfoShown)
			y0 -= 13;

		++mouse_hide_level;
		scare_mouse_area(x0, y0, 360, 128);

		DrawScreenBox3D( x0, y0, x0 + 360, y0 + 120);
		if (objnum < 0)
		{
			SetColor(LIGHTGRAY);
			DrawScreenText( x0 + 120, y0 + 50, "(No Thing Selected)");

			--mouse_hide_level;
			unscare_mouse();
			break;
		}
		extra = GetThingExtra (Things[objnum].type);

//		y0 = ScrMaxY - 120;
//		if (InfoShown)
//			y0 -= 13;
		SetColor( YELLOW);
		DrawScreenText( x0 + 5, y0 + 5, "Selected Thing (#%d),  (%4d)", objnum, Things[ objnum].type);
		SetColor( BLACK);
		DrawScreenText( -1, y0 + 20, "Type     : %s", GetThingName( Things[ objnum].type));
		SetColor(LIGHTGRAY);
		if (extra & TX_POLYSPOT)
			DrawScreenText( -1, -1, "Position : (%d, %d, %d)", Things[ objnum].xpos, Things[ objnum].ypos, Things[ objnum].zpos);
		else
			DrawScreenText( -1, -1, "Position : %d %s (%d, %d, %d)", Things[ objnum].angle, GetAngleName( Things[ objnum].angle), Things[ objnum].xpos, Things[ objnum].ypos, Things[ objnum].zpos);
		SetColor( BLACK);
		DrawScreenText( -1, -1, "                 %s", GetThingFlagsName( Things[ objnum].flags, extra));
		DrawScreenText( -1, y0 + 40, "Flags    : %d", Things[objnum].flags);
		if (extra & TX_POLYSPOT)
			DrawScreenText (-1, -1, "Poly ID  : %d", Things[objnum].angle);
		else
			DrawScreenText( -1, -1, "TID      : %d", Things[objnum].tid);

		if (extra & TX_DEATHSCRIPT)
		{
			DrawScreenText (-1, -1, "Script   : %d", Things[objnum].special);
		}
		else
		{
			args = GetThingArgs (Things[objnum].type);
			if (args == NULL)
			{
				argtexts = GetArgText (Things[objnum].special);
				argtext = GetLineDefArgs (Things[objnum].special);
				enums = GetLineDefEnums (Things[objnum].special);
				DrawScreenText( -1, -1, "Special  : %d %s", Things[ objnum].special ,GetLineDefTypeName( Things[ objnum].special));
			}
			else
			{
				argtexts = GetThingArgText (Things[objnum].type);
				argtext = args;
				enums = GetThingEnums (Things[objnum].type);
				DrawScreenText( -1, -1, "");
			}

//			DrawScreenTextSmall (x0+122, y0+57, "");
			DrawArgs (argtext, argtexts, Things[objnum].arg, enums, x0+122, y0+57);
		}

		if (DispSprite)
		{
//			y0 = ScrMaxY - 120;
//			if (InfoShown)
//				y0 -= 13;
			SetColor(BLACK);

			unscare_mouse();
			scare_mouse_area(ScrMaxX - 120, y0, 120, 120);

			DrawScreenBox3D(ScrMaxX - 120, y0, ScrMaxX, y0 + 120);
			DisplayPicFit( ScrMaxX - 116, y0 + 4, ScrMaxX - 4, y0 + 98, GetThingSprite( Things[ objnum].type));
			if (Things[ objnum].flags & 0x01)
				DrawScreen3DText(ScrMaxX - 112, y0 + 108, RED, DARKERGRAY, "E");
			else
				DrawScreen3DText(ScrMaxX - 112, y0 + 108, DARKERGRAY, LIGHTGRAY, "E");
			if (Things[ objnum].flags & 0x02)
				DrawScreen3DText(ScrMaxX - 100, y0 + 108, RED, DARKERGRAY, "M");
			else
				DrawScreen3DText(ScrMaxX - 100, y0 + 108, DARKERGRAY, LIGHTGRAY, "M");
			if (Things[ objnum].flags & 0x04)
				DrawScreen3DText(ScrMaxX - 88, y0 + 108, RED, DARKERGRAY, "H");
			else
				DrawScreen3DText(ScrMaxX - 88, y0 + 108, DARKERGRAY, LIGHTGRAY, "H");
			if (Things[ objnum].flags & 0x20)
				DrawScreen3DText(ScrMaxX - 74, y0 + 108, BLUE, DARKERGRAY, "F");
			else
				DrawScreen3DText(ScrMaxX - 74, y0 + 108, DARKERGRAY, LIGHTGRAY, "F");
			if (Things[ objnum].flags & 0x40)
				DrawScreen3DText(ScrMaxX - 62, y0 + 108, BLUE, DARKERGRAY, "C");
			else
				DrawScreen3DText(ScrMaxX - 62, y0 + 108, DARKERGRAY, LIGHTGRAY, "C");
			if (Things[ objnum].flags & 0x80)
				DrawScreen3DText(ScrMaxX - 50, y0 + 108, BLUE, DARKERGRAY, "M");
			else
				DrawScreen3DText(ScrMaxX - 50, y0 + 108, DARKERGRAY, LIGHTGRAY, "M");
			if (Things[ objnum].flags & 0x100)
				DrawScreen3DText(ScrMaxX - 36, y0 + 108, GREEN, DARKERGRAY, "S");
			else
				DrawScreen3DText(ScrMaxX - 36, y0 + 108, DARKERGRAY, LIGHTGRAY, "S");
			if (Things[ objnum].flags & 0x200)
				DrawScreen3DText(ScrMaxX - 24, y0 + 108, GREEN, DARKERGRAY, "C");
			else
				DrawScreen3DText(ScrMaxX - 24, y0 + 108, DARKERGRAY, LIGHTGRAY, "C");
			if (Things[ objnum].flags & 0x400)
				DrawScreen3DText(ScrMaxX - 12, y0 + 108, GREEN, DARKERGRAY, "D");
			else
				DrawScreen3DText(ScrMaxX - 12, y0 + 108, DARKERGRAY, LIGHTGRAY, "D");
		}

		--mouse_hide_level;
		unscare_mouse();

		break;

	case OBJ_LINEDEFS:
		strcpy( foo, " ");
		x0 = 0;
		y0 = ScrMaxY - 130;
		if (InfoShown)
			y0 -= 13;

		++mouse_hide_level;
		scare_mouse_area(x0, y0, 258, 130);

		DrawScreenBox3D (x0, y0, x0 + 258, y0 + 130);
		if(objnum < 0)
		{
			SetColor(LIGHTGRAY);
			DrawScreenText( x0 + 45, y0 + 35, "(No LineDef Selected)");

			--mouse_hide_level;
			unscare_mouse();
			break;
		}

		SetColor( YELLOW);
		DrawScreenText( x0 + 5, y0 + 5, "Selected LineDef (#%d)", objnum);
		SetColor( BLACK);
		DrawScreenText( -1, y0 + 20, "Flags    : %s", GetLineDefFlagsName( LineDefs[ objnum].flags));
		switch ((LineDefs[objnum].flags & 0x1c00) >> 10)
		{
			case 0:
				strcpy (foo, "Player Crosses");
				break;
			case 1:
				strcpy (foo, "Player Uses");
				break;
			case 2:
				strcpy (foo, "Monster Crosses");
				break;
			case 3:
				strcpy (foo, "Projectile Hits");
				break;
			case 4:
				strcpy (foo, "Player Pushes");
				break;
			case 5:
				strcpy (foo, "Projectile Crosses");
				break;
			case 6:
				strcpy (foo, "Player Uses (Pass)");
				break;
			default:
				strcpy (foo, "Reserved");
				break;
		}
		DrawScreenText (-1, -1, "Trigger  :");
		if (LineDefs[objnum].special == 0)
			SetColor (DARKERGRAY);
		DrawScreenText( -1, y0+30, "           %s", foo);
		SetColor (BLACK);
		strcpy( foo, " ");
		sd1 = LineDefs[ objnum].sidedef1;
		sd2 = LineDefs[ objnum].sidedef2;
		s1 = LineDefs[ objnum].start;
		s2 = LineDefs[ objnum].end;
		DrawScreenText( -1, -1, "Vertexes : (#%d, #%d)", s1, s2);
		SetColor( LIGHTGRAY);
		n = ComputeDist( Vertexes[ s2].x - Vertexes[ s1].x, Vertexes[ s2].y - Vertexes[ s1].y);
		DrawScreenText( -1, -1, "Length   : %d",n);
		SetColor( BLACK);
		DrawScreenText( -1, -1, "SideDefs : (#%d, #%d)", sd1, sd2);
		val = LineDefs[ objnum].special;
		DrawScreenText( -1, -1, "Special  : %d", val);

		argtext = GetLineDefArgs(LineDefs[ objnum].special);
		DrawArgs (argtext, GetArgText (LineDefs[objnum].special), LineDefs[objnum].arg,
		          GetLineDefEnums(LineDefs[objnum].special), x0+122, y0+67);

		// [RH] Draw the special name with the small font so that we can use longer names.
		DrawScreenTextSmall(x0 + 122, y0 + 67 , GetLineDefTypeName (val));

		if (sd1 >= 0)
			s1 = SideDefs[ sd1].sector;
		else
			s1 = -1;
		if (sd2 >= 0)
			s2 = SideDefs[ sd2].sector;
		else
			s2 = -1;

		x0 = 259;
		y0 = ScrMaxY - 90;
		if (InfoShown)
			y0 -= 13;

		unscare_mouse();
		scare_mouse_area(x0, y0, 190, 90);

		DrawScreenBox3D( x0, y0, x0 + 190, y0 + 90);
		if (sd1 >= 0) {
			SetColor( YELLOW);
			DrawScreenText( x0 + 5, y0 + 5, "First SideDef (#%d)", sd1);
			SetColor( BLACK);
			texname[ 8] = '\0';
			strncpy( texname, SideDefs[ sd1].tex3, 8);
			DrawScreenText( -1, y0 + 20, "Norm. Texture: %s", texname);
			strncpy( texname, SideDefs[ sd1].tex1, 8);
			if (s1 >= 0 && s2 >= 0 && Sectors[ s1].ceilh > Sectors[ s2].ceilh) {
				if (texname[ 0] == '-' && texname[ 1] == '\0')
					SetColor( RED);
			}
			else
				SetColor(GRAY);
			DrawScreenText( -1, -1, "Upper Texture: %s", texname);
			SetColor( BLACK);
			strncpy( texname, SideDefs[ sd1].tex2, 8);
			if (s1 >= 0 && s2 >= 0 && Sectors[ s1].floorh < Sectors[ s2].floorh) {
				if (texname[ 0] == '-' && texname[ 1] == '\0')
					SetColor( RED);
			}
			else
				SetColor(GRAY);
			DrawScreenText( -1, -1, "Lower Texture: %s", texname);
			SetColor( BLACK);
			DrawScreenText( -1, -1, "Tex. X Offset: %d", SideDefs[ sd1].xoff);
			DrawScreenText( -1, -1, "Tex. Y Offset: %d", SideDefs[ sd1].yoff);
			SetColor( LIGHTGRAY);
			DrawScreenText( -1, -1, "Sector Height: %d", Sectors[ s1].ceilh - Sectors[s1].floorh);
			SetColor( BLACK);
			DrawScreenText( -1, -1, "Sector       : #%d", s1);
			SetColor( DARKGRAY);
			if (DispSprite) {
			}
		}
		else {
			SetColor(LIGHTGRAY);
			DrawScreenText( x0 + 24, y0 + 35, "(No First SideDef)");
			DrawScreenButton(x0 + 45, y0 + 60, 100, 16, "Add 1st SideDef");
		}
		x0 = 450;
//		y0 = ScrMaxY - 90;
//		if (InfoShown)
//			y0 -= 13;

		unscare_mouse();
		scare_mouse_area(x0, y0, 189, 90);

		DrawScreenBox3D( x0, y0, x0 + 189, y0 + 90);
		if (sd2 >= 0) {
			SetColor( YELLOW);
			DrawScreenText( x0 + 5, y0 + 5, "Second SideDef (#%d)", sd2);
			SetColor( BLACK);
			texname[ 8] = '\0';
			strncpy( texname, SideDefs[ sd2].tex3, 8);
			
			DrawScreenText( -1, y0 + 20, "Norm. Texture: %s", texname);
			strncpy( texname, SideDefs[ sd2].tex1, 8);
			if (s1 >= 0 && s2 >= 0 && Sectors[ s2].ceilh > Sectors[ s1].ceilh) {
				if (texname[ 0] == '-' && texname[ 1] == '\0')
					SetColor( RED);
			}
			else
				SetColor(GRAY);
			DrawScreenText( -1, -1, "Upper Texture: %s", texname);
			SetColor( BLACK);
			strncpy( texname, SideDefs[ sd2].tex2, 8);
			if (s1 >= 0 && s2 >= 0 && Sectors[ s2].floorh < Sectors[ s1].floorh) {
				if (texname[ 0] == '-' && texname[ 1] == '\0')
					SetColor( RED);
			}
			else
				SetColor(GRAY);
			DrawScreenText( -1, -1, "Lower Texture: %s", texname);
			SetColor( BLACK);
			DrawScreenText( -1, -1, "Tex. X Offset: %d", SideDefs[ sd2].xoff);
			DrawScreenText( -1, -1, "Tex. Y Offset: %d", SideDefs[ sd2].yoff);
			SetColor( LIGHTGRAY);
			DrawScreenText( -1, -1, "Sector Height: %d", Sectors[ s2].ceilh - Sectors[s2].floorh);
			SetColor( BLACK);
			DrawScreenText( -1, -1, "Sector       : #%d", s2);
			if (DispSprite) {
			}
		}
		else {
			SetColor(LIGHTGRAY);
			DrawScreenText( x0 + 20, y0 + 35, "(No Second SideDef)");
			if (sd1 >= 0)
				DrawScreenButton(x0 + 45, y0 + 60, 100, 16, "Add 2nd SideDef");
		}

		if(DispSprite && (sd1 >= 0 || sd2 >= 0))
		{
			x0 = ScrMaxX - 140;
			y0 = ScrMaxY - 210;
			if (VideoMode < 3)
				y0 -= 91;
			if (InfoShown) 
				y0 -= 13;

			unscare_mouse();
			scare_mouse_area(x0, y0, ScrMaxX-x0, ScrMaxY-y0);

			DrawScreenBox3D(x0, y0, ScrMaxX, y0 + 210);

			if(s1 >= 0)
			{
				DisplayTexture(x0 + 4, y0 + 4 , x0 + 68, y0 + 4 + 68, FindTexture(SideDefs[ sd1].tex1, WTexture, WTexNum));
				DisplayTexture(x0 + 4, y0 + 73, x0 + 68, y0 + 73 + 68, FindTexture(SideDefs[ sd1].tex3, WTexture, WTexNum));
				DisplayTexture(x0 + 4, y0 + 142, x0 + 68, y0 + 142 + 68, FindTexture(SideDefs[ sd1].tex2, WTexture, WTexNum));
			}

			if(sd2 >= 0)
			{
				x0 = ScrMaxX - 72;

				DisplayTexture( x0 + 4, y0 + 4 , x0 + 68, y0 + 4 + 64, FindTexture(SideDefs[sd2].tex1, WTexture, WTexNum));
				DisplayTexture( x0 + 4, y0 + 73, x0 + 68, y0 + 73 + 64, FindTexture(SideDefs[sd2].tex3, WTexture, WTexNum));
				DisplayTexture( x0 + 4, y0 + 142, x0 + 68, y0 + 142 + 64, FindTexture(SideDefs[sd2].tex2, WTexture, WTexNum));
			}
		}

		unscare_mouse();
		--mouse_hide_level;

		break;

    case OBJ_VERTEXES:
		x0 = 0;
		y0 = ScrMaxY - 30;
		if (InfoShown)
			y0 -= 13;

		++mouse_hide_level;
		scare_mouse_area(x0, y0, 220, 30);

		DrawScreenBox3D( x0, y0, x0 + 220, y0 + 30);
		if (objnum < 0) {
			SetColor(LIGHTGRAY);
			DrawScreenText( x0 + 30, y0 + 12, "(No Vertex Selected)");
		}
		else {
			SetColor( YELLOW);
			DrawScreenText( x0 + 5, y0 + 5, "Selected Vertex (#%d)", objnum);
			SetColor( BLACK);
			DrawScreenText( -1, y0 + 20, "Coordinates: (%d, %d)", Vertexes[ objnum].x, Vertexes[ objnum].y);
		}

		unscare_mouse();
		--mouse_hide_level;

		break;

    case OBJ_SECTORS:
		x0 = 0;
		y0 = ScrMaxY - 90;
		if (InfoShown)
			y0 -= 13;

		++mouse_hide_level;
		scare_mouse_area(x0, y0, 278, 90);

		DrawScreenBox3D (x0, y0, x0 + 278, y0 + 90);
		if (objnum < 0)
		{
			SetColor(LIGHTGRAY);
			DrawScreenText( x0 + 60, y0 + 42, "(No Sector Selected)");

			unscare_mouse();
			--mouse_hide_level;
			break;
		}
		SetColor( YELLOW);
		DrawScreenText( x0 + 5, y0 + 5, "Selected Sector (#%d)", objnum);
		SetColor( BLACK);
		ftexname[ 8] = '\0';
		ctexname[ 8] = '\0';

		strncpy( ftexname, Sectors[ objnum].floort, 8);
    	DrawScreenText( -1, y0 + 20, "Floor    : %d", Sectors[ objnum].floorh);
    	DrawScreenText( x0 + 151, y0 + 20, "%s", ftexname);

		strncpy( ctexname, Sectors[ objnum].ceilt, 8);
		DrawScreenText( x0 + 5, y0 + 30, "Ceiling  : %d", Sectors[ objnum].ceilh);
    	DrawScreenText( x0 + 151, y0 + 30, "%s", ctexname);

		SetColor(LIGHTGRAY);
		DrawScreenText( x0 + 5, y0 + 40, "Height   : %d", Sectors[ objnum].ceilh - Sectors[objnum].floorh);
		SetColor(BLACK);
		DrawScreenText( -1, -1, "Light    : %d", Sectors[ objnum].light);
		tag = Sectors[ objnum].tag;
		if (tag == 0)
		{
			n = NumLineDefs;
		}
		else
		{
			for (n = 0; n < NumLineDefs; n++)
			{
				char *argtext = GetLineDefArgs (LineDefs[n].special);
				int o;
				for (o = 0; o < 5; ++o)
				{
					if ((argtext[o] == '$' || argtext[o] == '%') && LineDefs[n].arg[o] == tag)
					{
						break;
					}
				}
				if (o < 5)
				{
					break;
				}
			}
		}
		if (n < NumLineDefs)
            DrawScreenText( -1, -1, "Tag      : %d (#%d)", tag, n);
		else
            DrawScreenText( -1, -1, "Tag      : %d (none)", tag);
	    DrawScreenText( -1, -1, "Special  : %d %s", Sectors[objnum].special & 255, GetSectorTypeName( Sectors[ objnum].special));
		DrawScreenText (-1, -1, "Flags    : %s", GetSectorFlags (Sectors[objnum].special));

		if (DispSprite)
		{
			unscare_mouse();
			scare_mouse_area(ScrMaxX - 72, ScrMaxY - 152, 72, 152);

			if (InfoShown)
			{
				DrawScreenBox3D( ScrMaxX - 72, ScrMaxY - 152, ScrMaxX, ScrMaxY - 13);
				DisplayTexture( ScrMaxX - 68, ScrMaxY - 148 , ScrMaxX - 4, ScrMaxY - 148 + 64, FindTexture(ctexname, FTexture, FTexNum));
				DisplayTexture( ScrMaxX - 68, ScrMaxY - 80 , ScrMaxX - 4, ScrMaxY - 80 + 64, FindTexture(ftexname, FTexture, FTexNum));
			}
			else
			{
				DrawScreenBox3D( ScrMaxX - 72, ScrMaxY - 139, ScrMaxX, ScrMaxY);
				DisplayTexture( ScrMaxX - 68, ScrMaxY - 135 , ScrMaxX - 4, ScrMaxY - 135 + 64, FindTexture(ctexname, FTexture, FTexNum));
				DisplayTexture( ScrMaxX - 68, ScrMaxY - 67 , ScrMaxX - 4, ScrMaxY - 67 + 64, FindTexture(ftexname, FTexture, FTexNum));
			}
		}

		unscare_mouse();
		--mouse_hide_level;

		break;
    }

	release_bitmap(buffer);
}



/* end of file */
