/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.

   LEVELS.C - Level loading and saving routines.
   */

/* the includes */
#include "deu.h"
#include "struct_h.h"

/* the global data */
MDirPtr Level = NULL;				/* master dictionary entry for the level */
BCINT NumThings = 0;				/* number of things */
TPtr Things = NULL;					/* things data */
int BehaviorSize = 0;				/* size of behavior */
BPtr Behavior = NULL;				/* behavior data */
int ScriptsSize = 0;				/* size of scripts source code */
char *Scripts = NULL;				/* scripts source code */
int NumLineDefs = 0;				/* number of line defs */
LDPtr LineDefs = NULL;				/* line defs data */
int NumSideDefs = 0;				/* number of side defs */
SDPtr SideDefs = NULL;				/* side defs data */
int NumVertexes = 0;				/* number of vertexes */
VPtr Vertexes = NULL;				/* vertex data */
int NumSectors = 0;				/* number of sectors */
SPtr Sectors = NULL;				/* sectors data */
int NumSegs = 0;					/* number of segments */
SEPtr Segs = NULL;					/* list of segments */
SEPtr LastSeg = NULL;				/* last segment in the list */
int NumSSectors = 0;				/* number of subsectors */
SSPtr SSectors = NULL;				/* list of subsectors */
SSPtr LastSSector = NULL;			/* last subsector in the list */
int NumNodes = 0;					/* number of Nodes */
NPtr Nodes = NULL;					/* nodes tree */

Tex *_WTexture;
int TotalTextures;

int WTexNum = 0;
int NumWTexture = 0;				/* number of wall textures */
Tex **WTexture;					/* array of wall texture names */
int FTexNum = 0;
int NumFTexture = 0;				/* number of floor/ceiling textures */
Tex **FTexture;					/* array of texture names */
BCINT MapMaxX = -32767;				/* maximum X value of map */
BCINT MapMaxY = -32767;				/* maximum Y value of map */
BCINT MapMinX = 32767;				/* minimum X value of map */
BCINT MapMinY = 32767;				/* minimum Y value of map */
Bool MadeChanges = FALSE;			/* made changes? */
Bool MadeMapChanges = FALSE;		/* made changes that need rebuilding? */
SelPtr errld = NULL;				/* LineDefs in error (Nodes builder) */

SList Texture_sections = NULL;
SList Ftexture_sections = NULL;

int NumTextureX;
int NumTX;
int NumFlats;
int NumPnames;
int NumSprites;

/* An empty behavior lump containing no scripts or strings */
static unsigned char NullBehavior[16] = "ACS\0\x8\0\0\0\0\0\0\0\0\0\0\0";

/*
   read in the level data
   */

void ReadLevelData()
{
    MDirPtr dir;
    BCINT n;
    BCINT val;
    BCINT OldNumVertexes;
    BCINT *VertexUsed = NULL;

    /* No objects are needed: they may be swapped after they have been read */

/*
    DisplayMessage( -1, -1, "Reading Data For Level %s...", LevelName);
*/
    Level = FindMasterDir( MasterDir, LevelName);	//jff try the name
    if (!Level)
		Level = FindMasterDir( MasterDir, "MAP01");	//jff try DOOM2 default
	if (!Level)
		Level = FindMasterDir (MasterDir, "E1M1");	//jff try DOOM default
	if (!Level)
        ProgError( "level data not found");			//jff if no level, error

	InitTIDList ();

	{
		int i;

		/* read in the Behavior data */
		dir = Level;
		for (i = 0; i < 11 && dir; i++)
			dir = dir->next;

		if (dir && !strncmp (dir->dir.name, "BEHAVIOR", 8))
		{
			BehaviorSize = dir->dir.size;
			if (BehaviorSize > 0)
			{
				Behavior = (BPtr)myalloc (BehaviorSize);
				BasicWadSeek (dir->wadfile, dir->dir.start);
				BasicWadRead (dir->wadfile, Behavior, BehaviorSize);
			}

			/* read in the Scripts data */
			dir = dir->next;
			if (dir && !strncmp (dir->dir.name, "SCRIPTS", 8))
			{
				ScriptsSize = dir->dir.size;
				if (ScriptsSize > 0)
				{
					Scripts = (char *)myalloc (ScriptsSize);
					BasicWadSeek (dir->wadfile, dir->dir.start);
					BasicWadRead (dir->wadfile, Scripts, ScriptsSize);
				}
			}
		}
		else
		{
			BehaviorSize = 0;
			ScriptsSize = 0;
			InitConversion ();
		}
	}

    /* get the number of Vertices */
    dir = FindMasterDir( Level, "VERTEXES");
    if (dir != NULL)
        OldNumVertexes = (BCINT) (dir->dir.size / 4L);
    else
        OldNumVertexes = 0;
    if (OldNumVertexes > 0)
	{
        VertexUsed = (BCINT*)myalloc( OldNumVertexes * sizeof( BCINT));
        for (n = 0; n < OldNumVertexes; n++)
            VertexUsed[ n] = FALSE;
    }

    /* read in the Things data */
    dir = FindMasterDir( Level, "THINGS");
    if (dir != 0)
	{
		if (BehaviorSize)
			NumThings = (BCINT) (dir->dir.size / 20L);
		else
			NumThings = (BCINT) (dir->dir.size / 10L);
    }
	else
	{
        NumThings = 0;
	}
    if (NumThings > 0)
	{
        Things = (TPtr)myalloc( (unsigned long) NumThings * sizeof( struct Thing));
        BasicWadSeek( dir->wadfile, dir->dir.start);
		if (BehaviorSize)
		{
        	for (n = 0; n < NumThings; n++)
			{
				BCINT tid;
				BasicWadRead( dir->wadfile, &(tid), 2);
				BasicWadRead( dir->wadfile, &(Things[ n].xpos), 2);
				BasicWadRead( dir->wadfile, &(Things[ n].ypos), 2);
				BasicWadRead( dir->wadfile, &(Things[ n].zpos), 2);
				BasicWadRead( dir->wadfile, &(Things[ n].angle), 2);
				BasicWadRead( dir->wadfile, &(Things[ n].type), 2);
				BasicWadRead( dir->wadfile, &(Things[ n].flags), 2);
				BasicWadRead( dir->wadfile, &(Things[ n].special), 1);
				BasicWadRead( dir->wadfile, &(Things[ n].arg[0]), 5);

				Things[n].tid = 0;
				SetThingTID (n, tid);
			}
		}
		else
		{
			for (n = 0; n < NumThings; n++)
			{
				BCINT flags;
				memset (&Things[n], 0, sizeof(struct Thing));
				BasicWadRead (dir->wadfile, &(Things[n].xpos), 2);
				BasicWadRead (dir->wadfile, &(Things[n].ypos), 2);
				BasicWadRead (dir->wadfile, &(Things[n].angle), 2);
				BasicWadRead (dir->wadfile, &(Things[n].type), 2);
				BasicWadRead (dir->wadfile, &flags, 2);
				Things[n].flags = ((flags & 0xf) | 0x7e0);
				if (flags & 0x0010)
					Things[n].flags &= ~0x0100;
				if (flags & 0x0020)
					Things[n].flags &= ~0x0400;
				if (flags & 0x0040)
					Things[n].flags &= ~0x0200;

				Things[n].nextByID = -1;
				Things[n].prevByID = -1;
			}
		}
    }

    /* read in the LineDef information */
    dir = FindMasterDir( Level, "LINEDEFS");
    if (dir != NULL)
	{
		if (BehaviorSize)
			NumLineDefs = (BCINT) (dir->dir.size / 16L);
		else
			NumLineDefs = (BCINT) (dir->dir.size / 14L);
    }
	else
	{
        NumLineDefs = 0;
	}
    if (NumLineDefs > 0)
	{
        LineDefs = (LDPtr)myalloc( (unsigned long) NumLineDefs * sizeof( struct LineDef));
        BasicWadSeek( dir->wadfile, dir->dir.start);
        for (n = 0; n < NumLineDefs; n++)
		{
            BasicWadRead( dir->wadfile, &(LineDefs[ n].start), 2);
            VertexUsed[ LineDefs[ n].start] = TRUE;
            BasicWadRead( dir->wadfile, &(LineDefs[ n].end), 2);
            VertexUsed[ LineDefs[ n].end] = TRUE;
            BasicWadRead( dir->wadfile, &(LineDefs[ n].flags), 2);
			if (BehaviorSize)
			{
				BasicWadRead( dir->wadfile, &(LineDefs[ n].special), 1);
				BasicWadRead( dir->wadfile, &(LineDefs[ n].arg[0]), 5);
			}
			else
			{
				BCINT typeAndTag[2];
				BasicWadRead (dir->wadfile, typeAndTag, 4);
				TranslateLineDef (n, typeAndTag[0], typeAndTag[1]);
			}
            BasicWadRead( dir->wadfile, &(LineDefs[ n].sidedef1), 2);
            BasicWadRead( dir->wadfile, &(LineDefs[ n].sidedef2), 2);
        }
    }

    /* read in the SideDef information */
    dir = FindMasterDir( Level, "SIDEDEFS");
    if (dir != NULL)
        NumSideDefs = (BCINT) (dir->dir.size / 30L);
    else
        NumSideDefs = 0;
    if (NumSideDefs > 0)
	{
        SideDefs = (SDPtr)myalloc( (unsigned long) NumSideDefs * sizeof( struct SideDef));
        BasicWadSeek( dir->wadfile, dir->dir.start);
        for (n = 0; n < NumSideDefs; n++)
		{
            BasicWadRead( dir->wadfile, &(SideDefs[ n].xoff), 2);
            BasicWadRead( dir->wadfile, &(SideDefs[ n].yoff), 2);
            BasicWadRead( dir->wadfile, &(SideDefs[ n].tex1), 8);
            BasicWadRead( dir->wadfile, &(SideDefs[ n].tex2), 8);
            BasicWadRead( dir->wadfile, &(SideDefs[ n].tex3), 8);
            BasicWadRead( dir->wadfile, &(SideDefs[ n].sector), 2);
        }
    }

	/* Read the vertices. If the wad has been run through a nodes
	   builder, there is a bunch of vertices at the end that are not
	   used by any linedefs. Those vertices have been created by the
	   nodes builders for the segs. We ignore them, because they're
	   useless to the level designer. However, we do NOT ignore
	   unused vertices in the middle because that's where the
	   "string art" bug came from.

	   Note that there is absolutely no guarantee that the nodes
	   builders add their own vertices at the end, instead of at the
	   beginning or right in the middle, AFAIK. It's just that they
	   all seem to do that (1). What if some don't ? Well, we would
	   end up with many unwanted vertices in the level data. Nothing
	   that a good CheckCrossReferences() couldn't take care of. */
#if 1
	{
		int lastVertex = -1;
		for (n = 0; n < OldNumVertexes; n++)
			if (VertexUsed[n])
				lastVertex = n;
		NumVertexes = lastVertex + 1;

		if (NumVertexes > 0)
		{
			Vertexes = (VPtr)myalloc ((unsigned long) NumVertexes * sizeof(struct Vertex));
			dir = FindMasterDir(Level, "VERTEXES");
			BasicWadSeek(dir->wadfile, dir->dir.start);
			MapMaxX = -32767;
			MapMaxY = -32767;
			MapMinX = 32767;
			MapMinY = 32767;
			for (n = 0; n < NumVertexes; n++)
			{
				BasicWadRead (dir->wadfile, &val, 2);
				if (val < MapMinX)
					MapMinX = val;
				if (val > MapMaxX)
					MapMaxX = val;
				Vertexes[n].x = val;

				BasicWadRead (dir->wadfile, &val, 2);
				if (val < MapMinY)
					MapMinY = val;
				if (val > MapMaxY)
					MapMaxY = val;
				Vertexes[n].y = val;
			}
			for (; n < OldNumVertexes; n++)
			{
				BasicWadRead (dir->wadfile, &val, 2);
				BasicWadRead (dir->wadfile, &val, 2);
			}
		}

		if (OldNumVertexes > 0)
			free (VertexUsed);
	}
#else
    /* read in the Vertices which are all the corners of the level, but ignore the */
    /* Vertices not used in any LineDef (they usually are at the end of the list). */
    NumVertexes = 0;
    for (n = 0; n < OldNumVertexes; n++)
        if (VertexUsed[n])
            NumVertexes++;
    if (NumVertexes > 0)
	{
        Vertexes = (VPtr)myalloc ((unsigned long) NumVertexes * sizeof(struct Vertex));
        dir = FindMasterDir(Level, "VERTEXES");
        BasicWadSeek(dir->wadfile, dir->dir.start);
        MapMaxX = -32767;
        MapMaxY = -32767;
        MapMinX = 32767;
        MapMinY = 32767;
        m = 0;
        for (n = 0; n < OldNumVertexes; n++)
		{
			if (VertexUsed[n])
			{
				BasicWadRead (dir->wadfile, &val, 2);
				if (val < MapMinX)
					MapMinX = val;
				if (val > MapMaxX)
					MapMaxX = val;
				Vertexes[m].x = val;

				BasicWadRead (dir->wadfile, &val, 2);
				if (val < MapMinY)
					MapMinY = val;
				if (val > MapMaxY)
					MapMaxY = val;
				Vertexes[m].y = val;

				VertexUsed[n] = m++;
			}
			else
			{
				BasicWadRead (dir->wadfile, &val, 2);
				BasicWadRead (dir->wadfile, &val, 2);
			}
        }
        if (m != NumVertexes)
            ProgError ("inconsistency in the Vertex data\n");
    }

    if (OldNumVertexes > 0)
	{
        /* update the Vertex numbers in the LineDefs, in case there were unused
		   vertices in the middle that were removed above */
        for (n = 0; n < NumLineDefs; n++)
		{
            LineDefs[n].start = VertexUsed[LineDefs[n].start];
            LineDefs[n].end = VertexUsed[LineDefs[n].end];
        }
        free (VertexUsed);
    }
#endif

    /* ignore the Segs, SSectors and Nodes */

    /* read in the Sectors information */
    dir = FindMasterDir( Level, "SECTORS");
    if (dir != NULL)
        NumSectors = (BCINT) (dir->dir.size / 26L);
    else
        NumSectors = 0;
    if (NumSectors > 0)
	{
        Sectors = (SPtr)myalloc( (unsigned long) NumSectors * sizeof( struct Sector));
        BasicWadSeek( dir->wadfile, dir->dir.start);
        for (n = 0; n < NumSectors; n++)
		{
            BasicWadRead( dir->wadfile, &(Sectors[ n].floorh), 2);
            BasicWadRead( dir->wadfile, &(Sectors[ n].ceilh), 2);
            BasicWadRead( dir->wadfile, &(Sectors[ n].floort), 8);
            BasicWadRead( dir->wadfile, &(Sectors[ n].ceilt), 8);
            BasicWadRead( dir->wadfile, &(Sectors[ n].light), 2);
            BasicWadRead( dir->wadfile, &(Sectors[ n].special), 2);
            BasicWadRead( dir->wadfile, &(Sectors[ n].tag), 2);
        }
		if (BehaviorSize == 0)
		{
			ConvertSectors ();
		}
    }
	if (BehaviorSize == 0)
	{
		FinishConversion ();
	}
}



/*
   forget the level data
   */

void ForgetLevelData() /* SWAP! */
{
    /* forget the Things */
    NumThings = 0;
    if (Things)
        free( Things);
    Things = NULL;

    /* forget the Vertices */
    NumVertexes = 0;
    if (Vertexes)
        free( Vertexes);
    Vertexes = NULL;

    /* forget the LineDefs */
    NumLineDefs = 0;
    if (LineDefs)
        free( LineDefs);
    LineDefs = NULL;

    /* forget the SideDefs */
    NumSideDefs = 0;
    if (SideDefs)
        free( SideDefs);
    SideDefs = NULL;

    /* forget the Sectors */
    NumSectors = 0;
    if (Sectors)
        free( Sectors);
    Sectors = NULL;

	/* forget the behavior */
	BehaviorSize = 0;
	if (Behavior)
		free( Behavior);
	Behavior = NULL;

	/* forget the scripts source */
	ScriptsSize = 0;
	if (Scripts)
		free (Scripts);
	Scripts = NULL;
}


/*
   save the level data to a PWAD file
   */

void SaveLevelData( char *outfile)
{
    FILE   *file;
    MDirPtr dir;
    long    counter = 11;  
    BCINT   n; 
    long    size;
    long    dirstart;
    long    blocksize;
    long    rejectsize;
    BCINT   oldNumVertexes;
    long    behaviorsize;
    long    mapinfosize = 0;
    long    sndinfosize = 0;

	counter = 12;
	if (ScriptsSize || (BehaviorSize == 0))
		counter++;
	if (KeepInfo) {
		if (FindMasterDir (Level, "MAPINFO"))
			counter++;
		if (FindMasterDir (Level, "SNDINFO"))
			counter++;
	}

	HideMousePointer();

    DisplayMessage( -1, -1, "Saving data to \"%s\"...", outfile);
    LogMessage( ": Saving data to \"%s\"...\n", outfile);
    oldNumVertexes = NumVertexes;
    /* open the file */
    if ((file = fopen( outfile, "wb")) == NULL)
        ProgError( "Unable to open file \"%s\"", outfile);
    WriteBytes( file, "PWAD", 4L);     /* PWAD file */
    WriteBytes( file, &counter, 4L);   /* # of dir entries */
    WriteBytes( file, &counter, 4L);   /* fix this up later */
    counter = 12L; 
    dir = Level->next;

    /* output the things data */
    for (n = 0; n < NumThings; n++) {
        WriteBytes( file, &(Things[ n].tid), 2L);
        WriteBytes( file, &(Things[ n].xpos), 2L);
        WriteBytes( file, &(Things[ n].ypos), 2L);
        WriteBytes( file, &(Things[ n].zpos), 2L);
        WriteBytes( file, &(Things[ n].angle), 2L);
        WriteBytes( file, &(Things[ n].type), 2L);
        WriteBytes( file, &(Things[ n].flags), 2L);
        WriteBytes( file, &(Things[ n].special), 1L);
        WriteBytes( file, &(Things[ n].arg[0]), 5L);
        counter += 20L;
    }
    dir = dir->next;

    /* update MapMinX, MapMinY, MapMaxX, MapMaxY */
    MapMaxX = -32768;
    MapMaxY = -32768;
    MapMinX = 32767;
    MapMinY = 32767;
    for (n = 0; n < NumVertexes; n++) {
        if (Vertexes[ n].x < MapMinX)
            MapMinX = Vertexes[ n].x;
        if (Vertexes[ n].x > MapMaxX)
            MapMaxX = Vertexes[ n].x;
        if (Vertexes[ n].y < MapMinY)
            MapMinY = Vertexes[ n].y;
        if (Vertexes[ n].y > MapMaxY)
            MapMaxY = Vertexes[ n].y;
    }


    /* output the LineDefs */
    for (n = 0; n < NumLineDefs; n++) {
        WriteBytes( file, &(LineDefs[ n].start), 2L);
        WriteBytes( file, &(LineDefs[ n].end), 2L);
        WriteBytes( file, &(LineDefs[ n].flags), 2L);
        WriteBytes( file, &(LineDefs[ n].special), 1L);
        WriteBytes( file, &(LineDefs[ n].arg[0]), 5L);
        WriteBytes( file, &(LineDefs[ n].sidedef1), 2L);
        WriteBytes( file, &(LineDefs[ n].sidedef2), 2L);
        counter += 16L;
    }
    dir = dir->next;

    /* output the SideDefs */
    for (n = 0; n < NumSideDefs; n++) {
        WriteBytes( file, &(SideDefs[ n].xoff), 2L);
        WriteBytes( file, &(SideDefs[ n].yoff), 2L);
        WriteBytes( file, &(SideDefs[ n].tex1), 8L);
        WriteBytes( file, &(SideDefs[ n].tex2), 8L);
        WriteBytes( file, &(SideDefs[ n].tex3), 8L);
        WriteBytes( file, &(SideDefs[ n].sector), 2L);
        counter += 30L;
    }
    dir = dir->next;

    if (MadeMapChanges) {
        /* output the Vertices */
        for (n = 0; n < NumVertexes; n++) {
            WriteBytes( file, &(Vertexes[ n].x), 2L);
            WriteBytes( file, &(Vertexes[ n].y), 2L);
            counter += 4L;
        }
    }
    else {
        /* copy the Vertices */
        size = dir->dir.size;
        counter += size;
        BasicWadSeek( dir->wadfile, dir->dir.start);
        CopyBytes( file, dir->wadfile->fileinfo, size);
    }
    dir = dir->next;

	/* copy the Segs, SSectors and Nodes */
	for (n = 0; n < 3; n++)
	{
		size = dir->dir.size;
		counter += size;
		BasicWadSeek( dir->wadfile, dir->dir.start);
		CopyBytes( file, dir->wadfile->fileinfo, size);
		dir = dir->next;
	}

    /* output the Sectors */
    for (n = 0; n < NumSectors; n++) {
        WriteBytes( file, &(Sectors[ n].floorh), 2L);
        WriteBytes( file, &(Sectors[ n].ceilh), 2L);
        WriteBytes( file, &(Sectors[ n].floort), 8L);
        WriteBytes( file, &(Sectors[ n].ceilt), 8L);
        WriteBytes( file, &(Sectors[ n].light), 2L);
        WriteBytes( file, &(Sectors[ n].special), 2L);
        WriteBytes( file, &(Sectors[ n].tag), 2L);
        counter += 26L;
    }
    dir = dir->next;

	/* copy the Reject data */
	rejectsize = dir->dir.size;
	size = rejectsize;
	counter += size;
	BasicWadSeek( dir->wadfile, dir->dir.start);
	CopyBytes( file, dir->wadfile->fileinfo, size);
	dir = dir->next;

	/* copy the blockmap data */
	blocksize = dir->dir.size;
	size = blocksize;
	counter += size;
	BasicWadSeek( dir->wadfile, dir->dir.start);
	CopyBytes( file, dir->wadfile->fileinfo, size);
	dir = dir->next;

	/* output the behavior data */
	if (BehaviorSize)
	{
		behaviorsize = BehaviorSize;
		counter += behaviorsize;
		WriteBytes (file, Behavior, behaviorsize);

		/* output the scripts source */
		if (ScriptsSize)
		{
			counter += ScriptsSize;
			WriteBytes (file, Scripts, ScriptsSize);
		}
	}
	else
	{
		behaviorsize = sizeof(NullBehavior);
		counter += behaviorsize;
		WriteBytes (file, NullBehavior, behaviorsize);

		counter += 1;
		WriteBytes (file, "\n", 1);
	}

	if (KeepInfo == TRUE) {
		/* copy the mapinfo data */
    	dir = FindMasterDir( Level, "MAPINFO");
		if (dir) {
			mapinfosize = dir->dir.size;
			size = mapinfosize;
			counter += size;
			BasicWadSeek( dir->wadfile, dir->dir.start);
			CopyBytes( file, dir->wadfile->fileinfo, size);
		}

		/* copy the sndinfo data */
    	dir = FindMasterDir( Level, "SNDINFO");
		if (dir) {
			sndinfosize = dir->dir.size;
			size = sndinfosize;
			counter += size;
			BasicWadSeek( dir->wadfile, dir->dir.start);
			CopyBytes( file, dir->wadfile->fileinfo, size);
		}
	}

    /* output the actual directory */
    dirstart = counter;
    counter = 12L;
    size = 0L; /*12L*/
    dir = Level;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, &(dir->dir.name), 8L);
    /*counter += 12;*/
    dir = dir->next;

    size = (long) NumThings * 20L;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, "THINGS\0\0", 8L);
    counter += size;
    dir = dir->next;

    size = (long) NumLineDefs * 16L;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, "LINEDEFS", 8L);
    counter += size;
    dir = dir->next;

    size = (long) NumSideDefs * 30L;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, "SIDEDEFS", 8L);
    counter += size;
    dir = dir->next;

    if (MadeMapChanges)
        size = (long) NumVertexes * 4L;
    else
        size = dir->dir.size;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, "VERTEXES", 8L);
    counter += size;
    dir = dir->next;

	size = dir->dir.size;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, "SEGS\0\0\0\0", 8L);
    counter += size;
    dir = dir->next;

	size = dir->dir.size;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, "SSECTORS", 8L);
    counter += size;
    dir = dir->next;

	size = dir->dir.size;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, "NODES\0\0\0", 8L);
    counter += size;
    dir = dir->next;

    size = (long) NumSectors * 26L;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, "SECTORS\0", 8L);
    counter += size;
    dir = dir->next;

    size = rejectsize;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, "REJECT\0\0", 8L);
    counter += size;
    dir = dir->next;

    size = blocksize;
    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &size, 4L);
    WriteBytes( file, "BLOCKMAP", 8L);
    counter += size;
    dir = dir->next;

    WriteBytes( file, &counter, 4L);
    WriteBytes( file, &behaviorsize, 4L);
    WriteBytes( file, "BEHAVIOR", 8L);
    counter += behaviorsize;

	if (ScriptsSize)
	{
		WriteBytes (file, &counter, 4L);
		WriteBytes (file, &ScriptsSize, 4L);
		WriteBytes (file, "SCRIPTS", 8L);
		counter += ScriptsSize;
	}
	else if (BehaviorSize == 0)
	{
		int temp = 1;
		WriteBytes (file, &counter, 4L);
		WriteBytes (file, &temp, 4L);
		WriteBytes (file, "SCRIPTS", 8L);
		counter += 1;
	}

    if (KeepInfo == TRUE) {
		if (mapinfosize) {
			size = mapinfosize;
			WriteBytes( file, &counter, 4L);
			WriteBytes( file, &size, 4L);
			WriteBytes( file, "MAPINFO", 8L);
			counter += size;
		}
		
		if (sndinfosize) {
			size = sndinfosize;
			WriteBytes( file, &counter, 4L);
			WriteBytes( file, &size, 4L);
			WriteBytes( file, "SNDINFO", 8L);
			counter += size;
		}
    }

    /* fix up the directory start information */
    if (fseek( file, 8L, SEEK_SET))
        ProgError( "error writing to file");
    WriteBytes( file, &dirstart, 4L);

    /* close the file */
    fclose( file);

    NumSegs = 0;
    NumSSectors = 0;
    NumNodes = 0;

    /* delete the vertices added by the Nodes builder */
    if (NumVertexes != oldNumVertexes) {
        NumVertexes = oldNumVertexes;
        myrealloc( Vertexes, NumVertexes * sizeof( struct Vertex));
    }

    /* the file is now up to date */
    MadeChanges = FALSE;

    /* update pointers in Master Directory */
    OpenPatchWad( outfile);

    /* this should free the old "*.BAK" file */
    CloseUnusedWadFiles();

	ShowMousePointer();
}


/* SO 14/4/95:
   SortTextures will set this if a name is repeated in the texture list,
   so that we can look for it and eliminate duplicates */
int dupname;


/*
   function used by qsort to sort the texture names
   */
static int SortTextures(const void *a, const void *b)
{
	int i = stricmp((*(Tex**)a)->name, (*(Tex**)b)->name);
	if(i == 0)
		++dupname;
	return i;
}

static int sort_list(const void *a, const void *b)
{
	return stricmp(((Tex*)a)->name, ((Tex*)b)->name);
}


static void ReadTextureNamesIn(char *section, Tex **list, int *size, int type)
{
    MDirPtr dir,cdir;/*jff*/
	BCINT n, m;
    char start[16];
    char end[16];

	strcpy(start, section); strcat(start, "_START");
	strcpy(end, section); strcat(end, "_END");

	/* while blocks delimited by start and end exist */
	cdir = MasterDir; 	/*JFF use a variable start pointer to support loop */
	while (cdir && (dir = FindMasterDir (cdir, start)))
	{												/*JFF keep looking for more */
		cdir = dir = dir->next;
		/*	count the names */

		if (dir && !isamark(dir->dir.name))
		{
			for (n = 0; dir && !isamark(dir->dir.name); n++)
				dir = dir->next;		/* keep moving until points to end tag */

			/* if end tag is found, and non-empty section */
			if (dir && n>0)	
			{
				/* add the names from master dir. */
				dir = cdir;
				*list = myrealloc(*list, ((*size) + n) * sizeof(Tex));
				for (m = 0; dir &&  m < n; m++)
				{	/* add each new flat name */
					memcpy((*list)[(*size) + m].name, dir->dir.name, 8);
					(*list)[(*size) + m].name[8] = '\0';
					(*list)[(*size) + m].type = type;
					(*list)[(*size) + m].bmp = FALSE;
					dir = dir->next;
				}
				*size += m;
				cdir = dir; /*JFF continue search where last it left off */
			}
		}
	}
}

/*
   read in the wall texture names
   */
void ReadTextureNames()
{
	int i = 0, j, n;
	void *resource;
	Texture *p;
	SList l;

/*
For walls:
1. TX_START/TX_END
2. Textures
3. Flats
4. Patches (PNAMES)
5. Sprites

For floors/ceilings:
1. TX_START/TX_END
2. Flats
3. Textures
4. Patches (PNAMES)
5. Sprites
*/
	/* Add the fake '-' texture */
	TotalTextures = 1;
	_WTexture = myalloc(sizeof(Tex) * TotalTextures);
	strcpy(_WTexture[0].name, "-");
	_WTexture[0].type = TYPE_ALL;
	_WTexture[0].bmp = NULL;

	/* Read in TX_START/TX_END entries */
	j = TotalTextures;
	ReadTextureNamesIn("TX", &_WTexture, &TotalTextures, TYPE_TX);
	NumTX = TotalTextures - j;

	qsort(&(_WTexture[j]), NumTX, sizeof(Tex), sort_list);

	/* Count the number of textures in the TEXTUREx lumps */
	NumTextureX = 0;
	for(l = Texture_sections; l; l = l->next)
	{
    	resource = GetResource(l->string);
		if(!resource)
			continue;

		n = *((long *)resource);
		NumTextureX += n;
	}

	/* Add the TEXTUREx lumps */
	j = TotalTextures;
	TotalTextures += NumTextureX;
	_WTexture = myrealloc(_WTexture, sizeof(Tex) * TotalTextures);
	for(l = Texture_sections; l; l = l->next) 
	{
	    resource = GetResource(l->string);
		if(!resource)
			continue;

	    n = *((long *)resource);

		for(i = 1; i <= n; i++)
		{
			p = (Texture *)(((long *)resource)[i]);
			p = (Texture *)((long)p + (long)resource);

			memcpy(_WTexture[j+i-1].name, p->name, 8);
			_WTexture[j+i-1].name[8] = '\0';
			_WTexture[j+i-1].type = TYPE_TEXTURE;
			_WTexture[j+i-1].bmp = NULL;
		}

		qsort(&(_WTexture[j]), n, sizeof(Tex), sort_list);
		j += n;
	}


	/* Read in the flats */
	j = TotalTextures;
    for(l = Ftexture_sections; l; l = l->next)
		ReadTextureNamesIn(l->string, &_WTexture, &TotalTextures, TYPE_FLAT);
	NumFlats = TotalTextures - j;

	qsort(&(_WTexture[j]), NumFlats, sizeof(Tex), sort_list);


	/* Get PNAMES */
	resource = GetResource("PNAMES");
	if(resource)
	{
		void *pnames;

		NumPnames = *(long*)resource;
		pnames = resource + 4;

		j = TotalTextures;
		TotalTextures += NumPnames;

		_WTexture = myrealloc(_WTexture, sizeof(Tex) * TotalTextures);
		for(i = 0;i < NumPnames;++i)
		{
			memcpy(_WTexture[j+i].name, pnames + i*8, 8);
			_WTexture[j+i].name[8] = '\0';
			_WTexture[j+i].type = TYPE_PNAME;
			_WTexture[j+i].bmp = NULL;
		}

		qsort(&(_WTexture[j]), NumPnames, sizeof(Tex), sort_list);
	}


	/* And finally, get sprites */
	j = TotalTextures;
	ReadTextureNamesIn("S", &_WTexture, &TotalTextures, TYPE_SPRITE);
	NumSprites = TotalTextures - j;

	qsort(&(_WTexture[j]), NumSprites, sizeof(Tex), sort_list);


	/* Build the wall texture list(it's the same as the main list) */
	WTexNum = TotalTextures;
	WTexture = myalloc(sizeof(Tex*) * WTexNum);
	for(i = 0;i < WTexNum;++i)
		WTexture[i] = &(_WTexture[i]);

	/* Remove bad textures(duplicates) */
	j = WTexNum;
	for(i = 0;i < WTexNum;++i)
	{
		int i2;

		for(i2 = i+1;i2 < WTexNum;++i2)
		{
			if(strcmp(WTexture[i2]->name, WTexture[i]->name) == 0)
			{
				int t = i2;

				--WTexNum;
				while(t+1 < WTexNum)
				{
					WTexture[t] = WTexture[t+1];
					++t;
				}
				--i2;
			}
		}
	}
	if(j != WTexNum)
		WTexture = myrealloc(WTexture, sizeof(Tex*) * WTexNum);


	/* Build the floor texture list */
	FTexNum = TotalTextures - 1; /* Floor doesn't use '-' */
	FTexture = myalloc(sizeof(Tex*) * FTexNum);

	/* Copy TX_* first */
	j = 0;
	for(i = 0;i < NumTX;++i, ++j)
		FTexture[j] = &(_WTexture[j+1]);

	/* Flats */
	for(i = 0;i < NumFlats;++i, ++j)
		FTexture[j] = &(_WTexture[j+1+NumTextureX]);

	/* TEXTUREx */
	for(i = 0;i < NumTextureX;++i, ++j)
		FTexture[j] = &(_WTexture[j+1-NumFlats]);

	/* PNAMES */
	for(i = 0;i < NumPnames;++i, ++j)
		FTexture[j] = &(_WTexture[j+1]);

	/* SPRITES */
	for(i = 0;i < NumSprites;++i, ++j)
		FTexture[j] = &(_WTexture[j+1]);

	/* Remove bad textures (duplicates) */
	j = FTexNum;
	for(i = 0;i < FTexNum;++i)
	{
		int i2;

		for(i2 = i+1;i2 < FTexNum;++i2)
		{
			if(strcmp(FTexture[i2]->name, FTexture[i]->name) == 0)
			{
				int t = i2;

				--FTexNum;
				while(t+1 < FTexNum)
				{
					FTexture[t] = FTexture[t+1];
					++t;
				}
				--i2;
			}
		}
	}
	if(j != FTexNum)
		FTexture = myrealloc(FTexture, sizeof(Tex*) * FTexNum);

	if(!SeperateTypes)
	{
		qsort(WTexture, WTexNum, sizeof(Tex*), SortTextures);
		qsort(FTexture, FTexNum, sizeof(Tex*), SortTextures);
	}
}



/*
   forget the wall texture names
   */
void ForgetTextureNames()
{
	if(_WTexture)
	{
		int i;
		for(i = 0;i < TotalTextures;++i)
			if(_WTexture[i].bmp)
				destroy_bitmap(_WTexture[i].bmp);

		free(_WTexture);
		_WTexture = NULL;
	}
	TotalTextures = 0;

	if(WTexture)
	{
		free(WTexture);
		WTexture = NULL;
	}
    NumWTexture = WTexNum = 0;
	if(FTexture)
	{
		free(FTexture);
		FTexture = NULL;
	}
    NumFTexture = FTexNum = 0;
}

/* end of file */
