/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.

   Menu utilities by Raphal Quinet <quinet@montefiore.ulg.ac.be>
   Drop-down menus by Dewi Morgan <d.morgan@bradford.ac.uk>

   MENUS.C - Simple menus for DEU (or other programs).
   */

/* the includes */
#include "deu.h"
#include "levels.h"

/*
   menu of miscellaneous operations
   */



BCINT MiscOperations( BCINT x0, BCINT y0, BCINT objtype, SelPtr *list, BCINT CurObject)
{
    char   msg[ 80];
    BCINT	m;
    BCINT	angle, scale;
    
    sprintf( msg, "Rotate and Scale %s", GetEditModeName( objtype));
    if (objtype == OBJ_VERTEXES) {
		m = DisplayMenu( x0, y0, ((x0 == -1) ? "Misc. Operations" : NULL),
						  "Find First Free Tag Number",
						  "-",
						  msg,
						  "Delete Vertex And Join LineDefs",
						  "Merge Several Vertices Into One",
						  "Add A LineDef And Split Sector",
						  NULL);
    }
    else if (objtype == OBJ_LINEDEFS) {
		m = DisplayMenu( x0, y0, ((x0 == -1) ? "Misc. Operations" : NULL),
						  "Find First Free Tag Number",
						  "Find First Free Line ID",
						  "Find First Free Poly ID",
						  "-",
						  msg,
						  "Split LineDef And Add A Vertex",
						  "Split LineDefs And Add A Sector",
						  "Delete LineDefs And Join Sectors",
						  "Connect Linedefs And Split Donut",
						  "Connect Linedefs And Make Hallway",
						  "-",
						  "Flip LineDef",
						  "Swap SideDefs",
						  "-",
						  "Align Textures On Y",
						  "Align Textures On X",
						  "-",
						  "Search and Replace Textures",
						  "-",
						  "Renumber Selected LineDefs Lowest",
						  NULL);
    }
    else if (objtype == OBJ_SECTORS) {
		m = DisplayMenu( x0, y0, ((x0 == -1) ? "Misc. Operations" : NULL),
						  "Find First Free Tag Number",
						  "-",
						  msg,
						  "Make Door From Sector",
						  "Make Lift From Sector",
						  "-",
						  "Distribute Sector Floor Heights",
						  "Distribute Sector Ceiling Heights",
						  "Distribute Sector Light Levels",
						  "-",
						  "Search and Replace Textures",
						  "-",
						  "Clear all Secret Sectors",
						  "-",
						  "Renumber Selected Sectors Lowest",
						  NULL);
    }														 
    else {
		m = DisplayMenu( x0, y0, ((x0 == -1) ? "Misc. Operations" : NULL),
						  "Find First Free Tag Number",
						  "Find First Free Tid Number",
						  "Find First Free Poly ID",
						  "-",
						  msg,
						  NULL);
    }
    if (m <= 0) /*m == -1 || m == -2) */
    	return m;
    if (m > 1 && m != 3 && m != 2 && *list == NULL) {
		DEU_Beep();
		sprintf( msg, "You Must Select At Least One %s", GetObjectTypeName( objtype));
		Notify( -1, -1, msg, NULL);
		return 0;
    }
    if (objtype == OBJ_THINGS) {
		switch(m) {
			case 1:
				sprintf( msg, "First Free Tag Number: %d", FindFreeTag());
				Notify( -1, -1, msg, NULL);
				break;
			case 2:
				sprintf( msg, "First Free Tid Number: %d", FindFreeTid());
				Notify( -1, -1, msg, NULL);
				break;
			case 3:
				sprintf( msg, "First Free Poly ID: %d", FindFreePolyID());
				Notify( -1, -1, msg, NULL);
				break;
			case 5:
				if ((*list)->next == NULL) {
					DEU_Beep();
					sprintf( msg, "You Must Select More Than One Thing");
					Notify( -1, -1, msg, NULL);
					return 0;
				}
				angle = 0;
				scale = 100;
				if (Input2Numbers( -1, -1, "Rotation Angle ()", "Scale (%)", 360, 1000, &angle, &scale))
					Backup(BTHINGS);
					strcpy(BackupText, "Rotate And Scale Things");
					RotateAndScaleObjects( objtype, *list, (double) angle * 0.0174533, (double) scale * 0.01, 0);
				break;
		}
    }
    else if (objtype == OBJ_LINEDEFS) {
		switch(m) {
			case 1:
				sprintf( msg, "First Free Tag Number: %d", FindFreeTag());
				Notify( -1, -1, msg, NULL);
				break;
			case 2:
				sprintf (msg, "First Free Line ID: %d", FindFreeLineID());
				Notify (-1, -1, msg, NULL);
				break;
			case 3:
				sprintf (msg, "First Free Poly ID: %d", FindFreePolyID());
				Notify (-1, -1, msg, NULL);
				break;
			case 5:
				angle = 0;
				scale = 100;
				if (Input2Numbers( -1, -1, "Rotation Angle ()", "Scale (%)", 360, 1000, &angle, &scale))
					Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS);
					strcpy (BackupText, "Rotate And Scale LineDefs");
					RotateAndScaleObjects( objtype, *list, (double) angle * 0.0174533, (double) scale * 0.01, 0);
				break;
			case 6:
				Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS);
				strcpy (BackupText, "Split LineDefs");
				SplitLineDefs( *list);
				break;
			case 7:
				if ((*list)->next == NULL || (*list)->next->next != NULL) {
					DEU_Beep();
					Notify( -1, -1, "You Must Select Exactly Two LineDefs", NULL);
				}
				else {
					Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS | BSECTORS);
					strcpy (BackupText, "Split LineDefs And Add Sector");
					SplitLineDefsAndSector( (*list)->objnum, (*list)->next->objnum);
					ForgetSelection( list);
				}
				break;
			case 8:
				Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS | BSECTORS);
				strcpy (BackupText, "Delete LineDefs and Join Sectors");
				DeleteLineDefsJoinSectors( list);
				break;
			case 9:
				Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS | BSECTORS);
				strcpy (BackupText, "Connect LineDefs and Split Donut");
				FunctionConnectLineDefsSplitDonut(*list);
				ForgetSelection( list);
				break;
			case 10:
				Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS | BSECTORS);
				strcpy (BackupText, "Connect LineDefs and Make Hallway");
				FunctionConnectLineDefsMakeHallway(*list);
				ForgetSelection( list);
				break;
			case 12:
				Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS);
				strcpy (BackupText, "Flip SideDefs");
				FlipLineDefs( *list, TRUE);
				break;
			case 13:
				Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS);
				strcpy (BackupText, "Swap SideDefs");
				FlipLineDefs( *list, FALSE);
				break;
			case 15:
				YAlignmentMenu( list);
				break;
			case 16:
				XAlignmentMenu( list);
				break;
			case 18:
				{
					char origtexture[9];
					int newtexture;				//jff add Replace Any texture functions
					int n, count, val;
					int *sdfs=NULL;
					SelPtr cur;

					// Make a table of selected sidedefs

					sdfs = myalloc(NumSideDefs*sizeof(int));
					memset(sdfs,0,NumSideDefs*sizeof(int));	// default none selected

					// if something selected other than CurObject
					if (*list && ((*list)->next!=NULL || (*list)->objnum!=CurObject)) 
						for (cur=*list;cur;cur=cur->next)			// restrict list to selected
						{																			// sidedefs
							if (LineDefs[cur->objnum].sidedef1 >=0)
								sdfs[LineDefs[cur->objnum].sidedef1]=1;
							if (LineDefs[cur->objnum].sidedef2 >=0)
								sdfs[LineDefs[cur->objnum].sidedef2]=1;
						}
					else
						memset(sdfs,1,NumSideDefs*sizeof(int));	// default all selected

					val = DisplayMenu( x0, y0, ((x0 == -1) ? "Search For Which Texture?" : NULL),
								  "Normal Texture (Match with 1st SideDef)",
								  "Upper Texture            -- \" --",
								  "Lower Texture            -- \" --",
								  "All Textures from Normal (1st Sidedef)",
								  "All Textures from Upper  -- \" --",
								  "All Textures from Lower  -- \" --",
								  "Normal Texture (Match with 2nd SideDef)",
								  "Upper Texture            -- \" --",
								  "Lower Texture            -- \" --",
								  "All Textures from Normal (2nd Sidedef)",
								  "All Textures from Upper  -- \" --",
								  "All Textures from Lower  -- \" --",
								  NULL);
					switch(val) {
						case 1:
						case 7:
							if (val == 1)
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef1].tex3, 8);
							else
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef2].tex3, 8);
							newtexture = BrowseWalls(origtexture);
							if (newtexture >= 0) {
								Backup (BSIDEDEFS);
								count = 0;
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex3, origtexture, 8) == 0) {
										count ++;
										strncpy(SideDefs[n].tex3, WTexture[newtexture]->name, 8);
									}
								sprintf (BackupText, "Replace %d Normal Wall Textures", count);
							}
							break;
						case 2:
						case 8:
							if (val == 2)
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef1].tex1, 8);
							else
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef2].tex1, 8);
							newtexture = BrowseWalls(origtexture);
							if (newtexture >= 0) {
								Backup (BSIDEDEFS);
								count = 0;
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex1, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex1, WTexture[newtexture]->name, 8);
										count++;
									}
								sprintf (BackupText, "Replace %d Upper Wall Textures", count);
							}
							break;
						case 3:
						case 9:
							if (val == 3)
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef1].tex2, 8);
							else
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef2].tex2, 8);
							newtexture = BrowseWalls(origtexture);
							if (newtexture >= 0) {
								Backup (BSIDEDEFS);
								count = 0;
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex2, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex2, WTexture[newtexture]->name, 8);
										count++;
									}
								sprintf (BackupText, "Replace %d Lower Wall Textures", count);
							}
							break;
						case 4:
						case 10:
							if (val == 4)
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef1].tex3, 8);
							else
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef2].tex3, 8);
							newtexture = BrowseWalls(origtexture);
							if (newtexture >= 0) {
								Backup (BSIDEDEFS);
								count = 0;
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex2, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex2, WTexture[newtexture]->name, 8);
										count++;
									}
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex1, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex1, WTexture[newtexture]->name, 8);
										count++;
									}
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex3, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex3, WTexture[newtexture]->name, 8);
										count++;
									}
								sprintf (BackupText, "Replace %d Wall Textures", count);
							}
							break;
						case 5:
						case 11:
							if (val == 5)
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef1].tex1, 8);
							else
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef2].tex1, 8);
							newtexture = BrowseWalls(origtexture);
							if (newtexture >= 0) {
								Backup (BSIDEDEFS);
								count = 0;
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex2, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex2, WTexture[newtexture]->name, 8);
										count++;
									}
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex1, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex1, WTexture[newtexture]->name, 8);
										count++;
									}
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex3, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex3, WTexture[newtexture]->name, 8);
										count++;
									}
								sprintf (BackupText, "Replace %d Wall Textures", count);
							}
							break;
						case 6:
						case 12:
							if (val == 6)
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef1].tex2, 8);
							else
								strncpy(origtexture, SideDefs[LineDefs[CurObject].sidedef2].tex2, 8);
							newtexture = BrowseWalls(origtexture);
							if (newtexture >= 0) {
								Backup (BSIDEDEFS);
								count = 0;
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex2, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex2, WTexture[newtexture]->name, 8);
										count++;
									}
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex1, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex1, WTexture[newtexture]->name, 8);
										count++;
									}
								for (n = 0; n < NumSideDefs; n++)
									if (sdfs[n] && strncmp(SideDefs[n].tex3, origtexture, 8) == 0) {
										strncpy(SideDefs[n].tex3, WTexture[newtexture]->name, 8);
										count++;
									}
								sprintf (BackupText, "Replace %d Wall Textures", count);
							}
							break;
						}

				free(sdfs);
				break;
				}
			case 20:
				FunctionRenumberLineDefsLowest(CurObject,*list);
				ForgetSelection (list);
				CurObject = -1;
				break;

		}
    }
    else if (objtype == OBJ_VERTEXES) {
		switch(m) {
			case 1:
				sprintf( msg, "First Free Tag Number: %d", FindFreeTag());
				Notify( -1, -1, msg, NULL);
				break;
			case 3:
				if ((*list)->next == NULL) {
					DEU_Beep();
					sprintf( msg, "You Must Select More Than One Vertex.");
					Notify( -1, -1, msg, NULL);
					return 0;
				}
				angle = 0;
				scale = 100;
				if (Input2Numbers( -1, -1, "Rotation Angle ()", "Scale (%)", 360, 1000, &angle, &scale))
					Backup (BVERTEXES);
					strcpy (BackupText, "Rotate And Scale Vertexes");
					RotateAndScaleObjects( objtype, *list, (double) angle * 0.0174533, (double) scale * 0.01, 0);
				break;
			case 4:
				Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS);
				strcpy (BackupText, "Delete Vertexes And Join LineDefs");
				DeleteVerticesJoinLineDefs( *list);
				ForgetSelection( list);
				break;
			case 5:
				Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS);
				strcpy (BackupText, "Merge Vertexes");
				MergeVertices( list);
				break;
			case 6:
				if ((*list)->next == NULL || (*list)->next->next != NULL) {
					DEU_Beep();
					Notify( -1, -1, "You Must Select Exactly Two Vertices", NULL);
				}
				else {
					Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS | BSECTORS);
					strcpy (BackupText, "Split Sector Between 2 Vertexes");
					SplitSector( (*list)->objnum, (*list)->next->objnum);
					ForgetSelection( list);
				}
				break;
		}
    }
    else if (objtype == OBJ_SECTORS) {
		switch(m) {
			case 1:
				sprintf( msg, "First Free Tag Number: %d", FindFreeTag());
				Notify( -1, -1, msg, NULL);
				break;
			case 3:
				angle = 0;
				scale = 100;
				if (Input2Numbers( -1, -1, "Rotation Angle ()", "Scale (%)", 360, 1000, &angle, &scale))
					Backup (BSECTORS);
					strcpy (BackupText, "Rotate And Scale Sectors");
					RotateAndScaleObjects( objtype, *list, (double) angle * 0.0174533, (double) scale * 0.01, 0);
				break;
			case 4:
				if ((*list)->next != NULL) {
					DEU_Beep();
					Notify( -1, -1, "You Must Select Exactly One Sector", NULL);
				}
				else {
					Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS | BSECTORS);
					strcpy (BackupText, "Make Door From Sector");
					MakeDoorFromSector( (*list)->objnum);
				}
				break;
			case 5:
				if ((*list)->next != NULL) {
					DEU_Beep();
					Notify( -1, -1, "You Must Select Exactly One Sector", NULL);
				}
				else {
					Backup (BVERTEXES | BLINEDEFS | BSIDEDEFS | BSECTORS);
					strcpy (BackupText, "Make Lift From Sector");
					MakeLiftFromSector( (*list)->objnum);
				}
				break;
			case 7:
				if ((*list)->next == NULL || (*list)->next->next == NULL) {
					DEU_Beep();
					Notify( -1, -1, "You Must Select Three Or More Sectors", NULL);
				}
				else {
					Backup (BSECTORS);
					strcpy (BackupText, "Distribute Sector Floor Heights");
					DistributeSectorFloors( *list);
				}
				break;
			case 8:
				if ((*list)->next == NULL || (*list)->next->next == NULL) {
					DEU_Beep();
					Notify( -1, -1, "You Must Select Three Or More Sectors", NULL);
				}
				else {
					Backup (BSECTORS);
					strcpy (BackupText, "Distribute Sector Ceiling Heights");
					DistributeSectorCeilings( *list);
				}
				break;
			case 9:
				if ((*list)->next == NULL || (*list)->next->next == NULL) {
					DEU_Beep();
					Notify( -1, -1, "You Must Select Three Or More Sectors", NULL);
				}
				else {
					Backup (BSECTORS);
					strcpy (BackupText, "Distribute Sector Light Levels");
					DistributeLightLevels( *list);
				}
				break;
			case 11:
			{
				char origtexture[9];
				int newtexture;
				int n, count, val;
				val = DisplayMenu( x0, y0, ((x0 == -1) ? "Replace Which Texture?" : NULL),
							  "Floor Texture",
							  "Ceiling Texture",
							  "Both From Floor",
							  "Both From Ceiling",
							  NULL);
				switch(val) {
					case 1:
						strncpy(origtexture, Sectors[CurObject].floort, 8);
						newtexture = BrowseFlats(origtexture);
						if (newtexture >= 0) {
							Backup (BSECTORS);
							count = 0;
							for (n = 0; n < NumSectors; n++)
								if (strncmp(Sectors[n].floort, origtexture, 8) == 0) {
									strncpy(Sectors[n].floort, FTexture[newtexture]->name, 8);
									count++;
								}
							sprintf (BackupText, "Replace %d Floor Textures", count);
						}
						break;
					case 2:
						strncpy(origtexture, Sectors[CurObject].ceilt, 8);
						newtexture = BrowseFlats(origtexture);
						if (newtexture >= 0) {
							Backup (BSECTORS);
							count = 0;
							for (n = 0; n < NumSectors; n++)
								if (strncmp(Sectors[n].ceilt, origtexture, 8) == 0) {
									strncpy(Sectors[n].ceilt, FTexture[newtexture]->name, 8);
									count++;
								}
							sprintf (BackupText, "Replace %d Ceiling Textures", count);
						}
						break;
					case 3:
					{
						int DoSel;

						DoSel = *list && ((*list)->next!=NULL || (*list)->objnum!=CurObject);
						strncpy(origtexture, Sectors[CurObject].floort, 8);
						newtexture = BrowseFlats(origtexture);
						if (newtexture >= 0) {
							Backup (BSECTORS);
							count = 0;
							for (n = 0; n < NumSectors; n++)
								if (!DoSel || IsSelected(*list,n))
									if (strncmp(Sectors[n].floort, origtexture, 8) == 0)
									{
										strncpy(Sectors[n].floort, FTexture[newtexture]->name, 8);
										count++;
									}
							for (n = 0; n < NumSectors; n++)
								if (!DoSel || IsSelected(*list,n))
									if (strncmp(Sectors[n].ceilt, origtexture, 8) == 0)
									{
										strncpy(Sectors[n].ceilt, FTexture[newtexture]->name, 8);
										count++;
									}
							sprintf (BackupText, "Replace %d Flats", count);
						}
						break;
					}
					case 4:
					{
						int DoSel;

						DoSel = *list && ((*list)->next!=NULL || (*list)->objnum!=CurObject);
						strncpy(origtexture, Sectors[CurObject].ceilt, 8);
						newtexture = BrowseFlats(origtexture);
						if (newtexture >= 0) {
							Backup (BSECTORS);
							count = 0;
							for (n = 0; n < NumSectors; n++)
								if (!DoSel || IsSelected(*list,n))
									if (strncmp(Sectors[n].floort, origtexture, 8) == 0)
									{
										strncpy(Sectors[n].floort, FTexture[newtexture]->name, 8);
										count++;
									}
							for (n = 0; n < NumSectors; n++)
								if (!DoSel || IsSelected(*list,n))
									if (strncmp(Sectors[n].ceilt, origtexture, 8) == 0)
									{
										strncpy(Sectors[n].ceilt, FTexture[newtexture]->name, 8);
										count++;
									}
							sprintf (BackupText, "Replace %d Flats", count);
						}
						break;
					}
				}
				break;
			}
			case 13:
			{
				int n,count;

				Backup (BSECTORS);
				count = 0;
				for (n=0; n< NumSectors; n++)
				{
					Sectors[n].special &= ~0x0400;
				}
				sprintf (BackupText, "Cleared %d Sector Secrets",count);
				break;
			}
			case 15:
				FunctionRenumberSectorsLowest(CurObject,*list);
				ForgetSelection(list);
				break;
		}
    }
    return 0;
}



/*
   display a "Display Options" menu 
*/

void Options( BCINT x0, BCINT y0)
{
    char   *menustr[ MENUARRAYMAX];
    BCINT	 dummy[ MENUARRAYMAX];
    BCINT	 n, val;
    const BCINT items = 20;
	char menulines[21][80];
    Bool	done = FALSE;
    
    if (x0 < 0)
		x0 = (ScrMaxX - 50 * 8) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY + items * 10 + 20 ) / 2;
	for (n = 0; n < items + 1; n++)
		menustr[ n] = menulines[n];
	do {
		sprintf( menustr[items], "Display Options");
		sprintf( menustr[ 0], "LineDef Dimensioning   (%s)", (ShowDistance) ? "On" : "Off");
		sprintf( menustr[ 1], "AutoMap Display        (%s)", (AutoMap) ? "On" : "Off");
		sprintf( menustr[ 2], "Graphical Things       (%s)", (ShowThings) ? "On" : "Off");
		sprintf( menustr[ 3], "Information Bar        (%s)", (InfoShown) ? "On" : "Off");
		sprintf( menustr[ 4], "Alternative Info-bar   (%s)", (AltInfo) ? "On" : "Off");
		sprintf( menustr[ 5], "Graphical Information  (%s)", (DispSprite) ? "On" : "Off");
		sprintf( menustr[ 6], "Draw Things as Circles (%s)", (DrawThingCircles) ? "On" : "Off");
		sprintf( menustr[ 7], "Detailed Map Scrolling (%s)", (DetailScroll) ? "On" : "Off");
		sprintf( menustr[ 8], "-");
		sprintf( menustr[ 9], "Grid                   (%s)", (GridShown) ? "On" : "Off");
		sprintf( menustr[10], "Grid Dashed            (%s)", (GridDashed) ? "On" : "Off");
		sprintf( menustr[11], "Grid Always Accurate   (%s)", GridAccurate ? "Yes" : "No");
		sprintf( menustr[12], "Min Grid               (%d)", MinGrid);
		sprintf( menustr[13], "Max Grid               (%d)", MaxGrid);
		sprintf( menustr[14], "-");
		sprintf( menustr[15], "Set Video Mode         (%d)", VideoMode);
		sprintf( menustr[16], "Set Gamma Correction   (%d)", Gamma);
		sprintf( menustr[17], "Set Scroll Factor      (%d)", DefaultLargeScroll);
		sprintf( menustr[18], "-");
		sprintf( menustr[19], "Exit Menu");
		val = DisplayMenuArray( -1, -1, "Display Options"/*menustr[items]*/, items, NULL, menustr, dummy);

		if (val <= 0)
			done = TRUE;
		switch (val)
		{
		case 1:
			ShowDistance = !ShowDistance;
			break;
		case 2:
			AutoMap = !AutoMap;
			break;
		case 3:
			ShowThings = !ShowThings;
			break;
		case 4:
			InfoShown = !InfoShown;
			break;
		case 5:
			AltInfo = !AltInfo;
			break;
		case 6:
			DispSprite = !DispSprite;
			break;
		case 7:
			DrawThingCircles = !DrawThingCircles;
			break;
		case 8:
			DetailScroll = !DetailScroll;
			break;
		case 10:
			GridShown = !GridShown;
			break;
		case 11:
			GridDashed = !GridDashed;
			break;
		case 12:
			GridAccurate = !GridAccurate;
			break;
		case 13:
			n = InputIntegerValue( -1, -1, 1, 8, MinGrid);
			if (n >= 1)
				MinGrid = n;
			break;
		case 14:
			n = InputIntegerValue( -1, -1, 256, 2048, MaxGrid);
			if (n >= 256)
				MaxGrid = n;
			break;
		case 16:
/*			n = InputIntegerValue( -1, -1, 2, 6, VideoMode);
			if (n >= 2)
				VideoMode = n;*/
			//sprintf(menustr[items], "Video Modes");
			sprintf(menustr[0], " 640x480  (2)");
			sprintf(menustr[1], " 800x600  (3)");
			sprintf(menustr[2], "1024x768  (4)");
			sprintf(menustr[3], "1280x960  (5)");
			sprintf(menustr[4], "1600x1200 (6)");
			sprintf(menustr[5], "-");
			sprintf(menustr[6], "Exit Menu");
			n = DisplayMenuArray(-1, -1, "Video Modes"/*menustr[items]*/, 7, NULL, menustr, dummy);

			if(n <= 5)
				VideoMode = n+1;
			break;
		case 17:
			n = InputIntegerValue( -1, -1, 0, 4, Gamma);
			if (n >= 0) {
				Gamma = n;
				SetDoomPalette(0);
			}
			break;
		case 18:
			n = InputIntegerValue( -1, -1, 2, 32, DefaultLargeScroll);
			if (n >= 2) {
				DefaultLargeScroll = n;
				MoveSpeed = DefaultLargeScroll;
			}
			break;
		case 20:	
			done = TRUE;
			break;
		}			   
		MenuDefault = val - 1;
    } while (done == FALSE);

    if (VideoMode != RealVideoMode) {
		RealVideoMode = VideoMode;
		InitGfx();
	}
    MenuDefault = 0;
}

void PrefabPrefs(BCINT x0, BCINT y0)
{
    char   *menustr[ MENUARRAYMAX];
    BCINT	 dummy[ MENUARRAYMAX];
    BCINT	 n, val;
    char	texname[ 9];
    int		tex;
	char menulines[9][120];
    
    if (x0 < 0)
		x0 = (ScrMaxX - 50 * 8 - 19) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY - 5 * 10 - 28) / 2;
	for (n = 0; n < 8; n++)
		menustr[ n] = menulines[n];
	sprintf( menustr[ 7], "Prefab Preferences");
	sprintf( menustr[ 0], "Load Texture Scheme  (%s)", CurrentScheme == NULL ? "None" : CurrentScheme);
	sprintf( menustr[ 1], "-");
	sprintf( menustr[ 2], "Normal Texture       (%s)", DefaultWallTexture);
	sprintf( menustr[ 3], "Upper Texture        (%s)", DefaultUpperTexture);
	sprintf( menustr[ 4], "Lower Texture        (%s)", DefaultLowerTexture);
	sprintf( menustr[ 5], "Floor Texture        (%s)", DefaultFloorTexture);
	sprintf( menustr[ 6], "Ceiling Texture      (%s)", DefaultCeilingTexture);
	val = DisplayMenuArray( -1, -1, menustr[ 7], 7, NULL, menustr, dummy);
	switch (val) {
	case 1:
		SelectScheme();
		break;
	case 3:
		strcpy( texname, DefaultWallTexture);
		tex = BrowseWalls(texname);
		if (tex >= 0)
			strcpy(DefaultWallTexture, WTexture[tex]->name);
		break;
	case 4:
		strcpy( texname, DefaultUpperTexture);
		tex = BrowseWalls(texname);
		if (tex >= 0)
			strcpy(DefaultUpperTexture, WTexture[tex]->name);
		break;
	case 5:
		strcpy( texname, DefaultLowerTexture);
		tex = BrowseWalls(texname);
		if (tex >= 0)
			strcpy(DefaultLowerTexture, WTexture[tex]->name);
		break;
	case 6:
		strcpy( texname, DefaultFloorTexture);
		tex = BrowseFlats(texname);
		if (tex >= 0)
			strcpy(DefaultFloorTexture, FTexture[tex]->name);
		break;
	case 7:
		strcpy( texname, DefaultCeilingTexture);
		tex = BrowseFlats(texname);
		if (tex >= 0)
			strcpy(DefaultCeilingTexture, FTexture[tex]->name);
		break;
	}
}

void TexturePrefs( BCINT x0, BCINT y0)
{
    char   *menustr[ MENUARRAYMAX];
    BCINT	 dummy[ MENUARRAYMAX];
    BCINT	 n, val, tex;
    char	texname[ 9];
	char menulines[16][90];
    
    if (x0 < 0)
		x0 = (ScrMaxX - 50 * 8 - 19) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY - 5 * 10 - 28) / 2;
    for (n = 0; n < 16; n++)
		menustr[ n] = menulines[n];
    sprintf( menustr[ 0], "Change Default Floor Height     (%d)", (int)DefaultFloorHeight & 0xffff);
    sprintf( menustr[ 1], "Change Default Ceiling Height   (%d)", (int)DefaultCeilingHeight & 0xffff);
	sprintf( menustr[ 2], "Change Default Lighting         (%d)", (int)DefaultLighting & 0xffff);
    sprintf( menustr[ 3], "-");	// jff add default lighting
    sprintf( menustr[ 4], "Change Door Track Texture       (%s)", DefaultDoorTrack);
    sprintf( menustr[ 5], "Change Door Texture             (%s)", DefaultDoorTexture);
	sprintf( menustr[ 6], "Change Door Base Texture        (%s)", DefaultDoorBase);
	sprintf( menustr[ 7], "-");
    sprintf( menustr[ 8], "Change Upper Step Texture       (%s)", DefaultUpperStepTexture);
    sprintf( menustr[ 9], "Change Lower Step Texture       (%s)", DefaultLowerStepTexture);
	sprintf( menustr[10], "-");
    sprintf( menustr[11], "Change Teleport Texture         (%s)", DefaultTeleport.Texture);
	sprintf( menustr[12], "Change Teleport Side Texture    (%s)", DefaultTeleport.Side);
	sprintf( menustr[13], "Change Teleport Top Offset      (%d)", DefaultTeleport.Top);
	sprintf( menustr[14], "Change Teleport Bottom Offset   (%d)", DefaultTeleport.Bottom);
	sprintf( menustr[15], "Change Teleport Decoration      (%d: %s)", DefaultTeleport.Decoration, DefaultTeleport.Decoration ? GetThingName(DefaultTeleport.Decoration) : "None");
    val = DisplayMenuArray( -1, -1, "Other Prefrab Preferences", 16, NULL, menustr, dummy);
    switch (val)
	{
    case 1:
		val = InputIntegerValue( -1, -1, -512, 511, DefaultFloorHeight);
		if (val >= -512)
			DefaultFloorHeight = val;
		break;
    case 2:
		val = InputIntegerValue( -1, -1, -512, 511, DefaultCeilingHeight);
		if (val >= -512)
			DefaultCeilingHeight = val;
		break;
	case 3:	//jff add default lighting
		val = InputIntegerValue (-1, -1, 0, 255, DefaultLighting);
		if (val >= 0)
			DefaultLighting = val;
		break;
    case 5:
		strcpy( texname, DefaultDoorTrack);
		tex = BrowseWalls(texname);
		if (tex >= 0)
			strcpy(DefaultDoorTrack, WTexture[tex]->name);
		break;
    case 6:
		strcpy( texname, DefaultDoorTexture);
		tex = BrowseWalls(texname);
		if (tex >= 0)
			strcpy(DefaultDoorTexture, WTexture[tex]->name);
		break;
	case 7:
		strcpy( texname, DefaultDoorBase);
		tex = BrowseWalls(texname);
		if (tex >= 0)
			strcpy(DefaultDoorBase, WTexture[tex]->name);
		break;
    case 9:
		strcpy( texname, DefaultUpperStepTexture);
		tex = BrowseWalls(texname);
		if (tex >= 0)
			strcpy(DefaultUpperStepTexture, WTexture[tex]->name);
		break;
    case 10:
		strcpy( texname, DefaultLowerStepTexture);
		tex = BrowseWalls(texname);
		if (tex >= 0)
			strcpy(DefaultLowerStepTexture, WTexture[tex]->name);
		break;
    case 12:
		strcpy( texname, DefaultTeleport.Texture);
		tex = BrowseFlats(texname);
		if (tex >= 0)
			strcpy(DefaultTeleport.Texture, FTexture[tex]->name);
		break;
	case 13:
		strcpy( texname, DefaultTeleport.Side);
		tex = BrowseWalls(texname);
		if (tex >= 0)
			strcpy(DefaultTeleport.Side, WTexture[tex]->name);
		break;
	case 14:
		val = InputIntegerValue( -1, -1, 0, 24, DefaultTeleport.Top);
		if (val >= 0)
			DefaultTeleport.Top = val;
		break;
	case 15:
		val = InputIntegerValue( -1, -1, 0, 24, DefaultTeleport.Bottom);
		if (val >= 0)
			DefaultTeleport.Bottom = val;
		break;
	case 16:
		val = SelectThingType ();
		if (val >= 0)
			DefaultTeleport.Decoration = val;
		break;
    }	   
}

/*
   display a "Preferences" menu (change default textures, etc.)
   */

void MiscOptions( BCINT x0, BCINT y0)
{
    char   *menustr[ MENUARRAYMAX];
    BCINT	 dummy[ MENUARRAYMAX];
    BCINT	 n, val;
    Bool	done = FALSE;
	char menulines[23][100];
    
    if (x0 < 0)
		x0 = (ScrMaxX - 50 * 8 - 19) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY - 6 * 10 - 28) / 2;
	for (n = 0; n < 23; n++)
		menustr[ n] = menulines[n];
	do {
		sprintf( menustr[ 0], "Sector Increment              (%d)", (int)Increment);
		sprintf( menustr[ 1], "How Close To Line To Split    (%d)", (int)CloseToLine);
		sprintf( menustr[ 2], "Marquee Stretch Delay         (%d)", (int)StretchDelay);
		sprintf( menustr[ 3], "-");
		sprintf( menustr[ 4], "Graphical Front End           (%s)", (GraphFront) ? "Yes" : "No");
		sprintf( menustr[ 5], "Front End - Prompt Only       (%s)", (UsePromptOnly) ? "Yes" : "No");
		sprintf( menustr[ 6], "Confirmation Of Vertex Join   (%s)", (VertConf) ? "On" : "Off");
		sprintf( menustr[ 7], "Editor Mode                   (%s)", (Expert) ? "Expert" : "Novice");
		sprintf( menustr[ 8], "'Q' Key Is...                 (%s)", (QisQuit) ? "Quit" : "Save");
		sprintf( menustr[ 9], "Additive Selection Box        (%s)", (AdditiveSelBox) ? "Yes" : "No");
		sprintf( menustr[10], "Select Object 0               (%s)", (Select0) ? "Yes" : "No");
		sprintf( menustr[11], "Copy SideDefs                 (%s)", (CopySideDefs) ? "Yes" : "No");
		sprintf( menustr[12], "Swap Mouse Buttons            (%s)", (SwapButtons) ? "Yes" : "No");
		sprintf( menustr[13], "Nest Pull-Down Menus          (%s)", (NestMenus) ? "Yes" : "No");
		sprintf( menustr[14], "Persistant Editing Menus      (%s)", (PersistantMenus) ? "On" : "Off");
		sprintf( menustr[15], "Debugging Mode                (%s)", (Debug) ? "On" : "Off");
		sprintf( menustr[16], "Cache Textures                (%s)", (Caching) ? "On" : "Off");
		sprintf( menustr[17], "Automaticaly Save ini File    (%s)", (SaveIniFile) ? "Yes" : "No");
		sprintf( menustr[18], "Choose Game at Start          (%s)", (ChooseGame) ? "Yes" : "No");
		sprintf( menustr[19], "Delete LineDefs with Sectors  (%s)", DeleteSideLines ? "Yes" : "No");
		sprintf( menustr[20], "Delete Vertexes with LineDefs (%s)", DeleteLineVerts ? "Yes" : "No");
		sprintf( menustr[21], "-");
		sprintf( menustr[22], "Exit Menu");
		val = DisplayMenuArray( -1, -1, "Other ini File Settings", 23, NULL, menustr, dummy);
		if (val <= 0)
			done = TRUE;
		switch (val) {
		case 1:
			n = InputIntegerValue( -1, -1, 1, 256, Increment);
			if (n >= 1)
				Increment = n;
			break;
		case 2:
			n = InputIntegerValue( -1, -1, 0, 16, CloseToLine);
			if (n >= 0)
				CloseToLine = n;
			break;
		case 3:
			n = InputIntegerValue( -1, -1, 1, 100, StretchDelay);
			if (n >= 0)
				StretchDelay = n;
			break;
		case 5:
			GraphFront = !GraphFront;
			break;
		case 6:
			UsePromptOnly = !UsePromptOnly;
			break;
		case 7:
			VertConf = !VertConf;
			break;
		case 8:
			Expert = !Expert;
			break;
		case 9:
			QisQuit = !QisQuit;
			break;
		case 10:
			AdditiveSelBox = !AdditiveSelBox;
			break;
		case 11:
			Select0 = !Select0;
			break;
		case 12:
			CopySideDefs = !CopySideDefs;
			break;
		case 13:
			SwapButtons = !SwapButtons;
			break;
		case 14:
			NestMenus = !NestMenus;
			break;
		case 15:
			PersistantMenus = !PersistantMenus;
			break;
		case 16:
			Debug = !Debug;
			break;
		case 17:
			Caching = !Caching;
			break;
		case 18:
			SaveIniFile = !SaveIniFile;
			break;
		case 19:
			ChooseGame = !ChooseGame;
			break;
		case 20:
			DeleteSideLines = !DeleteSideLines;
			break;
		case 21:
			DeleteLineVerts = !DeleteLineVerts;
			break;
		case 23:
			done = TRUE;
			break;
		}
		MenuDefault = val - 1;
    } while (done == FALSE);
    MenuDefault = 0;
}


void Preferences( BCINT x0, BCINT y0)
{
    BCINT	 dummy[ MENUARRAYMAX];
    BCINT	 val;
    char   *menustr[ MENUARRAYMAX] =
	{
		"Prefab Texture Settings",
		"Other Prefab Settings",
		"-",
		"Display Options",
		"Other ini File Settings"
	};

    if (x0 < 0)
		x0 = (ScrMaxX - 50 * 8 - 19) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY - 5 * 10 - 28) / 2;
    val = DisplayMenuArray( -1, -1, "Preferences", 5, NULL, menustr, dummy);
    switch (val) {
    case 1:
    	PrefabPrefs(-1, -1);
		break;
    case 2:
    	TexturePrefs(-1, -1);
		break;
    case 4:
    	Options(-1, -1);
		break;
    case 5:
    	MiscOptions(-1, -1);
		break;
    }			   
}

BCINT BrowseFlats(char *CurFlat)
{
	int maxnumx, maxnumy, val, n, k;
	int nx, ny, x, y, page;
	int oldval = 0;
	int PointerX, PointerY, buttons;
	char foo[9];
	int currentx = 0;
	int currenty = 0;
	Bool NotThisPage = FALSE;
	Texture *t;

/*	switch (VideoMode)
	{
	case 3:
		maxnumx = 11;
		maxnumy = 7;
		break;
	case 4:
		maxnumx = 14;
		maxnumy = 9;
		break;
	case 5:
		maxnumx = 17;
		maxnumy = 11;
		break;
	case 6:
		maxnumx = 22;
		maxnumy = 14;
		break;
	default:
		maxnumx = 8;
		maxnumy = 5;
		break;
	}*/
	maxnumx = ScrMaxX/72;
	maxnumy = ScrMaxY/80 - 1;

	page = 0;
	strncpy(foo, CurFlat, 8);
	foo[8] = '\0';
	for (n = 0; n < FTexNum; n++) 
		if (strcmp(FTexture[n]->name, foo) == 0) {
			page = (n / (maxnumx * maxnumy));
			break;
		}
	while (1 == 1)
	{
		HideMousePointer();
		clear_to_color(GetBuffer(), TranslateToGameColor(BLACK));
		NotThisPage = TRUE;
		for (ny = 0; ny < maxnumy; ny++)
		{
			for (nx = 0; nx < maxnumx; nx++)
			{
				x = (nx * 72) + 8;
				y = (ny * 80);				
				n = (nx + (ny * maxnumx)) + (maxnumx * maxnumy * page);
				if (n < FTexNum)
				{
					if (strcmp(FTexture[n]->name, foo) == 0)
					{
						SetColor(YELLOW);
						DrawScreenFrame(x - 1, y + 1, x + 64, y + 66);
						DrawScreenFrame(x - 2, y, x + 65, y + 67);
						SetColor(RED);
						DrawScreenBox(x - 2, y + 68, x + 65, y + 78);
						currentx = nx;
						currenty = ny;
						NotThisPage = FALSE;
					}
					SetColor(WHITE);
					DisplayTexture(x, y + 2, x + 64, y + 66, FTexture[n]);
					DrawScreenText(x + 0, y + 70, "%s", FTexture[n]->name);
				}
			}
		}
		if (page > 0) {
			DrawScreenButton(ScrMaxX - 140, ScrMaxY - 16, 40, 16, "<");
			DrawScreenButton(ScrMaxX - 180, ScrMaxY - 16, 30, 16, "<<");
		}
		else {
			DrawScreenButtonGray(ScrMaxX - 140, ScrMaxY - 16, 40, 16, "<");
			DrawScreenButtonGray(ScrMaxX - 180, ScrMaxY - 16, 30, 16, "<<");
		}
		if (n < FTexNum) {
			DrawScreenButton(ScrMaxX - 90, ScrMaxY - 16, 40, 16, ">");
			DrawScreenButton(ScrMaxX - 40, ScrMaxY - 16, 30, 16, ">>");
		}
		else {
			DrawScreenButtonGray(ScrMaxX - 90, ScrMaxY - 16, 40, 16, ">");
			DrawScreenButtonGray(ScrMaxX - 40, ScrMaxY - 16, 30, 16, ">>");
		}
		DrawScreenButton(ScrMaxX - 250, ScrMaxY - 16, 60, 16, "Cancel");
		if (NotThisPage) {
			currentx = 0;
			currenty = 0;
			SetColor(YELLOW);
			x = (currentx * 72) + 8;
			y = (currenty * 80);				
			DrawScreenFrame(x - 1, y + 1, x + 64, y + 66);
			DrawScreenFrame(x - 2, y, x + 65, y + 67);
		}
		ShowMousePointer();
		while (1 == 1) {
			if (keypressed()) {
				k = readkey();
				if (k == PAGEDOWN && n < FTexNum) {
					page++;
					break;
				}
				if (k == PAGEUP && page >= 1) {
					page--;
					break;
				}
				if (k == HOME) {
					page = 0;
					break;
				}
				if (k == END) {
					page = FTexNum / (maxnumx * maxnumy);
					break;
				}
				if (k>>8 == KEY_ESC)
					return -1;

				if (k == CURS_RIGHT || k == CURS_LEFT || k == CURS_DOWN || k == CURS_UP) {
					x = (currentx * 72) + 8;
					y = (currenty * 80);				
					SetColor(BLACK);
					HideMousePointer();
					DrawScreenFrame(x - 1, y + 1, x + 64, y + 66);
					DrawScreenFrame(x - 2, y, x + 65, y + 67);
					ShowMousePointer();

					if (k == CURS_RIGHT)
						currentx++;

					if (k == CURS_LEFT)
						currentx--;

					if (k == CURS_UP)
						currenty--;

					if (k == CURS_DOWN)
						currenty++;

					if (currentx >= maxnumx) {
						currentx = maxnumx - 1;
					}
					if (currentx < 0) {
						currentx = 0;
					}
					if (currenty >= maxnumy)
						currenty = maxnumy - 1;

					if (currenty < 0)
						currenty = 0;
				}

				val = (currentx + (currenty * maxnumx));
				val += maxnumx * maxnumy * page;

				if (k == ENTER)
					return val;

				if (val >= FTexNum) {
					currentx = 0;
					currenty = 0;
					val = 0; /*jff*/
				}

				x = (currentx * 72) + 8;
				y = (currenty * 80);

				SetColor(YELLOW);
				HideMousePointer();
				DrawScreenFrame(x - 1, y + 1, x + 64, y + 66);
				DrawScreenFrame(x - 2, y, x + 65, y + 67);

				SetColor(BLACK);
				DrawScreenBox(0, ScrMaxY - 12, ScrMaxX-251, ScrMaxY);

				SetColor(WHITE);
				DrawScreenText(0, ScrMaxY - 12, "%s", FTexture[val]->name);

				t = GetTextureInfo(FTexture[val]);
				if(t)
				{
					int sx = t->scalex ? t->scalex : 8;
					int sy = t->scaley ? t->scaley : 8;

					DrawScreenText(0, ScrMaxY - 12, "         %ix%i", t->width, t->height);

					if(sx != 8 || sy != 8)
						DrawScreenText(0, ScrMaxY - 12, "                   XScale: %.2f YScale: %.2f",
						               sx/8.0, sy/8.0);
				}

				ShowMousePointer();
			}
			GetMouseCoords(&PointerX, &PointerY, &buttons);
			val = (((PointerX - 8) / 72) + ((PointerY / 80) * maxnumx));
			val += maxnumx * maxnumy * page;

			if (oldval != val) {
				HideMousePointer();

				SetColor(BLACK);
				DrawScreenBox(0, ScrMaxY - 12, ScrMaxX-251, ScrMaxY);
				if (val < FTexNum && val >= 0)
				{
//					SetColor(WHITE);
//					DrawScreenText(0, ScrMaxY - 12, "%s", FTexture[val]->name);
					SetColor(WHITE);
					DrawScreenText(0, ScrMaxY - 12, "%s", FTexture[val]->name);

					t = GetTextureInfo(FTexture[val]);
					if(t)
					{
						int sx = t->scalex ? t->scalex : 8;
						int sy = t->scaley ? t->scaley : 8;

						DrawScreenText(0, ScrMaxY - 12, "         %ix%i", t->width, t->height);

						if(sx != 8 || sy != 8)
							DrawScreenText(0, ScrMaxY - 12, "                   XScale: %.2f YScale: %.2f",
							               sx/8.0, sy/8.0);
					}
				}

				ShowMousePointer();
				oldval = val;
			}
			if (buttons == 1) {
				if(MouseInArea(ScrMaxX - 250, ScrMaxY - 16, 60, 16)) {
					DrawScreenButtonIn(ScrMaxX - 250, ScrMaxY - 16, 60, 16, "Cancel");
					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					if(MouseInArea(ScrMaxX - 250, ScrMaxY - 16, 60, 16))
						return -1;
					break;
				}
				else if(MouseInArea(ScrMaxX - 140, ScrMaxY - 16, 40, 16) && page >= 1) {
					DrawScreenButtonIn(ScrMaxX - 140, ScrMaxY - 16, 40, 16, "<");
					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					if(MouseInArea(ScrMaxX - 140, ScrMaxY - 16, 40, 16))
						page--;
					break;
				}
				else if(MouseInArea(ScrMaxX - 90, ScrMaxY - 16, 40, 16) && n < FTexNum) {
					DrawScreenButtonIn(ScrMaxX - 90, ScrMaxY - 16, 40, 16, ">");
					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					if(MouseInArea(ScrMaxX - 90, ScrMaxY - 16, 40, 16))
						page++;
					break;
				}
				else if(MouseInArea(ScrMaxX - 180, ScrMaxY - 16, 30, 16) && page >= 1) {
					DrawScreenButtonIn(ScrMaxX - 180, ScrMaxY - 16, 30, 16, "<<");
					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					if(MouseInArea(ScrMaxX - 180, ScrMaxY - 16, 30, 16))
						page = 0;
					break;
				}
				else if(MouseInArea(ScrMaxX - 40, ScrMaxY - 16, 30, 16) && page != FTexNum / (maxnumx * maxnumy)) {
					DrawScreenButtonIn(ScrMaxX - 40, ScrMaxY - 16, 30, 16, ">>");
					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					if(MouseInArea(ScrMaxX - 40, ScrMaxY - 16, 30, 16))
						page = FTexNum / (maxnumx * maxnumy);
					break;
				}
				else if (PointerY < ScrMaxY - 20) {
					if (val >= FTexNum)
						val = -1;

					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					return val;
				}
			}		

			rest(1);
		}
	}
}

void WriteLog (char *logstring)
{
	logst = fopen ("zethlog.txt", "a");
	if (logst)
	{
		fprintf (logst, "%s\n", logstring);
		fclose (logst);
	}
}

BCINT BrowseWalls(char *CurTexture)
{
	int maxnumx, maxnumy, val, n, k;
	int nx, ny, x, y, page;
	int oldval = 0;
	char foo[9];
	int PointerX, PointerY, buttons;
	int currentx = 0;
	int currenty = 0;
	Bool NotThisPage = FALSE;
	Texture *t;

/*	switch (VideoMode)
	{
	case 3:
		maxnumx = 11;
		maxnumy = 7;
		break;
	case 4:
		maxnumx = 14;
		maxnumy = 9;
		break;
	case 5:
		maxnumx = 17;
		maxnumy = 12;
		break;
	case 6:
		maxnumx = 22;
		maxnumy = 15;
		break;
	default:
		maxnumx = 8;
		maxnumy = 5;
		break;
	}*/
	maxnumx = ScrMaxX/72;
	maxnumy = ScrMaxY/80 - 1;

	page = 0;
	strncpy(foo, CurTexture, 8);
	foo[8] = '\0';
	for (n = 0; n < WTexNum; n++) 
		if (strncmp(WTexture[n]->name, foo, 8) == 0 )
		{
			page = (n / (maxnumx * maxnumy));
			break;
		}

	while (1 == 1)
	{
		NotThisPage = TRUE;
		HideMousePointer();
		clear_to_color(GetBuffer(), TranslateToGameColor(BLACK));
		for (ny = 0; ny < maxnumy; ny++)
		{
			y = (ny * 80);				
			for (nx = 0; nx < maxnumx; nx++)
			{
				x = (nx * 72) + 8;
				n = (nx + (ny * maxnumx)) + (maxnumx * maxnumy * page);
				if (n < WTexNum)
				{
					if (strcmp(WTexture[n]->name, foo) == 0)
					{
						SetColor(YELLOW);
						DrawScreenFrame(x - 1, y + 1, x + 64, y + 66);
						DrawScreenFrame(x - 2, y, x + 65, y + 67);
						SetColor(RED);
						DrawScreenBox(x - 2, y + 68, x + 65, y + 78);
						currentx = nx;
						currenty = ny;
						NotThisPage = FALSE;
					}
					SetColor(WHITE);
					DisplayTexture(x, y + 2, x + 64, y + 66, WTexture[n]);
					DrawScreenText(x, y + 70, "%s", WTexture[n]->name);
				}
			}
		}

		SetColor(BLACK);
		DrawScreenBox(0, ScrMaxY - 12, ScrMaxX, ScrMaxY);

/*		for (ny = 0; ny < maxnumy; ny++)
		{
			for (nx = 0; nx < maxnumx; nx++)
			{
				x = (nx * (64 + 8)) + 8;
				y = (ny * (64 + 14));				
				n = (nx + (ny * maxnumx)) + (maxnumx * maxnumy * page);
				if (n < WTexNum && strncmp(WTexture[n]->name, foo, 8) == 0 )
				{
					SetColor(YELLOW);
					DrawScreenFrame(x - 1, y + 1, x + 64, y + 66);
					DrawScreenFrame(x - 2, y, x + 65, y + 67);
					SetColor(RED);
					DrawScreenBox(x - 2, y + 68, x + 65, y + 78);
					SetColor(WHITE);
					DrawScreenText(x + 0, y + 64 + 6, "%s", WTexture[n]->name);
					currentx = nx;
					currenty = ny;
					NotThisPage = FALSE;
					break;
				}
			}
		}*/
		if (page > 0)
		{
			DrawScreenButton(ScrMaxX - 140, ScrMaxY - 16, 40, 16, "<");
			DrawScreenButton(ScrMaxX - 180, ScrMaxY - 16, 30, 16, "<<");
		}
		else
		{
			DrawScreenButtonGray(ScrMaxX - 140, ScrMaxY - 16, 40, 16, "<");
			DrawScreenButtonGray(ScrMaxX - 180, ScrMaxY - 16, 30, 16, "<<");
		}
		if (n < WTexNum)
		{
			DrawScreenButton(ScrMaxX - 90, ScrMaxY - 16, 40, 16, ">");
			DrawScreenButton(ScrMaxX - 40, ScrMaxY - 16, 30, 16, ">>");
		}
		else
		{
			DrawScreenButtonGray(ScrMaxX - 90, ScrMaxY - 16, 40, 16, ">");
			DrawScreenButtonGray(ScrMaxX - 40, ScrMaxY - 16, 30, 16, ">>");
		}
		DrawScreenButton(ScrMaxX - 250, ScrMaxY - 16, 60, 16, "Cancel");
		if (NotThisPage)
		{
			currentx = 0;
			currenty = 0;
			SetColor(YELLOW);
			x = (currentx * (64 + 8)) + 8;
			y = (currenty * (64 + 14));				
			DrawScreenFrame(x - 1, y + 1, x + 64, y + 66);
			DrawScreenFrame(x - 2, y, x + 65, y + 67);
		}
		ShowMousePointer();

		for (;;)
		{
			if (keypressed())
			{
				k = readkey();
				if (k == PAGEDOWN && n < WTexNum)
				{
					page++;
					break;
				}
				if (k == PAGEUP && page >= 1)
				{
					page--;
					break;
				}
				if (k == HOME)
				{
					page = 0;
					break;
				}
				if (k == END)
				{
					page = WTexNum / (maxnumx * maxnumy);
					break;
				}
				if((k>>8) == KEY_ESC)
					return -1;

				if (k == CURS_RIGHT || k == CURS_LEFT || k == CURS_DOWN || k == CURS_UP)
				{ 
					x = (currentx * (64 + 8)) + 8;
					y = (currenty * (64 + 16));
					SetColor(BLACK);
					HideMousePointer();
					DrawScreenFrame(x - 1, y + 1, x + 64, y + 66);
					DrawScreenFrame(x - 2, y, x + 65, y + 67);
					ShowMousePointer();

					if (k == CURS_RIGHT)
						currentx++;

					if (k == CURS_LEFT)
						currentx--;

					if (k == CURS_UP)
						currenty--;

					if (k == CURS_DOWN)
						currenty++;

					if (currentx >= maxnumx)
					{
						currentx = maxnumx - 1;
					}
					if (currentx < 0)
					{
						currentx = 0;
					}
					if (currenty >= maxnumy)
						currenty = maxnumy - 1;

					if (currenty < 0)
						currenty = 0;
				}

				val = (currentx + (currenty * maxnumx));
				val = val + (maxnumx * maxnumy * page);

				if (k == ENTER)
					return val;

				if (val >= WTexNum)
				{
					currentx = 0;
					currenty = 0;
					val = 0; /*jff*/
				}

				x = (currentx * (64 + 8)) + 8;
				y = (currenty * (64 + 16));
				SetColor(YELLOW);
				HideMousePointer();
				DrawScreenFrame(x - 1, y + 1, x + 64, y + 66);
				DrawScreenFrame(x - 2, y, x + 65, y + 67);
				SetColor(BLACK);
				DrawScreenBox(0, ScrMaxY - 12, ScrMaxX-251, ScrMaxY);

				SetColor(WHITE);
				DrawScreenText(0, ScrMaxY - 12, "%s", WTexture[val]->name);

				t = GetTextureInfo(WTexture[val]);
				if(t)
				{
					int sx = t->scalex ? t->scalex : 8;
					int sy = t->scaley ? t->scaley : 8;

					DrawScreenText(0, ScrMaxY - 12, "         %ix%i", t->width, t->height);
					if(sx != 8 || sy != 8)
						DrawScreenText(0, ScrMaxY - 12, "                   XScale: %.2f YScale: %.2f",
						               sx/8.0, sy/8.0);
				}

				ShowMousePointer();
			}
			GetMouseCoords( &PointerX, &PointerY, &buttons);
			val = (((PointerX - 8) / 72) + ((PointerY / 80) * maxnumx));
			val = val + (maxnumx * maxnumy * page);

			if (oldval != val)	/* jff val< not val<= */
			{
				HideMousePointer();

				SetColor(BLACK);
				DrawScreenBox(0, ScrMaxY - 12, ScrMaxX-251, ScrMaxY);
				if(val < WTexNum && val >= 0)
				{
//					SetColor(WHITE);
//					DrawScreenText(0, ScrMaxY - 12, "%s", WTexture[val]->name);
					SetColor(WHITE);
					DrawScreenText(0, ScrMaxY - 12, "%s", WTexture[val]->name);

					t = GetTextureInfo(WTexture[val]);
					if(t)
					{
						int sx = t->scalex ? t->scalex : 8;
						int sy = t->scaley ? t->scaley : 8;

						DrawScreenText(0, ScrMaxY - 12, "         %ix%i", t->width, t->height);

						if(sx != 8 || sy != 8)
							DrawScreenText(0, ScrMaxY - 12, "                   XScale: %.2f YScale: %.2f",
							               sx/8.0, sy/8.0);
					}
				}

				ShowMousePointer();
				oldval = val;
			}
			if(buttons == 1)
			{
				if(MouseInArea(ScrMaxX - 250, ScrMaxY - 16, 60, 16))
				{
					DrawScreenButtonIn(ScrMaxX - 250, ScrMaxY - 16, 60, 16, "Cancel");
					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					if(MouseInArea(ScrMaxX - 250, ScrMaxY - 16, 60, 16))
						return -1;
					break;
				}
				else if(MouseInArea(ScrMaxX - 140, ScrMaxY - 16, 40, 16) && page >= 1)
				{
					DrawScreenButtonIn(ScrMaxX - 140, ScrMaxY - 16, 40, 16, "<");
					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					if(MouseInArea(ScrMaxX - 140, ScrMaxY - 16, 40, 16))
						page--;
					break;
				}
				else if(MouseInArea(ScrMaxX - 90, ScrMaxY - 16, 40, 16) && n < WTexNum)
				{
					DrawScreenButtonIn(ScrMaxX - 90, ScrMaxY - 16, 40, 16, ">");
					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					if(MouseInArea(ScrMaxX - 90, ScrMaxY - 16, 40, 16))
						page++;
					break;
				}
				else if(MouseInArea(ScrMaxX - 180, ScrMaxY - 16, 30, 16) && page >= 1)
				{
					DrawScreenButtonIn(ScrMaxX - 180, ScrMaxY - 16, 30, 16, "<<");
					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					if(MouseInArea(ScrMaxX - 180, ScrMaxY - 16, 30, 16))
						page = 0;
					break;
				}
				else if(MouseInArea(ScrMaxX - 40, ScrMaxY - 16, 30, 16) && page != WTexNum / (maxnumx * maxnumy))
				{
					DrawScreenButtonIn(ScrMaxX - 40, ScrMaxY - 16, 30, 16, ">>");
					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					if(MouseInArea(ScrMaxX - 40, ScrMaxY - 16, 30, 16))
						page = WTexNum / (maxnumx * maxnumy);
					break;
				}
				else if (PointerY < ScrMaxY - 20)
				{
					if (val >= WTexNum)
						val = -1;

					while(buttons == 1)
						GetMouseCoords(&PointerX, &PointerY, &buttons);

					return val;
				}
			}		
	
			rest(1);
		}
	}
}

/*
   draws a line of text in a menu
   */

void DisplayMenuText( BCINT x0, BCINT y0, BCINT line, char *text, BCINT highlightnum, Bool shownumbers)
{
	HideMousePointer(); 
    if (shownumbers)
	{
		DrawScreenText( x0 + 26, y0 + 8 + line * 10, "- %s", text);
		SetColor( WHITE);
		if (line < 9)
			DrawScreenText( x0 + 10, y0 + 8 + line * 10, "%d", line + 1, text);
		else
			DrawScreenText( x0 + 10, y0 + 8 + line * 10, "%c", line + 56, text);
    }
    else
	{
		DrawScreenText( x0 + 10, y0 + 8 + line * 10, text);

		SetColor( WHITE);
		if (highlightnum > 0)
			DrawScreenText( x0 + 2 + highlightnum * 8, y0 + 8 + line * 10, "%c", text[highlightnum - 1]);
    }
	ShowMousePointer();
}

/*
   display and execute a menu
   */

BCINT DisplayMenuArray( BCINT x0, BCINT y0, char *menutitle, BCINT numitems, BCINT *okkeys, char *menustr[ MENUARRAYMAX], BCINT highlight[ MENUARRAYMAX])
{
	static BCINT mbX[] = { 0, 43, 81, 136, 177, 248, 344, 0 };
    int maxlen, line, oldline;
    Bool  ok;
    int k, PointerX, PointerY, buttons, oldbuttons, oldX, oldY;
    static char temp[ 10];
    static char temp2[ 10];
    int x1,y1;
	BCINT width,height;
    
    /* compute maxlen */
    if (menutitle)
		maxlen = (strlen(menutitle) * .85) - 2;
    else
		maxlen = 1;
    for (line = 0; line < numitems; line++)
		if ((signed)strlen( menustr[ line]) > maxlen)
			maxlen = strlen( menustr[ line]);
    
    /* display the menu */
    if (x0 < 0)
		x0 = (ScrMaxX - maxlen * 8 - (menutitle ? 30 : 0) - (okkeys ? 19 : 19)) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY - numitems * 10 - (menutitle ? 28 : 12)) / 2;
    if (x0 > ScrMaxX - maxlen * 8 - (menutitle ? 30 : 0) - (okkeys ? 19 : 19))
		x0 = ScrMaxX - maxlen * 8 - (menutitle ? 30 : 0) - (okkeys ? 19 : 19);

	HideMousePointer(); 
    DrawScreenBox3D( x0, y0, (menutitle ? 30 : 0) + x0 + maxlen * 8 + (okkeys ? 19 : 19), y0 + numitems * 10 + (menutitle ? 32 : 13));
    DrawScreenInset( x0 + 3, y0 + (menutitle ? 21 : 3) , (menutitle ? 30 : 0) + maxlen * 8 + (okkeys ? 19 : 19) - 6, numitems * 10 + (menutitle ? 8 : 7));
    if (menutitle)
	{
		SetColor(CYAN);
		DrawScreenBox( x0 + 3, y0 + 3, 30 + x0 + maxlen * 8 + (okkeys ? 19 : 19) - 3, y0 + 17);
		SetColor(BLACK);
		DrawScreenTextSmall(x0 + 10, y0 + 4, menutitle);
	}
    for (line = 0; line < numitems; line++)
	{
		SetColor( BLACK);
		if (menustr[line][0] == '-')
			DrawScreenBar( x0 + 7, y0 + (menutitle ? 29 : 10) + (line * 10), (menutitle ? 30 : 0) + (maxlen * 8) + 6);
		else
			DisplayMenuText( x0, y0 + (menutitle ? 19 : 0), line, menustr[ line], menutitle ? 0 : highlight[line], menutitle ? TRUE : FALSE);
    }
	ShowMousePointer();
    line = (MenuDefault>=0 && MenuDefault<numitems)? MenuDefault : 0;
    oldline = line;
	oldbuttons = 0;
	oldX = 0;
	oldY = 0;
    ok = FALSE;

	mbX[7] = ScrMaxX - maxlen * 8 - 19;

    while (!ok)
	{
		GetMouseCoords(&PointerX, &PointerY, &buttons);

		// [RH] allow selection of other top-level menus with the mouse
		// Yes, it's a big-big hack, but the GUI code in this program isn't very well structured
		// and I don't feel like rewriting it. Porting to Windows would probably be easier and
		// more effective.
		if (buttons == 1 && PointerY < 16 && y0 == 17 && NestMenus)
		{
			int slot = -1;

			if (PointerX < 12)			DEU_Beep();
			else if (PointerX < 42)		slot = 0;
			else if (PointerX < 80)		slot = 1;
			else if (PointerX < 135)	slot = 2;
			else if (PointerX < 176)	slot = 3;
			else if (PointerX < 247)	slot = 4;
			else if (PointerX < 344)	slot = 5;
			else if (PointerX < 440)	slot = 6;
			else if (PointerX < ScrMaxX - 60)	DEU_Beep();
			else						slot = 7;

			// Don't leave if the new menu is the same as the current one
			if (slot >= 0 && mbX[slot] != x0)
			{
				return -10-slot;
			}
		}

		/* right button = cancel */
		if (buttons == 2)
			return 0;

		/* release left button = accept selection */
		else if (buttons == 0 && oldbuttons == 1 && PointerY >= 16)
			ok = TRUE;

		/* select with mouse tracking */
//		else if (buttons == 1 && PointerX >= x0 && PointerX <= x0 + maxlen * 8 + 19)
		if(PointerX != oldX || PointerY != oldY)
			line = (PointerY - y0 - (menutitle ? 27 : 8)) / 10;

		if(line < 0)
			line = 0;
		else if(line >= numitems)
			line = numitems-1;//MenuDefault;

		oldbuttons = buttons;
		oldX = PointerX;
		oldY = PointerY;

		if (keypressed())
		{
			k = readkey();

			/* enter key = accept selection */
			if ((k >> 8) == KEY_ENTER)
				ok = TRUE;
			/* escape key = cancel */
			else if ((k >> 8) == KEY_ESC)
				return 0;
			else if (k >> 8 == KEY_RIGHT )
				return -1;
			else if (k >> 8 == KEY_LEFT )
				return -2;
			/* up arrow = select previous line */
			else if ((k & 0xFF00) == CURS_UP)
			{
				if (line > 0)
				{
					if (strncmp(menustr[line - 1], "-", 1) == 0)
						line--;
					line--;
				}
				else
					line = numitems - 1;
			}
			/* down arrow = select next line */
			else if ((k & 0xFF00) == CURS_DOWN)
			{
				if (line < numitems - 1)
				{
					if (strncmp(menustr[line + 1], "-", 1) == 0)
						line++;
					line++;
				}
				else
					line = 0;
			}
			/* home = select first line */
			else if ((k & 0xFF00) == HOME)
				line = 0;
			/* end = select last line */
			else if ((k & 0xFF00) == END)
				line = numitems - 1;
			/* pgup = select line - 5 */
			else if ((k & 0xFF00) == PAGEUP)
			{
				if (line >= 5)
					line -= 5;
				else
					line = 0;
			}
			/* pgdn = select line + 5 */
			else if ((k & 0xFF00) == PAGEDOWN)
			{
				if (line < numitems - 5)
					line += 5;
				else
					line = numitems - 1;
			}
			/* number or alphabetic key = enter selection */
/*			else if ((k & 0x00FF) >= '1' && (k & 0x00FF) <= '9' && ((k & 0x00FF) - '1') < numitems)
			{
				line = (k & 0x00FF) - '1';
				ok = TRUE;
			}
			else if ((k & 0x00FF) >= 'A' && (k & 0x00FF) <= 'Z' && ((k & 0x00FF) - 'A' + 9) < numitems)
			{
				line = (k & 0x00FF) - 'A' + 9;
				ok = TRUE;
			}
			else if ((k & 0x00FF) >= 'a' && (k & 0x00FF) <= 'z' && ((k & 0x00FF) - 'a' + 9) < numitems)
			{
				line = (k & 0x00FF) - 'a' + 9;
				ok = TRUE;
			}*/
			/* check the list of "hot keys" */
			else if (okkeys)
			{
				for (line = 0; line < numitems; line++)
					if (toupper( k & 0x00FF ) == okkeys[ line])
						break;
				if (line < numitems)
					ok = TRUE;
				else
				{
					line = oldline;
					DEU_Beep();
				}
			}
			/* other key */
			else
				DEU_Beep();
		}
		if (line != oldline)
		{
			HideMousePointer();

			SetColor(DARKGRAY);
			DrawScreenBox (x0 + 6, y0 + (menutitle ? 19 : 0) - 2 + 8 + oldline * 10, (menutitle ? 30 : 0) + x0 + maxlen * 8 + (okkeys ? 19 : 19) - 6, y0 + (menutitle ? 19 : 0) + 17 + oldline * 10);
			SetColor( BLACK);
			if (strncmp(menustr[oldline], "-", 1) == 0)
				DrawScreenBar( x0 + 7, y0 + (menutitle ? 29 : 10) + (oldline * 10), (menutitle ? 30 : 0) + (maxlen * 8) + 6);
			else
				DisplayMenuText( x0, y0 + (menutitle ? 19 : 0), oldline, menustr[ oldline], menutitle ? 0 : highlight[oldline], menutitle ? TRUE : FALSE);

			SetColor(BLUE);
			DrawScreenButtonOut(x0 + 6, y0 + (menutitle ? 19 : 0) - 2 + 8 + line * 10, (menutitle ? 30 : 0) + maxlen * 8 + (okkeys ? 19 : 19) - 12, 10);
			SetColor(WHITE);
			if (strncmp(menustr[line], "-", 1) == 0)
				DrawScreenBar( x0 + 7, y0 + (menutitle ? 29 : 10) + (line * 10), (menutitle ? 30 : 0) + (maxlen * 8) + 6);
			else
				DisplayMenuText( x0, y0 + (menutitle ? 19 : 0), line, menustr[ line], menutitle ? 1 : highlight[line], menutitle ? TRUE : FALSE);

			if (ShowSprite)
			{
				if (strcmp( menutitle, "Select Thing Type:") >= -1)
				{
					temp[ 0] = '\0';
					HideMousePointer();
					if (InfoShown)
					{
						DrawScreenBox3D( ScrMaxX - 120, ScrMaxY - 133, ScrMaxX, ScrMaxY - 13);
						strncpy( temp, menustr[ line], 6);
						strncpy( temp2, menustr[ line], 8);
						GetPicSize(temp, &width, &height);
						x1 = ((120 - width) / 2);
						y1 = ((120 - height) / 2);
						if (x1 < 0)
							x1 = 4;
						if (y1 < 0)
							y1 = 4;
						DisplayPic( ScrMaxX - 120 + x1, ScrMaxY - 133 + y1 , ScrMaxX - 4, ScrMaxY - 17, temp);
						DisplayPic( ScrMaxX - 120 + x1, ScrMaxY - 133 + y1 , ScrMaxX - 4, ScrMaxY - 17, temp2);
					}
					else
					{
						DrawScreenBox3D( ScrMaxX - 120, ScrMaxY - 120, ScrMaxX, ScrMaxY);
						strncpy( temp, menustr[ line], 6);
						strncpy( temp2, menustr[ line], 8);
						GetPicSize(temp, &width, &height);
						x1 = ((120 - width) / 2);
						y1 = ((120 - height) / 2);
						if (x1 < 0)
							x1 = 0;
						if (y1 < 0)
							y1 = 0;
						DisplayPic( ScrMaxX - 120 + x1, ScrMaxY - 120 + y1 , ScrMaxX - 4, ScrMaxY - 4, temp);
						DisplayPic( ScrMaxX - 120 + x1, ScrMaxY - 120 + y1 , ScrMaxX - 4, ScrMaxY - 4, temp2);
					}
					if (Debug)
					{
						SetColor(BLACK);
						DrawScreenBox(ScrMaxX - 80, ScrMaxY - 150, ScrMaxX, ScrMaxY -140);
						SetColor(ORANGE);
						DrawScreenText(ScrMaxX - 80, ScrMaxY - 150,"%d,%d", width, height);
					}
					ShowMousePointer();
				}
			}
			ShowMousePointer();

			oldline = line;
		}

		rest(1);
    }

    if (line >= 0 && line < numitems && menustr[line][0] != '-')
		return (line + 1);

	return 0;
}




/*
   display and execute a menu defined with variable arguments
   */

BCINT DisplayMenu( BCINT x0, BCINT y0, char *menutitle, ...)
{
    va_list args;
    BCINT num, val;
    char *menustr[ MENUARRAYMAX];
    BCINT dummy[ MENUARRAYMAX] = { 0 };
    
    /* put the va_args in the menustr table */
    num = 0;
    va_start( args, menutitle);
    while ((num < MENUARRAYMAX) && ((menustr[ num] = va_arg( args, char *)) != NULL))
		num++;
    va_end( args);
    
    /* display the menu */
    val = DisplayMenuArray( x0, y0, menutitle, num, NULL, menustr, dummy);
    if (val >= -2 || (val <= -10 && val >= -17))
    	return val;
    else
		return 0;
}



/* display and execute a dropdown menu (returns a key code) */

BCINT PullDownMenu( BCINT x0, BCINT y0, ...)
{
    va_list args;
    BCINT num;
    char *menustr[ MENUARRAYMAX];
    BCINT retnkeys[ MENUARRAYMAX];
    BCINT permkeys[ MENUARRAYMAX];
    BCINT highlight[ MENUARRAYMAX];

    /* put the va_args in the menustr table and the two strings */
    num = 0;
    va_start( args, y0);
    while ((num < MENUARRAYMAX) && ((menustr[ num] = va_arg( args, char *)) != NULL))
	{
		if (!(retnkeys[ num] = va_arg( args, int)))
			ProgError( "BUG: PullDownMenu() called with invalid arguments");
		if (!(permkeys[ num] = va_arg( args, int)))
			ProgError( "BUG: PullDownMenu() called with invalid arguments");
		if (!(highlight[ num] = va_arg( args, int)))
			ProgError( "BUG: PullDownMenu() called with invalid arguments");
		num++;
    }
    va_end( args);
    
    /* display the menu */
    num = DisplayMenuArray( x0, y0, NULL, num, permkeys, menustr, highlight);
    if (num == -1 || num == -2 || (num <= -10 && num >= -17))
    	return num;
    if (num >= 1)
		return retnkeys[ num - 1]; /* return a valid key */
    else
		return 0; /* return an invalid key */
}



/* display the integer input box */

BCINT InputInteger( BCINT x0, BCINT y0, BCINT *valp, BCINT minv, BCINT maxv)
{
    BCINT  k, val;
    Bool neg, ok, firstkey;
   	if (x0 < 0)
   		x0 = (ScrMaxX - 74) / 2;
   	if (y0 < 0)
   		y0 = (ScrMaxY) / 2;
    DrawScreenBoxHollow( x0, y0, x0 + 61, y0 + 13);
    neg = (*valp < 0);
    val = neg ? -(*valp) : *valp;
    firstkey = TRUE;
    for (;;) {
		ok = (neg ? -val : val) >= minv && (neg ? -val : val) <= maxv;
		SetColor( BLACK);
		DrawScreenBox( x0 + 1, y0 + 1, x0 + 60, y0 + 12);
		if (ok)
			SetColor(WHITE);
		else                                          
			SetColor(DARKGRAY);
		if (neg)
			DrawScreenText( x0 + 3, y0 + 3, "-%d", val);
		else
			DrawScreenText( x0 + 3, y0 + 3, "%d", val);
		k = readkey();
		if (firstkey && (k & 0x00FF) > ' ') {
			val = 0;
			neg = FALSE;
		}
		firstkey = FALSE;
		if (val < 3277 && (k & 0x00FF) >= '0' && (k & 0x00FF) <= '9')
			val = val * 10 + (k & 0x00FF) - '0';
		else if (val > 0 && (k & 0x00FF) == 0x0008)
			val = val / 10;
		else if (neg && (k & 0x00FF) == 0x0008)
			neg = FALSE;
		else if ((k & 0x00FF) == '-')
			neg = !neg;
		else if (ok && (k & 0x00FF) == 0x000D)
			break; 
		else if ((k & 0xFF00) == CURS_UP || (k & 0xFF00) == CURS_DOWN ||
				 (k & 0xFF00) == CURS_LEFT || (k & 0xFF00) == CURS_RIGHT ||
				 (k & 0x00FF) == 0x0009 || (k & 0xFF00) == 0x0F00)
			break; 
		else if ((k & 0x00FF) == 0x001B) {
			neg = FALSE;
			val = -16385; /*THIS WAS 0*/ 
			break;
		}
		else
			DEU_Beep();

		rest(1);
    }
    if (neg)
		*valp = -val;
    else
		*valp = val;
    return k;
}



/* ask for an integer value and check for minimum and maximum */

BCINT InputIntegerValue( BCINT x0, BCINT y0, BCINT minv, BCINT maxv, BCINT defv)
{
    BCINT  val, k;
    char prompt[ 80];
    
	HideMousePointer();
    sprintf( prompt, "Enter A Decimal Value Between %d And %d:", minv, maxv);
 	if (x0 < 0)
   		x0 = (ScrMaxX - 25 - 8 * strlen( prompt)) / 2;
   	if (y0 < 0)
   		y0 = (ScrMaxY - 55) / 2;
    DrawScreenBox3D( x0, y0, x0 + 25 + 8 * strlen( prompt), y0 + 55);
    SetColor( WHITE);
    DrawScreenText( x0 + 10, y0 + 8, prompt);
    val = defv;
	ShowMousePointer();

    while (((k = InputInteger( -1, -1, &val, minv, maxv)) & 0x00FF) != 0x000D && (k & 0x00FF) != 0x001B)
		DEU_Beep();

    return val;
}

BCINT InputIntegerValuePlain( BCINT x0, BCINT y0, BCINT minv, BCINT maxv, BCINT defv)
{
    BCINT  val, k;
    
    val = defv;
    while (((((k = InputInteger( x0, y0, &val, minv, maxv)) & 0x00FF) != 0x000D && (k & 0x00FF) != 0x001B)  && (k & 0x00FF) != 0x0009) && (k & 0xFF00) != 0x0F00)
		DEU_Beep();

    return val;
}

/* ask for a name in a given list */

void InputNameFromList(BCINT x0, BCINT y0, char *prompt, BCINT listsize, Tex **list, char *name)
{
    InputNameFromListWithFunc(x0, y0, prompt, listsize, list, 5, name, 0, 0, NULL);
}

/* ask for a name in a given list and call a function (for displaying objects, etc.)

   Arguments:
   x0, y0  : where to draw the box.
   prompt  : text to be displayed.
   listsize: number of elements in the list.
   list    : list of names (picture names, level names, etc.).
   listdisp: number of lines that should be displayed.
   name    : what we are editing...
   width   : \ width and height of an optional window where a picture
   height  : / can be displayed (used to display textures, sprites, etc.).
   hookfunc: function that should be called to display a picture.
   (x1, y1, x2, y2 = coordinates of the window in which the
   picture must be drawn, name = name of the picture).
   */
void InputNameFromListWithFunc(BCINT x0, BCINT y0, char *prompt,
	int listsize, Tex **list, BCINT listdisp, char *name,
	BCINT width, BCINT height,
	void (*hookfunc)(BCINT, BCINT, BCINT, BCINT, Tex*))
{
	int  k, n, t, l = 0;
	int  x1 = 0, y1 = 0, x2 = 0, y2 = 0;
	int  maxlen = 0;
	int  EditMode = OBJ_LINEDEFS;
    Bool ok, firstkey;
	char *types[5] = { "-TX", "-TEXTUREx", "-Flats", "-PNAMES", "-Sprites" };
	static int currtype = TYPE_TEXTURE;
	static Bool showall = TRUE;
	int fakesize;
	int PointerX, PointerY, buttons;
	BITMAP *buffer = GetBuffer();

	fakesize = 0;
	for(n = 0;n < listsize;++n)
	{
		if(list[n]->type & currtype || showall)
		++fakesize;
	}

	if(hookfunc)
	{
		for(n = 0;n < listsize;++n)
		{
			if(list[n]->type == TYPE_TX)
				types[0] = "TX";
			else if(list[n]->type == TYPE_TEXTURE)
				types[1] = "TEXTUREx";
			else if(list[n]->type == TYPE_FLAT)
				types[2] = "Flats";
			else if(list[n]->type == TYPE_PNAME)
				types[3] = "PNAMES";
			else if(list[n]->type == TYPE_SPRITE)
				types[4] = "Sprites";
		}
	}

    if (BrowseMap)
	{
		EditMode = 99;
		InitialScale = 9;
	}

    /* compute maxlen */
	if (!BrowseMap)
	{
		maxlen = 8;

		for (n = strlen(name) + 1; n <= maxlen; n++)
			name[n] = '\0';

		/* compute the minimum width of the dialog box */
		l = maxlen;
		if ((signed)strlen(prompt) > l + 12)
			l = strlen(prompt) - 12;
		l = l * 8 + 110;
		x1 = l + 3;
		y1 = 10 + 1;
		if (width > 0)
			l += 10 + width;
		if (height > 65)
			n = height + 20;
		else
			n = 85;
		if (x0 < 0)
			x0 = (ScrMaxX - l) / 2;
		if (y0 < 0)
			y0 = (ScrMaxY - n) / 2;
		x1 += x0;
		y1 += y0;
		if (x1 + width - 1 < ScrMaxX)
			x2 = x1 + width - 1;
		else
			x2 = ScrMaxX;
		if (y1 + height - 1 < ScrMaxY)
			y2 = y1 + height - 1;
		else
			y2 = ScrMaxY;

		HideMousePointer();
		/* draw the dialog box */
		DrawScreenBox3D( x0, y0, x0 + l, y0 + n);
		DrawScreenBoxHollow( x0 + 10, y0 + 28, x0 + 101, y0 + 41);
		DrawScreenText( x0 + 10, y0 + 8, prompt);

		if (width > 0)
		{
			DrawScreenBox( x1, y1, x2 + 1, y2 + 1);
			SetColor(DARKERGRAY);
			DrawScreenBox( x1 - 1, y1 - 1, x2, y2);
		}
		ShowMousePointer();
    }
    firstkey = TRUE;
	ok = TRUE;
    for (;;)
	{
		/* test if "name" is in the list */
		for(n = 0;n < listsize;++n)
		{
			if(strcmp(name, list[n]->name) == 0)
			{
				ok = TRUE;
				break;
			}
		}

		if(n == listsize)
		{
			for(n = 0; n < listsize; n++)
				if(((currtype & list[n]->type) || showall) &&
				   strcmp(name, list[n]->name) < 0)
					break;

			ok = FALSE;
		}

		if (n >= listsize)
			n = listsize - 1;

		/* Make sure we're not past the end of the list type */
		if(!showall)
		{
			t = n;
			while(t < listsize)
			{
				if(currtype & list[t]->type)
					break;
				++t;
			}

			if(t == listsize)
			{
				while(n >= 0)
				{
					if(currtype & list[n]->type)
						break;
					--n;
				}
				if(n < 0)
					n = 0;
			}
		}

		acquire_bitmap(buffer);

		/* display the "listdisp" next items in the list */
		if (!BrowseMap)
		{
			HideMousePointer();

			SetColor(DARKGRAY);
			DrawScreenBox( x0 + 110, y0 + 30, x0 + 110 + 8 * maxlen, y0 + 30 + 10 * listdisp);
			SetColor( BLACK);

			t = 0;
			l = 0;
			while(l < fakesize && l < listdisp && n+l+t < listsize)
			{
				if(showall || (currtype & list[n+l+t]->type))
				{
					DrawScreenText(x0 + 110, y0 + 30 + l * 10, list[n+l+t]->name);
					++l;
				}
				else
					++t;
			}

			l = strlen(name);
			DrawScreenBox(x0 + 11, y0 + 29, x0 + 100, y0 + 40);
			if (ok)
				SetColor(WHITE);
			else
				SetColor(DARKGRAY);
			DrawScreenText( x0 + 13, y0 + 31, name);

			ShowMousePointer();
		}
		/* call the function to display the picture, if any */
		if (BrowseMap)
		{
			if(ok)
			{
				strcpy(LevelName, name);
				ReadLevelData();
				HideMousePointer();
				Scale = (float) (1.0 / InitialScale); 
				CenterMapAroundCoords( (MapMinX + MapMaxX) / 2, (MapMinY + MapMaxY) / 2);
				ClearMapScreen(0,13);
				if (EditMode == 99)
					DrawMapQuick(); 
				else
					DrawMap(EditMode,0,0); 
				ShowMousePointer();
				ForgetLevelData(); 
			} 
		}
		if (hookfunc)
		{
			/* Draw the buttons */
			if(showall)
				DrawScreenButtonIn(x0+8, y0+128-18, 64, 16, "All");
			else
				DrawScreenButton(x0+8, y0+128-18, 64, 16, "All");

			for(l = 0;l < 5;++l)
			{
				if(types[l][0] != '-')
				{
					if((currtype & (1<<l)) && !showall)
						DrawScreenButtonIn(x0+8, y0+128+(l*18), 64, 16, types[l]);
					else
						DrawScreenButton(x0+8, y0+128+(l*18), 64, 16, types[l]);
				}
				else
					DrawScreenButtonGray(x0+8, y0+128+(l*18), 64, 16, &(types[l][1]));
			}

			/* Clear the area for the texture info */
			HideMousePointer();
			SetColor(DARKGRAY);
			DrawScreenBox(x0+8, y0+48, x0+96, y0+96);

			/* clear the window */
			SetColor(BLACK);
			DrawScreenBox( x1, y1, x2, y2);

			/* display the picture "name" */
			if(ok)
			{
				Texture *t;

				hookfunc(x1, y1, x2, y2, list[n]);

				t = GetTextureInfo(list[n]);
				if(t)
				{
					int sx = t->scalex ? t->scalex : 8;
					int sy = t->scaley ? t->scaley : 8;

					SetColor(BLACK);

					DrawScreenText(x0+8, y0+48, "%s", (list[n]->type == TYPE_TX ? "TX" :
					                                   (list[n]->type == TYPE_TEXTURE ? "TEXTUREx" :
													    (list[n]->type == TYPE_FLAT ? "Flat" :
													     (list[n]->type == TYPE_PNAME ? "PNAMES" :
													      (list[n]->type == TYPE_SPRITE ? "Sprite" : ""))))));
					DrawScreenText(-1, -1, "%4ix%i", t->width, t->height);
					if(sx != 8 || sy != 8)
						DrawScreenText(-1, -1, "Xscale: %.2f  Yscale: %.2f", sx/8.0, sy/8.0);
				}
			}
			ShowMousePointer();
		}

		release_bitmap(buffer);

		/* wait for keypress or mouse click */
waitkey:
		do {
			if(quitnow)
			{
				name[0] = '\0'; /* return an empty string */
				if(BrowseMap)
					SetMouseCoords(0,0);

				return;
			}

			GetMouseCoords(&PointerX, &PointerY, &buttons);

			rest(1);
		} while(!keypressed() && !buttons);

		/* If left mouse click */
		if(buttons & 1)
		{
			/* Clicked on the name list */
			if(PointerX >= x0+110 && PointerX < x1 &&
			   PointerY >= y0+30  && PointerY < y0+30+listdisp*10)
			{
				/* Find out which name was clicked */
				int offset = (PointerY - (y0+30)) / 10;

				/* Wait for button release */
				while(buttons & 1)
				{
					GetMouseCoords(&PointerX, &PointerY, &buttons);
					rest(1);
				}

				/* If the mouse clicked on a different name and is still over it */
				if(offset && (PointerY - (y0+30)) / 10 == offset)
				{
					/* If it's not the full list, we need to search it manually */
					if(!showall)
					{
						t = 0;

						/* Increment 't' for every good name hit */
						while(t < offset && n+t < listsize)
						{
							if(list[n+t]->type & currtype)
								++t;
							else
								++n;
						}
						/* update 'n' */
						n += t;

						/* Find the next good name */
						while(n < listsize)
						{
							if(list[n]->type & currtype)
								break;
							++n;
						}

						/* Copy it */
						if(n < listsize)
							strcpy(name, list[n]->name);
					}
					else
					{
						/* It's the full list, just jump forward */
						if((n += offset) >= listsize)
							n = listsize-1;

						strcpy(name, list[n]->name);
					}
				}
			}

			/* Clicked on the image viewer. Left click = select */
			else if(PointerX >= x1 && PointerX <= x2 && PointerY >= y1 && PointerY <= y2)
			{
				while(buttons & 1)
				{
					GetMouseCoords(&PointerX, &PointerY, &buttons);
					rest(1);
				}

				if(PointerX >= x1 && PointerX <= x2 && PointerY >= y1 && PointerY <= y2)
					return;
			}

			/* Clicked the texture type buttons */
			else
			{
				/* Too far to the sides */
				if(PointerX < x0+8 && PointerX > x0+8+64)
				{
					while(buttons & 1)
					{
						GetMouseCoords(&PointerX, &PointerY, &buttons);
						rest(1);
					}

					goto waitkey;
				}

				/* Check each button */
				if(PointerY > y0+128-18 && PointerY < y0+128-2)
				{
					/* Press the button in and wait for click release */
					DrawScreenButtonIn(x0+8, y0+128-18, 64, 16, "All");

					while(buttons & 1)
					{
						GetMouseCoords(&PointerX, &PointerY, &buttons);
						rest(1);
					}

					/* If it's still on the button, toggle it */
					if(PointerX > x0+8 && PointerX < x0+8+64 &&
					   PointerY > y0+128-18 && PointerY < y0+128-2)
					{
						showall = !showall;

						/* Get the number of textures for display */
						fakesize = 0;
						for(t = 0;t < listsize;++t)
						{
							if(list[t]->type & currtype || showall)
								++fakesize;
						}
					}

					continue;
				}

				for(l = 0;l < 5;++l)
				{
					/* This button's disabled, skip it */
					if(types[l][0] == '-')
						continue;

					/* Got it */
					if(PointerY > y0+128+(l*18) && PointerY < y0+128+(l*18)+16)
					{
						/* Press the button in and wait for click release */
						DrawScreenButtonIn(x0+8, y0+128+(l*18), 64, 16, types[l]);

						while(buttons & 1)
						{
							GetMouseCoords(&PointerX, &PointerY, &buttons);
							rest(1);
						}

						/* If it's still on the button, toggle it */
						if(PointerX > x0+8 && PointerX < x0+8+64 &&
						   PointerY > y0+128+(l*18) && PointerY < y0+128+(l*18)+16)
						{
							if(showall)
							{
								showall = FALSE;
								currtype = 1<<l;
							}
							else
								currtype ^= 1 << l;

							/* Get the number of textures for display */
							fakesize = 0;
							for(t = 0;t < listsize;++t)
							{
								if(list[t]->type & currtype)
									++fakesize;
							}
						}

						break;
					}
				}

				/* Wait for mouse button release */
				while(buttons & 1)
				{
					GetMouseCoords(&PointerX, &PointerY, &buttons);
					rest(1);
				}

				/* No button clicked */
				if(l == 5)
					goto waitkey;

				continue;
			}
		}
		else if(buttons & 2)
		{
			if(PointerX >= x1 && PointerX <= x2 && PointerY >= y1 && PointerY <= y2)
			{
				while(buttons & 2)
				{
					GetMouseCoords(&PointerX, &PointerY, &buttons);
					rest(1);
				}

				if(PointerX >= x1 && PointerX <= x2 && PointerY >= y1 && PointerY <= y2)
				{
					if(BrowseMap)
						SetMouseCoords(0, 0);

					name[0] = '\0';
					return;
				}
			}

			while(buttons & 2)
			{
				GetMouseCoords(&PointerX, &PointerY, &buttons);
				rest(1);
			}
		}

		if(!keypressed())
			continue;
		k = readkey();

		l = strlen(name);
		if(!BrowseMap)
		{
			if(firstkey && (k & 0x00FF) > ' ')
			{
				for(l = 0; l <= maxlen; l++)
					name[ l] = '\0';
				l = 0;
				firstkey = FALSE;
			}
		}
		else
		{
			if((k & 0x00FF) == 's' || (k & 0x00FF) == 'S')
			{
				EditMode = OBJ_SECTORS;
				continue;
			}
			else if((k & 0x00FF) == 't' || (k & 0x00FF) == 'T')
			{
				EditMode = OBJ_THINGS;
				continue;
			}
			else if((k & 0x00FF) == 'v' || (k & 0x00FF) == 'V')
			{
				EditMode = OBJ_VERTEXES;
				continue;
			}
			else if((k & 0x00FF) == 'l' || (k & 0x00FF) == 'L')
			{
				EditMode = OBJ_LINEDEFS;
				continue;
			}
			else if((k & 0x00FF) == 'q' || (k & 0x00FF) == 'Q')
			{
				EditMode = 99;
				continue;
			}
			else if((((k & 0x00FF) == '+' || (k & 0x00FF) == '=') && InitialScale > 1.0))
			{ 
				InitialScale = InitialScale - 1.0;
				continue;
			}
			else if((((k & 0x00FF) == '-' || (k & 0x00FF) == '_') && InitialScale < 16.0))
			{ 
				InitialScale = InitialScale + 1.0;
				continue;
			}
		}
		if(l < maxlen && (k & 0x00FF) >= 'a' && (k & 0x00FF) <= 'z')
		{
			name[l] = (k & 0x00FF) + 'A' - 'a';
			name[l + 1] = '\0';
		}
		else if(l < maxlen && (k & 0x00FF) > ' ')
		{
			name[l] = k & 0x00FF;
			name[l + 1] = '\0';
		}
		else if(l > 0 && (k >> 8) == KEY_BACKSPACE)
			name[l - 1] = '\0';
		else if(n < listsize - 1 && (k & 0xFF00) == CURS_DOWN)
		{
			while(++n < listsize)
				if(showall || (list[n]->type & currtype))
					break;

			if(n < listsize)
				strcpy(name, list[n]->name);

			if (BrowseMap)
				InitialScale = 9;
		}
		else if(n > 0 && (k & 0xFF00) == CURS_UP)
		{
			while(--n >= 0)
				if(showall || (list[n]->type & currtype))
					break;

			if(n >= 0)
				strcpy(name, list[n]->name);

			if (BrowseMap)
				InitialScale = 9;
		}
		else if(n < listsize - listdisp && (k & 0xFF00) == PAGEDOWN)
		{
			while((n += listdisp) < listsize)
				if(showall || (list[n]->type & currtype))
					break;

			if(n < listsize)
				strcpy(name, list[n]->name);

			if (BrowseMap)
				InitialScale = 9;
		}
		else if(n > 0 && (k & 0xFF00) == PAGEUP)
		{
			while((n -= listdisp) >= 0)
				if(showall || (list[n]->type & currtype))
					break;

			if (n >= 0)
				strcpy(name, list[n]->name);
			else
				strcpy(name, list[0]->name);

			if (BrowseMap)
				InitialScale = 9;
		}
		else if((k & 0xFF00) == END)
		{
			n = listsize;
			while(--n >= 0)
				if(showall || (list[n]->type & currtype))
					break;

			if (n >= 0)
				strcpy(name, list[n]->name);

			if (BrowseMap)
				InitialScale = 9;
		}
		else if((k & 0xFF00) == HOME)
		{
			n = -1;
			while(++n < listsize)
				if(showall || (list[n]->type & currtype))
					break;

			if (n < listsize)
				strcpy(name, list[n]->name);

			if (BrowseMap)
				InitialScale = 9;
		}
		else if((k & 0x00FF) == 0x0009)
		{
			strcpy(name, list[n]->name);
			if (BrowseMap)
				InitialScale = 9;
		}
		else if((k & 0x00FF) == 0x000D)
		{
			if (BrowseMap)
				SetMouseCoords(0,0);

			return;	 /* return "name" */
		}
		else if((k >> 8) == KEY_ESC)
		{
			name[0] = '\0'; /* return an empty string */
			if(BrowseMap)
				SetMouseCoords(0,0);

			break;
		}
		else
			if (!BrowseMap) 
				DEU_Beep();
    }
}

void InputCommand(BCINT x0, BCINT y0, char *prompt, BCINT maxlen, char *command)
{
    BCINT   k, l, boxlen;
    Bool  ok, firstkey;
    
    for (l = strlen(command) + 1; l <= maxlen; l++)
		command[ l] = '\0';
/*	if (maxlen > 20)
		boxlen = 20;
	else
		boxlen = maxlen;*/
    boxlen = maxlen;
    /* compute the width of the dialog box */
    if ((signed)strlen( prompt) > boxlen)
		l = strlen( prompt);
    else
		l = boxlen;
    if (x0 < 0)
		x0 = (ScrMaxX - 26 - 8 * l) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY - 50) / 2;
    /* draw the dialog box */
    DrawScreenBoxHollow( x0, y0, x0 + 5 + 8 * boxlen, y0 + 13);
    firstkey = TRUE;
    for (;;)
	{
		ok = TRUE;
		l = strlen( command);
		SetColor( BLACK);
		DrawScreenBox( x0 + 1, y0 + 1, x0 + 4 + 8 * boxlen, y0 + 12);
		SetColor (ok ? WHITE : DARKGRAY);
		if (l > boxlen)
		{
			DrawScreenText( x0 + 1, y0 + 3, "<");
			DrawScreenText( x0 + 3, y0 + 3, "<%s", command + (l - boxlen + 1));
		}
		else
		{
			DrawScreenText( x0 + 3, y0 + 3, command);
		}

		do {
			if(quitnow)
			{
				command[0] = '\0'; /* return an empty string */
				return;
			}

			rest(1);
		} while(!keypressed());

		k = readkey();
		if (firstkey && (k & 0x00FF) > ' ')
		{
			for (l = 0; l <= maxlen; l++)
				command[ l] = '\0';
			l = 0;
		}
		firstkey = FALSE;
		if (l < maxlen && (k & 0x00FF) >= 'a' && (k & 0x00FF) <= 'z')
		{
			command[ l] = k & 0x00FF;
			command[ l + 1] = '\0';
		}
		else if (l < maxlen && (k & 0x00FF) >= ' ')
		{
			command[ l] = k & 0x00FF;
			command[ l + 1] = '\0';
		}
		else if (l > 0 && (k & 0x00FF) == 0x0008)
		{
			command[ l - 1] = '\0';
		}
		else if (ok && (k & 0x00FF) == 0x000D)
		{
			break; /* return "filename" */
		}
		else if ((k & 0x00FF) == 0x001B)
		{
			command[ 0] = '\0'; /* return an empty string */
			break;
		}
		else
		{
			DEU_Beep();
		}
    }
}



/* ask for a filename */

void InputFileName( BCINT x0, BCINT y0, char *prompt, BCINT maxlen, char *filename)
{
    BCINT   k, l, boxlen;
    Bool  firstkey;
    char *p;
    
	HideMousePointer();
    for (l = strlen(filename) + 1; l <= maxlen; l++)
		filename[ l] = '\0';
    /* compute the width of the input box */
    if (maxlen > 20)
		boxlen = 20;
    else
		boxlen = maxlen;
    /* compute the width of the dialog box */
    if ((signed)strlen( prompt) > boxlen)
		l = strlen( prompt);
    else
		l = boxlen;
    if (x0 < 0)
		x0 = (ScrMaxX - 26 - 8 * l) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY - 50) / 2;
    /* draw the dialog box */
    DrawScreenBox3D( x0, y0, x0 + 26 + 8 * l, y0 + 50);
    DrawScreenBoxHollow( x0 + 10, y0 + 28, x0 + 15 + 8 * boxlen, y0 + 41);
    DrawScreenText( x0 + 10, y0 + 8, prompt);
	ShowMousePointer();

    firstkey = TRUE;
    for (;;) {
		/* check that "filename" looks like a valid file name */
		if (filename[ 1] == ':')
			p = filename + 2;
		else
			p = filename;

		l = strlen( filename);
		SetColor( BLACK);
		DrawScreenBox( x0 + 11, y0 + 29, x0 + 14 + 8 * boxlen, y0 + 40);
		SetColor( WHITE);
		if (l > boxlen) {
			DrawScreenText( x0 + 11, y0 + 31, "<");
			DrawScreenText( x0 + 13, y0 + 31, "<%s", filename + (l - boxlen + 1));
		}
		else
			DrawScreenText( x0 + 13, y0 + 31, filename);

		do {
			if(quitnow)
			{
				filename[0] = '\0'; /* return an empty string */
				return;
			}

			rest(1);
		} while(!keypressed());

		k = readkey();
		if (firstkey && (k & 0x00FF) >= ' ') {
			for (l = 0; l <= maxlen; l++)
				filename[ l] = '\0';
			l = 0;
		}
		firstkey = FALSE;
		/*if (l < maxlen && (k & 0x00FF) >= 'a' && (k & 0x00FF) <= 'z') {
			filename[ l] = (k & 0x00FF) + 'A' - 'a';
			filename[ l + 1] = '\0';
		}
		else*/ if (l < maxlen && (k & 0x00FF) >= ' ') {
			filename[ l] = k & 0x00FF;
			filename[ l + 1] = '\0';
		}
		else if (l > 0 && (k & 0x00FF) == 0x0008)
			filename[ l - 1] = '\0';
		else if ((k & 0x00FF) == 0x000D)
			break; /* return "filename" */
		else if ((k & 0x00FF) == 0x001B) {
			filename[ 0] = '\0'; /* return an empty string */
			break;
		}
		else
			DEU_Beep();
    }
}



/* ask for confirmation (prompt2 may be NULL) */

Bool Confirm( BCINT x0, BCINT y0, char *prompt1, char *prompt2)
{
    BCINT k;
    BCINT maxlen = 46;
    
	HideMousePointer();
    if ((signed)strlen( prompt1) > maxlen)
		maxlen = strlen( prompt1);
    if (prompt2 != NULL && (signed)strlen( prompt2) > maxlen)
		maxlen = strlen( prompt2);
    if (x0 < 0)
		x0 = (ScrMaxX - 22 - 8 * maxlen) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY - (prompt2 ? 53 : 43)) / 2;
    DrawScreenBox3D( x0, y0, x0 + 22 + 8 * maxlen, y0 + (prompt2 ? 53 : 43));
    SetColor( WHITE);
    DrawScreenText( x0 + 10, y0 + 8, prompt1);
    if (prompt2 != NULL)
		DrawScreenText( x0 + 10, y0 + 18, prompt2);
    SetColor( YELLOW);
    DrawScreenText( x0 + 10, y0 + (prompt2 ? 38 : 28), "Press Y To Confirm, Or Any Other Key To Cancel");
	ShowMousePointer();

	do {
		if(quitnow)
			return FALSE;

		rest(1);
	} while(!keypressed());
    k = readkey();

    return ((k & 0x00FF) == 'Y' || (k & 0x00FF) == 'y');
}



/* display a notification and wait for a key (prompt2 may be NULL) */

void Notify( BCINT x0, BCINT y0, char *prompt1, char *prompt2)
{
    BCINT maxlen = 30;
    
	HideMousePointer();
    if ((signed)strlen( prompt1) > maxlen)
		maxlen = strlen( prompt1);
    if (prompt2 != NULL && (signed)strlen( prompt2) > maxlen)
		maxlen = strlen( prompt2);
    if (x0 < 0)
		x0 = (ScrMaxX - 22 - 8 * maxlen) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY - (prompt2 ? 73 : 63)) / 2;
    DrawScreenBox3D( x0, y0, x0 + 22 + 8 * maxlen, y0 + (prompt2 ? 91 : 81));
    DrawScreenInset( x0 + 3, y0 + 21, 22 + (8 * maxlen) - 6, (prompt2 ? 73 : 63) - 6);
	SetColor(ORANGE);
    DrawScreenBox( x0 + 3, y0 + 3, x0 + 22 + (8 * maxlen) - 3, y0 + 17);
	SetColor(BLACK);
	DrawScreenTextSmall(x0 + 10, y0 + 4, "Message:");
    SetColor( WHITE);
    DrawScreenText( x0 + 10, y0 + 30, prompt1);
    if (prompt2 != NULL)
		DrawScreenText( x0 + 10, y0 + 40, prompt2);
    SetColor( YELLOW);
    DrawScreenText( x0 + 10, y0 + (prompt2 ? 75 : 65), "Press Any Key To Continue...");
	ShowMousePointer();

	do {
		if(quitnow)
			return;

		rest(1);
	} while(!keypressed());
    readkey();
}



/* clear the screen and display a message */

void DisplayMessage( BCINT x0, BCINT y0, char *msg, ...)
{
    char prompt[ 120];
    va_list args;
    
    va_start( args, msg);
    vsprintf( prompt, msg, args);
    va_end( args);
    
	HideMousePointer();
    clear_bitmap(GetBuffer());
    if (x0 < 0)
		x0 = (ScrMaxX - 40 - 8 * strlen( prompt)) / 2;
    if (y0 < 0)
		y0 = (ScrMaxY - 40) / 2;
    DrawScreenBox3D( x0, y0, x0 + 40 + 8 * strlen( prompt), y0 + 40);
    DrawScreenText( x0 + 20, y0 + 17, prompt);
	ShowMousePointer();
}

/* let's make the user angry... */

void NotImplemented()
{
	HideMousePointer();
    DrawScreenBox3D( 140, 220, 500, 260);
    SetColor( RED);
    DrawScreenText( 150, 230, "This Function Is Not Implemented... Yet!");
    SetColor( YELLOW);
    DrawScreenText( 150, 245, "Press Any Key To Return To The Editor.");
	ShowMousePointer();
    DEU_Beep();

	do {
		if(quitnow)
			return;

		rest(1);
	} while(!keypressed());
    readkey();
}

void XAlignmentMenu(SelPtr *list)
{
	BCINT	dummy[ MENUARRAYMAX];
	BCINT	val;
	int x0, y0;
	char	*menustr[ MENUARRAYMAX] =
	{
		"Align X With Texture Checking",
		"Align X With Texture Checking And Initial Offset",
		"Align X Without Texture Checking",
		"Align X Without Texture Checking And Initial Offset"
	};

	x0 = (ScrMaxX - 29 - 8 * 44) / 2;
	y0 = (ScrMaxY - 24 - 10 * 5) / 2;
	val = DisplayMenuArray( -1, -1, "X Texture Alignment", 4, NULL, menustr, dummy);

	if(val >= 1 && val <= 4) {
		Backup (BLINEDEFS | BSIDEDEFS);
		strcpy (BackupText, "Align Textures on X");
		val--;
		AlignTexturesX(list, 
					   (val & 2) ? TRUE : FALSE, 	
					   (val & 1) ? TRUE : FALSE);	
	}
}


void YAlignmentMenu(SelPtr *list)
{
	BCINT	dummy[ MENUARRAYMAX];
	BCINT	val;
	int x0, y0;
	char	*menustr[ MENUARRAYMAX] =
	{
		"Align Y With Texture Checking",
		"Align Y With Texture Checking And Initial Offset",
		"Align Y Without Texture Checking",
		"Align Y Without Texture Checking And Initial Offset"
	};

	x0 = (ScrMaxX - 29 - 8 * 44) / 2;
	y0 = (ScrMaxY - 24 - 10 * 5) / 2;
	val = DisplayMenuArray( -1, -1, "Y Texture Alignment", 4, NULL, menustr, dummy);

	if(val >= 1 && val <= 4) {
		Backup (BLINEDEFS | BSIDEDEFS);
		strcpy (BackupText, "Align Textures on Y");
		val--;
		AlignTexturesY(list, 
					   (val & 2) ? TRUE : FALSE, 	
					   (val & 1) ? TRUE : FALSE);	
	}
}

/* end of file */
