/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.

   TEXTURES.C - Textures in 256 colors.
   */

/* the includes */
#include <assert.h>
#include "deu.h"
#include "struct_h.h"

typedef struct _resource {
	char *name;
	void *data;
	char *ns;
	struct _resource *next;
} resource;

resource *ResourceList;

Bool ForgetAfter = FALSE;
/* whether DisplayPic should de-allocate the picture resource
   after displaying it */

BITMAP *CreatePatch(char *name, Bool forWall)
{
	void *p_resource = NULL;
	PColumn *c, **columns;
	PHeader *h;
	BITMAP *bmp;
	int x, y;
	int width, height;

	if(forWall)
	{
		p_resource = GetResourceInNamespace(name, "TX");
		if(!p_resource)
		{
			void *pnames = GetResource("PNAMES");
			if(pnames)
			{
				int i;
				for(i = 0;i < *(long*)pnames;++i)
				{
					if(strncmp(name, (pnames+4) + i*8, 8) == 0)
						break;
				}

				if(i == *(long*)pnames)
				{
					p_resource = GetResourceInNamespace(name, "S");
					if(!p_resource)
						return NULL;
				}
			}
		}
	}

	if(!p_resource)
	{
		p_resource = GetResourceInNamespace(name, "P");
		if(!p_resource)
			p_resource = GetResourceInNamespace(name, "S");
		if(!p_resource)
			p_resource = GetResource(name);
		if(!p_resource)
			return NULL;
	}

	h = (PHeader *)p_resource;
	columns = (PColumn **)(p_resource + sizeof(PHeader));

	width = h->width;
	height = 0;

	for (x = 0; x < width; x = x + 1)
	{
		int top = -1;	// [RH] Tall patch support
		c = (PColumn *)((long)columns[x] + p_resource);
		while(c->yofs != 255)
		{
			if (c->yofs <= top)
				top += c->yofs;
			else
				top = c->yofs;

			if (c->npixels + top > height)
				height = c->npixels + top;
			c = (PColumn *)((long)c + (long)(c->npixels) + 4L);
		}
	}

	bmp = NULL;
	if(width > 0 && height > 0)
		bmp = create_bitmap(width, height);
	if(!bmp)
	{
		if(Caching == FALSE)
			ForgetResource(name);
		return NULL;
	}
	clear_to_color(bmp, bitmap_mask_color(bmp));

	for(x = 0; x < width; ++x)
	{
		int top = -1;
		c = (PColumn *)((long)columns[x] + p_resource);
		while(c->yofs != 255)
		{
			BYTE *pixels = (BYTE *)((long)c + (long)sizeof(PColumn) + 1L);
			// [RH] Tall patch support
			if (c->yofs <= top)
				top += c->yofs;
			else
				top = c->yofs;

			for(y = 0; y < c->npixels; ++y)
				putpixel(bmp, x, y + top, palette_color[*pixels++]);
			c = (PColumn *)((long)c + (long)(c->npixels) + 4L);
		}
	}

	if(Caching == FALSE)
		ForgetResource(name);

	return bmp;
}

BITMAP *CreateTexture(char *name)
{
	int n;
	Texture  *tinfo;
	PatchDesc *pinfo;
	SList p;
	char picname[9];
	long patchname;
	void *Pnames;
	BITMAP *bmp;

	tinfo = NULL;
	for(p = Texture_sections; p && !tinfo; p = p->next)
	{
		void *tblock = GetResource(p->string);
		long *offsets = tblock;

		if(tblock)
		{
			int i;
			for(i = 1; i <= offsets[0]; i++)
			{
				Texture *t = (Texture *)((long)offsets[i] + (long)tblock);
				if(!strnicmp(t->name, name, 8)) /*jff don't care about case*/
				{
					tinfo = t;
					break;
				}
			}
		}
	}

	if(!tinfo)
		return NULL;

	Pnames = GetResource("PNAMES");
	if(!Pnames)
		return NULL;

	/* read the info for this texture */
	bmp = create_bitmap(tinfo->width, tinfo->height);
	if(!bmp)
		return NULL;
	clear_to_color(bmp, bitmap_mask_color(bmp));

	/* display the texture */
	pinfo = (PatchDesc *)((long)tinfo + sizeof(Texture));
	for (n = 0; n < tinfo->nPatches; n++, pinfo++)
	{
		BITMAP *p;

		/* OK, now look up the pic's name in Pnames. */
		patchname = (long)Pnames + 4L + ((long)(pinfo->PatchNumber) * 8);
		strncpy(picname, (char *)patchname, 8);
		picname[ 8] = '\0';

		p = CreatePatch(picname, FALSE);
		if(p)
		{
			masked_blit(p, bmp, 0, 0, pinfo->xOffset, pinfo->yOffset, p->w, p->h);
			destroy_bitmap(p);
		}
	}

	if(Caching == FALSE)
		ForgetResource(name);

	return bmp;
}

BITMAP *CreateFlat(char *name)
{
	MDirPtr dir = NULL;
	SList l;
	BITMAP *bmp;
	int dim, x, y;

    for(l = Ftexture_sections; l; l = l->next)
	{
		char start[9], end[9];

		strcpy(start, l->string);
		strcpy(end, l->string);
		strcat(start, "_START");
		strcat(end, "_END");

		dir = MasterDir;
		while((dir = FindMasterDir(dir, start)) != NULL)
		{
			while((dir = dir->next) != NULL)
			{
				if(strncmp(dir->dir.name, name, 8) == 0 || strncmp(dir->dir.name, end, 8) == 0)
					break;
			}
			if(strncmp(dir->dir.name, name, 8) == 0)
				goto found;
		}
	}

found:
	if (dir == NULL || dir->dir.size < 4096)
		return NULL;

	// High-res flats are determined by their size and can be
	// 128x128, 256x256, or 512x512. When rendered, they always
	// get scaled down to cover 64x64 world units.
	if(dir->dir.size < 128*128)
		dim = 64;
	else if(dir->dir.size < 256*256)
		dim = 128;
	else if(dir->dir.size < 512*512)
		dim = 256;
	else
		dim = 512;

	BasicWadSeek(dir->wadfile, dir->dir.start);

	bmp = create_bitmap(dim, dim);
	if(bmp)
	{
		unsigned char line[512*512];
		BasicWadRead(dir->wadfile, line, dim*dim);

		for(y = 0;y < dim;++y)
		{
			for(x = 0;x < dim;++x)
				putpixel(bmp, x, y, palette_color[line[y*dim + x]]);
		}
	}

	return bmp;
}


BITMAP *LoadTexture(Tex *tex)
{
	if(!tex)
		return NULL;

	if(tex->bmp)
	{
		if(bitmap_color_depth(tex->bmp) == bitmap_color_depth(screen))
			return tex->bmp;

		destroy_bitmap(tex->bmp);
		tex->bmp = NULL;
	}

	if(tex->type == TYPE_TX || tex->type == TYPE_PNAME || tex->type == TYPE_SPRITE)
		tex->bmp = CreatePatch(tex->name, TRUE);
	else if(tex->type == TYPE_FLAT)
		tex->bmp = CreateFlat(tex->name);
	else if(tex->type == TYPE_TEXTURE)
		tex->bmp = CreateTexture(tex->name);

	return tex->bmp;
}


BITMAP *LoadPatch(char *picname)
{
	return CreatePatch(picname, FALSE);
}

void GetPicSize(char *picname, BCINT *width, BCINT *height)
{	
	BITMAP *bmp;

	if (picname[0] == ':')
	{
		if(width)  *width = 32;
		if(height) *height = 32;
		return;
	}

	bmp = LoadPatch(picname);
	if(bmp)
	{
		if(width)  *width = (BCINT)bmp->w;
		if(height) *height = (BCINT)bmp->h;
		destroy_bitmap(bmp);
	}
	else
	{
		if(width)  *width = 0;
		if(height) *height = 0;
	}
}

/*
   display a picture "picname" at coords x0, y0 and not beyond x1, y1
   */


/* SO 2/4/95:
   DisplayPic now uses GetResource to get the sprite/patch to display.
   It checks the variable ForgetAfter to see whether it should
   call ForgetResource after displaying; this should be set by
   the caller.
   ChooseSprite sets this to TRUE; DisplayWallTexture sets it to FALSE.
   DisplayWallTexture then takes the responsibility for calling
   ForgetResource for all the patches in the texture being displayed.
   This enables us to buffer all the patches in case they are used
   more than once in a texture (e.g. LITE5 in Doom (extreme case!)) */
void DisplayPic(BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *picname)
{
	BITMAP *bmp;

	if(keypressed())
	    return;

	if (picname[0] == ':')
	{
		DrawIcon1 (x0, y0, picname+1);
		return;
	}

	bmp = LoadPatch(picname);
	if(!bmp)
		return;

	HideMousePointer();
	draw_sprite(GetBuffer(), bmp, x0, y0);
	ShowMousePointer();

	destroy_bitmap(bmp);
	(void)x1;
	(void)y1;
}

// [RH] Display a texture with fractional scaling
void DisplayPicFrac( BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *picname, long xstep, long ystep)
{	
	BITMAP *bmp;

	if(keypressed() && Caching == FALSE)
		return; 

	if((xstep == 65536 && ystep == 65536) || xstep == 0 || ystep == 0)
	{
		DisplayPic(x0, y0, x1, y1, picname);
		return;
	}

	if(picname[0] == ':')
	{
		DrawIconFrac (x0, y0, picname+1, xstep, ystep);
		return;
	}

	bmp = LoadPatch(picname);
	if(!bmp)
		return;

	HideMousePointer();
	stretch_sprite(GetBuffer(), bmp, x0, y0, bmp->w * 65536 / xstep, bmp->h * 65536 / ystep);
	ShowMousePointer();

	destroy_bitmap(bmp);
}

// [RH] Draw the sprite in the given region, picking a scale in order to keep it inside the box.
void DisplayPicFit( BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *picname)
{	
	long xmax, ymax, xstep, ystep, step;
	int width, height;

	if(keypressed())
	    return;

	if (picname[0] == ':')
	{
		width = height = 32;
	}
	else
	{
		BITMAP *bmp = LoadPatch(picname);
		if(!bmp)
			return;

		width = bmp->w;
		height = bmp->h;
		destroy_bitmap(bmp);
	}

	xmax = x1 - x0 + 1;
	ymax = y1 - y0;

	xstep = width * 65536 / xmax;
	ystep = height * 65536 / ymax;
	step = MAX(65536, MAX(xstep, ystep));

	DisplayPicFrac(x0+(xmax-width*65536/step)/2, ((y0*step)>>16) + ((ymax*step-65536*height)>>16)/2, x1, y1, picname, step, step);
}

void DisplaySprite( BCINT x0, BCINT y0, BCINT x1, BCINT y1, Tex *tex)
{
	DisplayPic(x0, y0, x1, y1, tex->name);
}

void *GetResource(char *name)
{
	void *t = NULL;
	resource *p, *new = (resource *)NULL;
	MDirPtr r;

	for(p = ResourceList; p; p = p->next)
		if(!strcmp(name, p->name) && !p->ns)
			return p->data;

//	r = FindMasterDir(MasterDir, name);
	r = MasterDir;
	while(r)
	{
		// If we hit a marker, skip to the end of the name section
		while(isamark(r->dir.name))
		{
			char m[9];
			char *p;

			strcpy(m, r->dir.name);
			p = strchr(m, '_');
			if(p)
			{
				strcpy(p+1, "END");

				while((r = r->next) != NULL)
					if(strncmp(r->dir.name, m, 8) == 0)
						break;

				if(r == NULL || (r = r->next) == NULL)
					return NULL;
			}
		}

		if(strncmp(r->dir.name, name, 8) == 0)
			break;

		r = r->next;
	}

	if(r && (t = malloc(r->dir.size)))
	{
		BasicWadSeek(r->wadfile, r->dir.start);
		BasicWadRead(r->wadfile, t, r->dir.size);

		new = (resource *)myalloc(sizeof(resource));

		new->data = t;
		new->name = strdup(name);
		new->next = ResourceList;
		ResourceList = new;
		return t;
	}

	return NULL;
}

void *GetResourceInNamespace(char *name, char *ns)
{
	void *t = NULL;
	resource *p, *new = (resource *)NULL;
	MDirPtr r;
	char start[9], end[9];

	if(!ns)
		return GetResource(name);

	for(p = ResourceList; p; p = p->next)
		if(!strcmp(name, p->name) && p->ns && !strcmp(ns, p->ns))
			return p->data;

	strcpy(start, ns);
	strcpy(end, ns);
	strcat(start, "_START");
	strcat(end, "_END");

	/* Get to the start of the namespace */
	r = FindMasterDir(MasterDir, start);
	if(!r) return NULL;

	/* search the wad directory.. */
	while((r = r->next) != NULL)
	{
		if(strncmp(r->dir.name, name, 8) == 0)
			break;

		/* if we reach the end, look for another start to this namespace */
		if(strncmp(r->dir.name, end, 8) == 0)
		{
			r = FindMasterDir(r->next, start);
			if(!r) return NULL;
		}
	}

	if(r && (t = malloc(r->dir.size)))
	{
		BasicWadSeek(r->wadfile, r->dir.start);
		BasicWadRead(r->wadfile, t, r->dir.size);

		new = (resource *)myalloc(sizeof(resource));
		new->data = t;
		new->name = strdup(name);
		new->ns = strdup(ns);

		new->next = ResourceList;
		ResourceList = new;

		if(strncmp(name, "ZETHTEST", 8) == 0)
		{
			while(r)
			{
				fprintf(stderr, "%s len %li (in %s)\n", r->dir.name, r->dir.size, ns);
				r = r->next;
			}
		}

		return t;
	}

	return NULL;
}

void ForgetResource(char *name)
{
	resource *p = ResourceList;
	resource **q = &ResourceList;
	while(p)
	{
		if(!strcmp(p->name, name))
		{
			*q = p->next;

			free(p->data);
			free(p->name);
			if(p->ns)
				free(p->ns);
			p->ns = NULL;

			free(p);
			p = *q;

			continue;
		}

		q = &(p->next);
		p = p->next;
	}
}

void ForgetAllResources()
{
	resource *p, *q;

	for(p = ResourceList; p; p = q)
	{
		q = p->next;
		free(p->data);
		free(p->name);
		if(p->ns)
			free(p->ns);
		free(p);
	}
	ResourceList = (resource *)NULL;
}


Tex *FindTexture(char *name, Tex **list, int size)
{
	int i;

	for(i = 0;i < size;++i)
	{
		if(strnicmp(list[i]->name, name, 8) == 0)
			return list[i];
	}

	return NULL;
}

Texture *GetTextureInfo(Tex *tex)
{
	long i;

	if(!tex)
		return NULL;

	if(tex->type == TYPE_TEXTURE)
	{
		SList p;

		for(p = Texture_sections; p; p = p->next)
		{
			void *tblock = GetResource(p->string);
			long *offsets = tblock;

			if(tblock)
			{
				for(i = 1; i <= offsets[0]; i++)
				{
					Texture *t = (Texture *)((long)offsets[i] + (long)tblock);
					if(!strnicmp(t->name, tex->name, 8)) /*jff don't care about case*/
						return t;
				}
			}
		}

		return NULL;
	}

	if(tex->type == TYPE_FLAT)
	{
		static Texture fakie = { "", 0, 0, 0, 0, 0, 0, 0, 0 };
		BITMAP *bmp;

		bmp = LoadTexture(tex);
		if(bmp)
		{
			strncpy(fakie.name, tex->name, 8);
			fakie.width = bmp->w;
			fakie.height = bmp->h;
			fakie.scalex = (fakie.width<<3) / 64;
			fakie.scaley = (fakie.height<<3) / 64;

			if(Caching == FALSE)
			{
				destroy_bitmap(tex->bmp);
				tex->bmp = NULL;
			}

			return &fakie;
		}

		return NULL;
	}

	if(tex->type == TYPE_TX || tex->type == TYPE_PNAME || tex->type == TYPE_SPRITE)
	{
		static Texture fakie = { "", 0, 0, 0, 0, 0, 0, 0, 0 };
		BITMAP *bmp;

		bmp = LoadTexture(tex);
		if(bmp)
		{
			strncpy(fakie.name, tex->name, 8);
			fakie.width = bmp->w;
			fakie.height = bmp->h;
			fakie.scalex = 8;
			fakie.scaley = 8;

			if(Caching == FALSE)
			{
				destroy_bitmap(tex->bmp);
				tex->bmp = NULL;
			}

			return &fakie;
		}

		return NULL;
	}

	return NULL;
}

// [RH] Named "Quick" because it's used by the fullscreen text browser--not because
// it's optimized for drawing textures quickly.
void DisplayWallQuick(BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *texname, BCINT scale)
{
	int	xsize, ysize;
	Texture  *tinfo;
	long xstep, ystep, xscale, yscale, step;
	Tex *tex;

	if (keypressed() && Caching == FALSE)
		return; 

	if (texname[0] == '-')
		return;

	if(scale <= 0)
		scale = 1;

	x1 = x0 + (x1-x0+1)/scale;
	y1 = y0 + (y1-y0)/scale;
	scale = 1;

	tex = FindTexture(texname, WTexture, WTexNum);
	tinfo = GetTextureInfo(tex);
	if (!tinfo)
	{
		SetColor(GRAY);
		DrawScreenLine(x0, y0, x1, y1 - 1);
		DrawScreenLine(x0, y1 - 1, x1, y0);
		return;
	}

	/* read the info for this texture */
	xsize = tinfo->width;
	ysize = tinfo->height;
	xscale = tinfo->scalex ? tinfo->scalex : 8;
	yscale = tinfo->scaley ? tinfo->scaley : 8;

	// [RH] Pick a scale appropriate for the cell
	xstep = (xsize<<19) / ((x1 - x0 + 1) * xscale);
	ystep = (ysize<<19) / ((y1 - y0) * yscale);
	step = MAX(65536, MAX(xstep, ystep));

	xstep = step * xscale / 8;
	ystep = step * yscale / 8;

	stretch_sprite(GetBuffer(), LoadTexture(tex), x0, y0, xsize * 65536 / xstep, ysize * 65536 / ystep);
	if(Caching == FALSE && tex->bmp)
	{
		destroy_bitmap(tex->bmp);
		tex->bmp = NULL;
	}
}


/*
   Function to get the size of a wall texture
   It's a lot smaller now that we have functions to get a pointer
   to the texture structure automagically :-)
   */

void GetWallTextureSize( BCINT *xsize_r, BCINT *ysize_r, char *texname)
{
	Texture *t = GetTextureInfo(FindTexture(texname, WTexture, WTexNum));

	if(t) {
		if(xsize_r) *xsize_r = t->width;
		if(ysize_r) *ysize_r = t->height;
	}
	else {
		/* texture data not found */
		if(xsize_r) *xsize_r = -1;
		if(ysize_r) *ysize_r = -1;
	}
}

void GetWallTextureScale(int *x_scale, int *y_scale, char *texname)
{
	Texture *t = GetTextureInfo(FindTexture(texname, WTexture, WTexNum));

	if(t) {
		if(x_scale) *x_scale = (int)(t->scalex ? t->scalex : 8);
		if(y_scale) *y_scale = (int)(t->scalex ? t->scalex : 8);
	}
	else {
		/* texture data not found */
		if(x_scale) *x_scale = 8;
		if(y_scale) *y_scale = 8;
	}
}


void DisplayTextureFit(BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *texname)
{
	int	xsize, ysize;
	Texture  *tinfo;
	long xstep, ystep, xscale, yscale, step;
	Tex *tex;

	if (keypressed() && Caching == FALSE)
		return; 

	if (texname[0] == '-')
		return;

	tex = FindTexture(texname, WTexture, WTexNum);
	tinfo = GetTextureInfo(tex);
	if (!tinfo)
	{
		SetColor(GRAY);
		DrawScreenLine(x0, y0, x1, y1 - 1);
		DrawScreenLine(x0, y1 - 1, x1, y0);
		return;
	}

	/* read the info for this texture */
	xsize = tinfo->width;
	ysize = tinfo->height;
	xscale = tinfo->scalex ? tinfo->scalex : 8;
	yscale = tinfo->scaley ? tinfo->scaley : 8;

	// [RH] Pick a scale appropriate for the cell
	xstep = (xsize<<19) / ((x1 - x0 + 1) * xscale);
	ystep = (ysize<<19) / ((y1 - y0) * yscale);
	step = MAX(65536, MAX(xstep, ystep));

	xstep = step * xscale / 8;
	ystep = step * yscale / 8;

	stretch_sprite(GetBuffer(), LoadTexture(tex), x0, y0, xsize * 65536 / xstep, ysize * 65536 / ystep);
	if(Caching == FALSE && tex->bmp)
	{
		destroy_bitmap(tex->bmp);
		tex->bmp = NULL;
	}
}

void DisplayTexture(BCINT x0, BCINT y0, BCINT x1, BCINT y1, Tex *tex)
{
	Texture  *tinfo;
	long xsize, ysize;
	long xstep, ystep, xscale, yscale, step;

	if (keypressed() && Caching == FALSE)
		return; 

	if(tex && tex->name[0] == '-')
		return;

	tinfo = GetTextureInfo(tex);
	if (!tinfo)
	{
		SetColor(GRAY);
		DrawScreenLine(x0, y0, x1, y1 - 1);
		DrawScreenLine(x0, y1 - 1, x1, y0);
		return;
	}

	/* read the info for this texture */
	xsize = tinfo->width;
	ysize = tinfo->height;
	xscale = tinfo->scalex ? tinfo->scalex : 8;
	yscale = tinfo->scaley ? tinfo->scaley : 8;

	// [RH] Pick a scale appropriate for the cell
	xstep = (xsize<<19) / ((x1 - x0) * xscale);
	ystep = (ysize<<19) / ((y1 - y0) * yscale);
	step = MAX(65536, MAX(xstep, ystep));

	xstep = step * xscale / 8;
	ystep = step * yscale / 8;

	stretch_sprite(GetBuffer(), LoadTexture(tex), x0, y0, xsize * 65536 / xstep, ysize * 65536 / ystep);
	if(Caching == FALSE && tex->bmp)
	{
		destroy_bitmap(tex->bmp);
		tex->bmp = NULL;
	}
}


/*
   choose a floor or ceiling texture
   */
void ChooseFloorTexture( BCINT x0, BCINT y0, char *prompt, int listsize, Tex **list, char *name)
{
	InputNameFromListWithFunc( x0, y0, prompt, listsize, list,
		/* [RH] */ ScrMaxY <= 480 ? 24  : 49,  name, ScrMaxX<800 ? 256 : 512,
		/* [RH] */ ScrMaxY <= 480 ? 256 : 512, DisplayTexture);
}

/*
   choose a wall texture
   */
void ChooseWallTexture( BCINT x0, BCINT y0, char *prompt, int listsize, Tex **list, char *name)
{
	InputNameFromListWithFunc( x0, y0, prompt, listsize, list,
		/* [RH] */ ScrMaxY <= 480 ? 24 : 49, name, ScrMaxX<800 ? 256 : 512/* jff */,
		/* [RH] */ ScrMaxY <= 480 ? 256 : 512, DisplayTexture);
}

/*
   function used by qsort to sort the sprite names
   */

/*
   choose a "sprite"
   */
void ChooseSprite( BCINT x0, BCINT y0, char *prompt, char *sname)
{
	return; /* disabled */

	MDirPtr dir;
	int n, i;
	char name[9];
	extern Tex *_WTexture;
	extern int NumSprites;
	extern int TotalTextures;
	Tex **list;

	memset(name, 0, sizeof(name));

	/* get the actual names from master dir. */
	dir = FindMasterDir( MasterDir, "S_START"); 
	if(!dir)
		return;

	dir = dir->next;

	i = TotalTextures - NumSprites;
	list = myalloc(NumSprites * sizeof(*list));
	for (n = 0; n < NumSprites; n++)
		list[n] = &(_WTexture[i++]);

	if (sname != NULL)
		strncpy(name, sname, 8);
	else
		strcpy(name, list[0]->name);

	ForgetAfter = TRUE;
	InputNameFromListWithFunc(x0, y0, prompt, NumSprites, list, 16, name, 320, 200, DisplaySprite);

	free(list);
}
